unit Shredder;
// Description: File/Disk Free Space Shredder (overwriter)
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.SDean12.org/
//
// -----------------------------------------------------------------------------
//


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  SDUGeneral, SDUProgress_U,
  FileList_U;

type
  TShredDetails = array [0..2] of byte;
  TShredBlock = array of byte;
  TShredFreeSpaceBlock = array of byte;

  TShredFreeSpaceBlockObj = class
  public
    BLANK_FREESPACE_BLOCK: TShredFreeSpaceBlock;
  end;

  // The array passed in is zero-indexed; populate elements zero to "bytesRequired"
  TGenerateOverwriteDataEvent = procedure (Sender: TObject; passNumber: integer; bytesRequired: cardinal; var outputBlock: TShredBlock) of object;


  TShredder = class(TComponent)
  private
    FFileDirUseInt: boolean;
    FFreeUseInt: boolean;
    FExtFileExe: string;
    FExtDirExe: string;
    FExtFreeSpaceExe: string;
    FExtShredFilesThenDir: boolean;
    FIntFirstBytes: longint;
    FIntGutmann: boolean;
    FIntPasses: integer;
    FIntFreeSpcFileSize: integer;
    FIntFreeSpcSmartFileSize: boolean;
    FIntFreeSpcFileCreationBlkSize: integer;
    FIntFileBufferSize: integer;

    FOnOverwriteDataReq: TGenerateOverwriteDataEvent;


    procedure ShredDir(dirname: string; silent: boolean);
    function  InternalShredFile(filename : string; quickShred: boolean; silent: boolean; silentProgressDlg: TSDUProgress_F; leaveFile: boolean = FALSE): integer;
    procedure DeleteFileOrDir(itemname: string);

    function  BytesPerCluster(filename: string): DWORD;

    procedure GetRandomDataBlock(passNum: integer; var outputBlock: TShredBlock);

    procedure GetGutmannBlock(passNum: integer; var outputBlock: TShredBlock);
    function  GetGutmannChars(passNum: integer): TShredDetails;
    function  SetGutmannDetails(nOne: integer; nTwo: integer; nThree: integer): TShredDetails;
    function  CreateEmptyFile(filename: string; size: int64; blankArray: TShredFreeSpaceBlockObj; progressDlg: TSDUProgress_F): integer;
    function  GetTempFilename(driveDir: string; serialNo: integer): string;
    function  WipeFileSlacksInDir(dirName: string; progressDlg: TSDUProgress_F; problemFiles: TStringList): integer;
    function  CountFiles(dirName: string): integer;
    function  GenerateRndDotFilename(path: string; origFilename: string): string;
  protected
    { Protected declarations }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy(); override;

    // Call this to shred all free space on the specified drive
    // Note: Does *not* wipe file slack - OverwriteAllFileSlacks(...) should be
    //       called before this function, if needed
    function  OverwriteDriveFreeSpace(driveLetter: char; silent: boolean = FALSE): integer;

    // Call this to a specific file's slack space
    function  OverwriteFileSlack(filename: string): boolean;

    // Call this to wipe all file slack on the specified drive
    function  OverwriteAllFileSlacks(driveLetter: char; silent: boolean = FALSE): integer;

    // Destroy the specified file or directory
    // Note that, with the internal shredder at any rate, calling WipeFileSlack
    // is not needed before calling this procedure
    procedure DestroyFileOrDir(itemname : string; quickShred: boolean; silent: boolean = FALSE);

    // Destroy the specified registry key
    procedure DestroyRegKey(key: string);

  published
    // Set this to TRUE to use the internal shredder when shredding
    // files/directories. Set to FALSE to use a 3rd party executable
    property FileDirUseInt: boolean read FFileDirUseInt write FFileDirUseInt;

    // Set this to TRUE to use the internal shredder when shredding
    // free space. Set to FALSE to use a 3rd party executable
    property FreeUseInt: boolean read FFreeUseInt write FFreeUseInt;

    // Command to be used when using a 3rd party executable to shred files
    property ExtFileExe : string read FExtFileExe  write FExtFileExe ;

    // Command to be used when using a 3rd party executable to shred directories
    property ExtDirExe : string read FExtDirExe  write FExtDirExe ;

    // Command to be used when using a 3rd party executable to free space
    property ExtFreeSpaceExe: string read FExtFreeSpaceExe write FExtFreeSpaceExe;

    // When shredding directories, and using a 3rd party executable to do so,
    // if all the files and subdirs within the directory to be destroyed must
    // be destroyed before the directory, this must be set to TRUE.
    property ExtShredFilesThenDir: boolean read FExtShredFilesThenDir write FExtShredFilesThenDir;

    // If "quickShred" is set to TRUE on any of the Destroy...(...) calls,
    // then only the first "FIntFirstBytes" bytes of files will be
    // overwritten before they are deleted
    property IntFirstBytes: longint read FIntFirstBytes write FIntFirstBytes;

    // Set to TRUE if Gutmann shredding should be used. Set to FALSE to use
    // a number of passes of pseudorandom data
    property IntGutmann: boolean read FIntGutmann write FIntGutmann;

    // If FIntGutmann is FALSE, then this number of passes of
    // pseudorandom data will be used for shredding
    property IntPasses : integer read FIntPasses  write FIntPasses ;

    // If FIntGutmann is FALSE, then (if set) this event will be called to
    // populate the buffer with pseudorandom data before writing.
    // IF SET, THE METHOD THAT THIS IS SET TO MUST POPULATE THE BUFFER IT IS
    // SUPPLIED WITH
    // If this is not set, then pseudorandom data will be used.
    property OnOverwriteDataReq: TGenerateOverwriteDataEvent read FOnOverwriteDataReq write FOnOverwriteDataReq default nil;

    // When shredding free space on a drive, the temp files that are created
    // and then overwritten should be of this size
    property IntFreeSpcFileSize: integer read FIntFreeSpcFileSize write FIntFreeSpcFileSize;

    // (Used in conjunction with IntFreeSpcFileSize)
    // If this is TRUE, then:
    //   f10 := The amount of free space is computed, and divided by 10
    //   fmax := max(5MB, f10)
    //   If fmax is less than IntFreeSpcFileSize, then fmax will be used
    //   If fmax is more than IntFreeSpcFileSize, then IntFreeSpcFileSize
    //   will be used
    // This means that if progress is being shown to the user, they'll always
    // get at least 10 "blocks" on the progress bar, making the application
    // appear more responsive; to actually be doing something
    property IntFreeSpcSmartFileSize: boolean read FIntFreeSpcSmartFileSize write FIntFreeSpcSmartFileSize;

    // When shredding free space on a drive, the temp files that are created
    // are made by repeatedly writing this blocksize amount of data until the
    // file reaches FIntFreeSpcFileSize, or the write fails (at which point,
    // the temp files are overwritten using the normal IntFileBufferSize as
    // with overwriting any other file)
    property IntFreeSpcFileCreationBlkSize: integer read FIntFreeSpcFileCreationBlkSize write FIntFreeSpcFileCreationBlkSize;

    // When shredding files, blocks of data which are of this buffer length are
    // written to the file to be destroyed until the whole file is overwritten
    property IntFileBufferSize: integer read FIntFileBufferSize write FIntFileBufferSize;


  end;

procedure Register;

implementation

uses Math,
     SDUFileIterator_U, SDUDirIterator_U,
     SDUDialogs,
     Registry;

type
  // Exceptions... These should *all* be handled internally
  EShredderError = Exception;
  EShredderErrorUserCancel = EShredderError;

  
const
  OVERWRITE_FREESPACE_TMP_DIR = '~STUfree';

  
procedure Register;
begin
  RegisterComponents('SDeanSecurity', [TShredder]);
end;


constructor TShredder.Create(AOwner: TComponent);
begin
  inherited;

  FFileDirUseInt:= TRUE;
  FFreeUseInt:= TRUE;
  FExtFileExe:= '';
  FExtDirExe:= '';
  FExtFreeSpaceExe:= '';
  FExtShredFilesThenDir:= FALSE;
  FIntFirstBytes:= 1024*1024;  // First MB by default
  FIntGutmann:= FALSE;
  FIntPasses:= 1;
  FIntFreeSpcFileSize:= 50*1024*1024;  // 50MB by default
  FIntFreeSpcSmartFileSize:= TRUE;
  FIntFreeSpcFileCreationBlkSize:= 1024*1024; // 1MB by default
  FIntFileBufferSize:= 1024*1024; // 1MB by default
  
end;


destructor TShredder.Destroy();
begin

  inherited;
end;



// Destroy a given file/dir, using the method specified in the INI file
// NOTE: When destroying dirs, QUICKSHRED IS ALWAYS FALSE - set in the dir
//       shredding procedure
// itemname       - the file/dir to be destroyed
// quickShred - ignored if not using internal shredding, otherwise if set to
//              FALSE then delete whole file; setting it to TRUE will only
//              delete the first n bytes
procedure TShredder.DestroyFileOrDir(itemname: string; quickShred: boolean; silent: boolean = FALSE);
var
{$IFDEF MSWINDOWS}
  fileAttributes : integer;
{$ENDIF}
  shredderCommandLine : string;
begin
  // Remove any hidden, system or readonly file attrib
{$IFDEF MSWINDOWS}
{$WARNINGS OFF}  // Useless warning about platform - we're already protecting
                 // against that!
  fileAttributes := FileGetAttr(itemname);
  fileAttributes := fileAttributes AND not(faReadOnly);
  fileAttributes := fileAttributes AND not(faHidden);
  fileAttributes := fileAttributes AND not(faSysFile);
  FileSetAttr(itemname, fileAttributes);
{$WARNINGS ON}
{$ENDIF}
{$IFDEF LINUX}
   xxx - to be implemented: remove any readonly attribute from the file
{$ENDIF}


  if (fileAttributes AND faDirectory)<>0 then
    begin
    ShredDir(itemname, silent);
    end
  else
    begin
    itemname := SDUConvertLFNToSFN(itemname);
    if FFileDirUseInt then
      begin
      InternalShredFile(itemname, quickShred, silent, nil);
      end
    else
      begin
      shredderCommandLine := format(FExtFileExe, [itemname]);
      WinExec(PChar(shredderCommandLine), SW_MINIMIZE);
      end;
    end;

end;


// Attempt to get the number of bytes/cluster on the drive the specified file is
// stored on
// Returns: The number of bytes per sector; or "1" if this cannot be
//          determined (e.g. the filename is a UNC)
function TShredder.BytesPerCluster(filename: string): DWORD;
var
  driveColon: string;
  retVal: DWORD;
  dwSectorsPerCluster: DWORD;
  dwBytesPerSector: DWORD;
  dwNumberOfFreeClusters: DWORD;
  dwTotalNumberOfClusters: DWORD;
begin
  // Attempt to get the number of bytes/sector
  retVal := 1;
  if (
      (Pos(':\', filename) = 2) or
      (Pos(':/', filename) = 2)
     ) then
    begin
    driveColon := filename[1]+':\';
    if GetDiskFreeSpace(
                        PChar(driveColon),       // address of root path
                        dwSectorsPerCluster,     // address of sectors per cluster
                        dwBytesPerSector,        // address of bytes per sector
                        dwNumberOfFreeClusters,  // address of number of free clusters
                        dwTotalNumberOfClusters  // address of total number of clusters
                       ) then
      begin
      retVal := (dwBytesPerSector * dwSectorsPerCluster);
      end;

    end;

  Result := retVal;
end;


// leaveFile - Set to TRUE to just leave the file after shredding; don't
//             delete it (default = FALSE). (Used when shredding files created
//             during drive freespace overwriting; the files must remain)
// silentProgressDlg - If silent is set to TRUE, then this may *optionally* be
//                     set to a progress dialog. This progress dialog will
//                     *only* be used for checking to see if the user's
//                     cancelled the operation
// Returns: -1 on error
//          -2 on user cancel
//           1 on success
function TShredder.InternalShredFile(
                                     filename: string;
                                     quickShred: boolean;
                                     silent: boolean;
                                     silentProgressDlg: TSDUProgress_F;
                                     leaveFile: boolean = FALSE
                                    ): integer;
var
  fileHandle : THandle;
  i: integer;
  blankingBytes: TShredBlock;
  bytesToShredLo: DWORD;
  bytesToShredHi: DWORD;
  bytesWritten: DWORD;
  numPasses: integer;
  progressDlg: TSDUProgress_F;
  gutmannStr: string;
  zero: DWORD;
  bpc: DWORD;
  bpcMod: DWORD;
  tmpDWORD: DWORD;
  bytesLeftToWrite: int64;
  failure: boolean;
  userCancel: boolean;
  retVal: integer;
  bytesToWriteNow: DWORD;
begin
  failure := FALSE;
  userCancel := FALSE;

  zero := 0;  // (Obviously!)
  bpc:= BytesPerCluster(filename);

  // Gutmann overwriting means 35 passes...
  numPasses := FIntPasses;
  if FIntGutmann then
    begin
    numPasses := 35;
    end;

  progressDlg:= TSDUProgress_F.create(nil);
  try
    try  // Exception handler for EShredderErrorUserCancel
      if not(silent) then
        begin
        progressDlg.Show();
        end;

      gutmannStr := '';
      if FIntGutmann then
        begin
        gutmannStr := 'Gutmann ';
        end;

      // Nuke any file attributes (e.g. readonly)
{$IFDEF MSWINDOWS}
{$WARN SYMBOL_PLATFORM OFF}  // Useless warning about platform - we're already 
                             // protecting against that!
      FileSetAttr(filename, 0);
{$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$IFDEF LINUX}
   xxx - to be implemented: strip off any readonly attribute from the file
{$ENDIF}

      fileHandle := CreateFile(PChar(filename),
                               GENERIC_READ or GENERIC_WRITE,
                               0,
                               nil,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL or FILE_FLAG_WRITE_THROUGH,
                               0);
      if (fileHandle = INVALID_HANDLE_VALUE) then
        begin
        failure := TRUE;
        end
      else
        begin
        try
          // Identify the size of the file being overwritten...
          bytesToShredLo := GetFileSize(fileHandle, @bytesToShredHi);
          // Check that the GetFileSize call was successful...
          if not( (bytesToShredLo = $FFFFFFFF) AND (GetLastError()<>NO_ERROR) ) then
            begin
            // Adjust to increase up to the nearest bytes per sector boundry
            // Note: This makes the reasonable assumption that ($FFFFFFFF + 1)
            //       will always be a multiple of "bpc"; so we don't need to
            //       bother involving the high DWORD of the file's size in this
            //       calculation
            bpcMod := (bytesToShredLo mod bpc);
            if (bpcMod>0) then
              begin
              tmpDWORD := bytesToShredLo + (bpc - bpcMod);
              // In case that causes an overflow...
              if (bytesToShredLo > tmpDWORD) then
                begin
                inc(bytesToShredHi);
                end;
              bytesToShredLo := tmpDWORD;
              end;

            // Sanity check - we can only handle files less than 2^63 bytes long
            bytesLeftToWrite := (bytesToShredHi shl 32) AND bytesToShredLo;
            if (bytesLeftToWrite < 0) then
              begin
              // Do nothing - retVal defaults to error
              end
            else
              begin
              if quickShred then
                begin
                // Note: Cast to int64 to ensure that no problems with truncation
                bytesLeftToWrite := min(bytesLeftToWrite, int64(FIntFirstBytes));
                bytesToShredHi := (bytesLeftToWrite shr 32);
                bytesToShredLo := (bytesLeftToWrite AND $FFFFFFFF);
                end;

              progressDlg.i64Min := 0;
              progressDlg.i64Max := bytesLeftToWrite;

              for i:=1 to numPasses do
                begin
                // Let user know what's going on...
                progressDlg.Caption := 'Shredding ('+gutmannStr+'pass '+inttostr(i)+'/'+inttostr(numPasses)+') '+filename;
                progressDlg.i64Position := 0;

                // Has user cancelled?
                Application.ProcessMessages();
                if progressDlg.Cancel then
                  begin
                  raise EShredderErrorUserCancel.Create('User cancelled');
                  end;
                if (silentProgressDlg <> nil) then
                  begin
                  if silentProgressDlg.Cancel then
                    begin
                    raise EShredderErrorUserCancel.Create('User cancelled');
                    end;
                  end;

                // Reset the file ptr
                SetFilePointer(fileHandle, 0, @zero, FILE_BEGIN);

                // Reset the total number of bytes to be written...
                bytesLeftToWrite := (bytesToShredHi shl 32) + bytesToShredLo;

                // Fill a block with random garbage
                SetLength(blankingBytes, FIntFileBufferSize);
                if not(assigned(FOnOverwriteDataReq)) then
                  begin
                  GetRandomDataBlock(i, blankingBytes);
                  end;


                while (
                       (bytesLeftToWrite > 0) AND
                       not(failure)
                      ) do
                  begin
                  progressDlg.i64Position := (progressDlg.i64Max - bytesLeftToWrite);

                  // Has user cancelled?
                  Application.ProcessMessages();
                  if progressDlg.Cancel then
                    begin
                    raise EShredderErrorUserCancel.Create('User cancelled');
                    end;
                  if (silentProgressDlg <> nil) then
                    begin
                    if silentProgressDlg.Cancel then
                      begin
                      raise EShredderErrorUserCancel.Create('User cancelled');
                      end;
                    end;

                  // Generate new data if user supplied routine...
                  if (assigned(FOnOverwriteDataReq)) then
                    begin
                    GetRandomDataBlock(i, blankingBytes);
                    end;

                  // Note: Cast to int64 to ensure no problems
                  bytesToWriteNow := (min(bytesLeftToWrite, int64(FIntFileBufferSize)) AND $FFFFFFFF);
                  failure := not(WriteFile(fileHandle, blankingBytes[0], bytesToWriteNow, bytesWritten, nil));

                  dec(bytesLeftToWrite, bytesWritten);

                  // Ensure that the buffer is flushed to disk (even through disk caching
                  // software) [from Borland FAQ]
  // xxx - this line commented out - prevents process from exiting?! It remains
  //       visible in task manager, and can't be killed?!
  //                FlushFileBuffers(fileHandle);
                  end;


                if (failure) then
                  begin
                  // Get out of loop...
                  break;
                  end;

                end;  // for i:=1 to numPasses do

              end;  // ELSE PART - if (bytesLeftToWrite < 0) then
            end;  // if not( (fileLengthLo = $FFFFFFFF) AND (GetLastError()<>NO_ERROR) ) then
        finally
  // xxx - this line commented out - prevents process from exiting?! It remains
  //       visible in task manager, and can't be killed?!
  //        FlushFileBuffers(fileHandle);
          CloseHandle(fileHandle);
        end;

        end;  // ELSE PART - if (fileHandle = INVALID_HANDLE_VALUE) then

    except
      on EShredderErrorUserCancel do
        begin
        // Ensure flag set
        userCancel := TRUE;
        end;

    end;

  finally
    progressDlg.Free();
  end;


  // Clean up file?
  if not(leaveFile) then
    begin
    DeleteFileOrDir(filename);
    end;


  // Determine return value...
  if (failure) then
    begin
    retVal := -1;
    end
  else if (userCancel) then
    begin
    retVal := -2;
    end
  else
    begin
    // Everything OK...
    retVal := 1;
    end;


  Result := retVal;

end;

// Simple rename a file/dir and then delete it.
procedure TShredder.DeleteFileOrDir(itemname: string);
const
  // This should be enough to overwrite any LFN directory entries (about 255 chars long)
  MASSIVE_FILENAME = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa.aaa';
var
  j: integer;
  deleteFilename: string;
  testRenameFilename: string;
  fileAttributes: integer;
  largeFilename: string;
  fileHandle : THandle;
  tmpFsp: string;
  zero: DWORD;
  fileIterator: TSDUFileIterator;
  currFile: string;
begin
  zero := 0; // Obviously!
  try
    tmpFsp := GenerateRndDotFilename(ExtractFilePath(itemname), ExtractFileName(itemname));
    tmpFsp := ExtractFilePath(itemname)+tmpFsp;
    if tmpFsp<>'' then
      begin
      if RenameFile(itemname, tmpFsp) then
        begin
        itemname := tmpFsp;
        end;
      end;

    deleteFilename := itemname;
    if (Win32Platform=VER_PLATFORM_WIN32_NT) then
      begin
      for j:=ord('a') to ord('z') do
        begin
        testRenameFilename := ExtractFilePath(itemname) + chr(j)+'.';
        if not(fileexists(testRenameFilename)) then
          begin
          deleteFilename := testRenameFilename;
          break;
          end;
        end;

      largeFilename := ExtractFilePath(itemname) + MASSIVE_FILENAME;
      if length(largeFilename)>MAX_PATH then
        begin
        Delete(largeFilename, MAX_PATH-1, length(largeFilename)-MAX_PATH+1);
        end;


      if RenameFile(itemname, largeFilename) then
        begin
        if not(RenameFile(largeFilename, deleteFilename)) then
          begin
          deleteFilename := largeFilename;
          end;
        end
      else
        begin
        deleteFilename := itemname;
        end;;

      end;

{$IFDEF MSWINDOWS}
{$WARN SYMBOL_PLATFORM OFF}  // Useless warning about platform - we're already 
                             // protecting against that!
    fileAttributes := FileGetAttr(deleteFilename);
    if (fileAttributes AND faDirectory)=0 then
{$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$IFDEF LINUX}
   xxx - to be implemented: test if it's *not* a directory
{$ENDIF}
      begin
      // Truncate the file to 0 bytes and set it's date/time to some junk
      fileHandle := CreateFile(PChar(deleteFilename),
                               GENERIC_READ or GENERIC_WRITE,
                               0,
                               nil,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL or FILE_FLAG_WRITE_THROUGH,
                               0);
      SetFilePointer(fileHandle, 0, @zero, FILE_BEGIN);
      SetEndOfFile(fileHandle);

      SDUSetAllFileTimes(fileHandle, DateTimeToFileDate(encodedate(1980, 1, 1)));

      CloseHandle(fileHandle);

      DeleteFile(deleteFilename);
      end
    else
      begin
      // Delete all files and dirs beneath the directory
      fileIterator := TSDUFileIterator.Create(nil);
      try
        fileIterator.Directory := deleteFilename;
        fileIterator.RecurseSubDirs := FALSE;  // We handle recursion
        fileIterator.IncludeDirNames := TRUE;
        fileIterator.OmitStartDirPrefix := FALSE;
        fileIterator.Reset();

        currFile := fileIterator.Next();
        while (currFile<>'') do
          begin
          DeleteFileOrDir(currFile);
          currFile := fileIterator.Next();
          end;

      finally
        fileIterator.Free();
      end;

      // Finally, remove the dir itsself...
      RemoveDir(deleteFilename);
      end;
  except
    // Nothing (i.e. ignore all exceptions, e.g. can't open file)
  else
    // Nothing (i.e. ignore all exceptions, e.g. can't open file)
  end;

end;


procedure TShredder.GetRandomDataBlock(passNum: integer; var outputBlock: TShredBlock);
begin
  if FIntGutmann then
    begin
    GetGutmannBlock(passNum, outputBlock);
    end
  else
    begin
    if assigned(FOnOverwriteDataReq) then
      begin
      FOnOverwriteDataReq(self, passNum, (high(outputBlock)-low(outputBlock)), outputBlock);
      end
    else
      begin
      // Gutmann 1st pass is pseudorandom bytes
      GetGutmannBlock(1, outputBlock);
      end;

    end;

end;



procedure TShredder.GetGutmannBlock(passNum: integer; var outputBlock: TShredBlock);
var
  i: integer;
  passDetails: TShredDetails;
begin
  if (passNum<5) OR (passNum>31) then
    begin
    for i:=low(outputBlock) to high(outputBlock) do
      begin
      outputBlock[i] := random(256);
      end;
    end
  else
    begin
    passDetails := GetGutmannChars(passNum);
    for i:=low(outputBlock) to high(outputBlock) do
      begin
      outputBlock[i] := passDetails[i mod 3];
      end;
    end;

end;

function TShredder.GetGutmannChars(passNum: integer): TShredDetails;
begin
  case passnum of
   5: Result := SetGutmannDetails($55, $55, $55);
   6: Result := SetGutmannDetails($aa, $aa, $aa);
   7: Result := SetGutmannDetails($92, $49, $24);
   8: Result := SetGutmannDetails($49, $24, $92);
   9: Result := SetGutmannDetails($24, $92, $49);
  10: Result := SetGutmannDetails($00, $00, $00);
  11: Result := SetGutmannDetails($11, $11, $11);
  12: Result := SetGutmannDetails($22, $22, $22);
  13: Result := SetGutmannDetails($33, $33, $33);
  14: Result := SetGutmannDetails($44, $44, $44);
  15: Result := SetGutmannDetails($55, $55, $55);
  16: Result := SetGutmannDetails($66, $66, $66);
  17: Result := SetGutmannDetails($77, $77, $77);
  18: Result := SetGutmannDetails($88, $88, $88);
  19: Result := SetGutmannDetails($99, $99, $99);
  20: Result := SetGutmannDetails($aa, $aa, $aa);
  21: Result := SetGutmannDetails($bb, $bb, $bb);
  22: Result := SetGutmannDetails($cc, $cc, $cc);
  23: Result := SetGutmannDetails($dd, $dd, $dd);
  24: Result := SetGutmannDetails($ee, $ee, $ee);
  25: Result := SetGutmannDetails($ff, $ff, $ff);
  26: Result := SetGutmannDetails($92, $49, $24);
  27: Result := SetGutmannDetails($49, $24, $92);
  28: Result := SetGutmannDetails($24, $92, $49);
  29: Result := SetGutmannDetails($6d, $b6, $db);
  30: Result := SetGutmannDetails($b6, $db, $6d);
  31: Result := SetGutmannDetails($db, $6d, $b6);
  end;

end;

function TShredder.SetGutmannDetails(nOne: integer; nTwo: integer; nThree: integer): TShredDetails;
begin
  Result[0] := nOne;
  Result[1] := nTwo;
  Result[2] := nThree;

end;

// Returns: -1 on error
//          -2 on user cancel
//           1 on success
function TShredder.OverwriteDriveFreeSpace(driveLetter: char; silent: boolean): integer;
const
  FIVE_MB = 5*1024*1024;
var
  drive: string;
  tempDriveDir: string;
  freeSpace: int64;
  fileNumber: integer;
  currFilename: string;
  blankArray: TShredFreeSpaceBlockObj;
  i: integer;
  lastFilename: string;
  shredderCommandLine: string;
  progressDlg: TSDUProgress_F;
  diskNumber: integer;
  userCancel: boolean;
  failure: boolean;
  retVal: integer;
  createOK: integer;
  internalShredOK: integer;
  useTmpFileSize: int64;
  prevCursor: TCursor;
begin
  userCancel := FALSE;
  failure := FALSE;

  if not(FFreeUseInt) then
    begin
    shredderCommandLine := format(FExtFreeSpaceExe, [driveLetter]);
    if (WinExec(PChar(shredderCommandLine), SW_RESTORE))<31 then
      begin
      failure := TRUE;
      SDUMessageDlg('Error running external (3rd party) free space shredder',
                 mtError,
                 [mbOK],
                 0);
      end;

    end
  else
    begin
    progressDlg := TSDUProgress_F.Create(nil);
    try
      blankArray := TShredFreeSpaceBlockObj.Create();
      try
        SetLength(blankArray.BLANK_FREESPACE_BLOCK, FIntFreeSpcFileCreationBlkSize);
        for i:=0 to FIntFreeSpcFileCreationBlkSize-1 do
          begin
          blankArray.BLANK_FREESPACE_BLOCK[i] := 0;
          end;

        // Create a subdir
        drive := uppercase(driveLetter);
        driveLetter := drive[1];
        tempDriveDir := driveLetter + ':\'+OVERWRITE_FREESPACE_TMP_DIR+inttostr(random(10000))+'.tmp';
        diskNumber := ord(drive[1])-ord('A')+1;

        CreateDir(tempDriveDir);
        fileNumber := 0;

        // While there is FIntFreeSpcFileSize (or smart) bytes diskspace
        // left, create a file FIntFreeSpcFileSize (or smart) big
        freeSpace := DiskFree(diskNumber);
        progressDlg.Caption := 'Shredding free space on drive '+driveLetter+':';
        progressDlg.i64Max := freeSpace;
        progressDlg.i64Min := 0;
        progressDlg.i64Position := 0;
        prevCursor := Screen.Cursor;
        if not(silent) then
          begin
          Screen.Cursor := crAppStart;

          progressDlg.Show();
          end;

        try  // Finally (mouse cursor revert)

          try  // Except (EShredderErrorUserCancel)

            try  // Finally

              if FIntFreeSpcSmartFileSize then
                begin
                useTmpFileSize:= max((freeSpace div 10), FIVE_MB);
                if (useTmpFileSize >= int64(FIntFreeSpcFileSize)) then
                  begin
                  useTmpFileSize := FIntFreeSpcFileSize;
                  end;
                end
              else
                begin
                useTmpFileSize:= FIntFreeSpcFileSize;
                end;



              // This is > and not >= so that the last file to be created (outside this
              // loop) isn't zero bytes long
              while (freeSpace>useTmpFileSize) do
                begin
                inc(fileNumber);
                currFilename := GetTempFilename(tempDriveDir, fileNumber);

                createOK := CreateEmptyFile(currFilename, useTmpFileSize, blankArray, progressDlg);
                if (createOK = -2) then
                  begin
                  raise EShredderErrorUserCancel.Create('User cancelled');
                  end
                else if (createOK = -1) then
                  begin
                  failure := TRUE;
                  // Quit loop...
                  break;
                  end;


                // Shred the file, but _don't_ _delete_ _it_
                // Note that this will overwrite any slack space at the end of the file
                internalShredOK := InternalShredFile(currFilename, FALSE, TRUE, progressDlg, TRUE);
                if (internalShredOK = -2) then
                  begin
                  raise EShredderErrorUserCancel.Create('User cancelled');
                  end
                else if (internalShredOK = -1) then
                  begin
                  failure := TRUE;
                  // Quit loop...
                  break;
                  end;


                freeSpace := DiskFree(diskNumber);
                progressDlg.i64InversePosition := freeSpace;

                // Check for user cancel...
                Application.ProcessMessages();
                if progressDlg.Cancel then
                  begin
                  raise EShredderErrorUserCancel.Create('User cancelled');
                  end;

                end;  // while (freeSpace>useTmpFileSize) do


              // Create a file with the remaining disk bytes
              if not(failure) then
                begin
                lastFilename := GetTempFilename(tempDriveDir, fileNumber+1);
                createOK := CreateEmptyFile(lastFilename, freeSpace, blankArray, progressDlg);
                if (createOK = -2) then
                  begin
                  raise EShredderErrorUserCancel.Create('User cancelled');
                  end
                else if (createOK = -1) then
                  begin
                  failure := TRUE;
                  end;

                end;  // if not(failure) then


              // Shred the last file
              if not(failure) then
                begin
                internalShredOK := InternalShredFile(lastFilename, FALSE, TRUE, progressDlg, TRUE);
                // Note that this will overwrite any slack space at the end of the file
                if (internalShredOK = -2) then
                  begin
                  raise EShredderErrorUserCancel.Create('User cancelled');
                  end
                else if (internalShredOK = -1) then
                  begin
                  failure := TRUE;
                  end;

                end;  // if not(failure) then


            finally
              // Remove any files that were created, together with the dir
              DeleteFileOrDir(tempDriveDir);
            end;

          except
            on EShredderErrorUserCancel do
              begin
              // Ensure flag set
              userCancel := TRUE;
              end;

          end;


        finally
          // Revert the mouse pointer, if we were showing overwrite
          // progress...
          if not(silent) then
            begin
            Screen.Cursor := prevCursor;
            end;
        end;

        
      finally
        blankArray.Free();
      end;

    finally
      progressDlg.Free();
    end;

  end; // use internal free space shredder


  // Determine return value...
  if (failure) then
    begin
    retVal := -1;
    end
  else if (userCancel) then
    begin
    retVal := -2;
    end
  else
    begin
    // Everything OK...
    retVal := 1;
    end;


  Result := retVal;
end;


// progressDlg - This may *optionally* be
//               set to a progress dialog. This progress dialog will
//               *only* be used for checking to see if the user's
//               cancelled the operation
// Returns: -1 on error
//          -2 on user cancel
//           1 on success
function TShredder.CreateEmptyFile(filename: string; size: int64; blankArray: TShredFreeSpaceBlockObj; progressDlg: TSDUProgress_F): integer;
var
  retVal: integer;
  userCancel: boolean;
  failure: boolean;
  fileHandle : THandle;
  bytesWritten: DWORD;
  bytesInBlock: DWORD;
  totalBytesWritten: int64;
begin
  failure := FALSE;
  userCancel := FALSE;

  fileHandle := CreateFile(PChar(filename),
                           GENERIC_READ or GENERIC_WRITE,
                           0,
                           nil,
                           CREATE_ALWAYS,
                           FILE_ATTRIBUTE_NORMAL or FILE_FLAG_WRITE_THROUGH,
                           0);

  if (fileHandle = INVALID_HANDLE_VALUE) then
    begin
    failure := TRUE;
    end
  else
    begin
    try
      // Fill out the file to the required size
      totalBytesWritten := 0;
      // Set bytesWritten to bootstrap loop
      bytesWritten := 1;
      while ( (totalBytesWritten<size) and (bytesWritten>0) )do
        begin
        bytesInBlock := min((size-totalBytesWritten), FIntFreeSpcFileCreationBlkSize);
        WriteFile(fileHandle, blankArray.BLANK_FREESPACE_BLOCK[0], bytesInBlock, bytesWritten, nil);
        inc(totalBytesWritten, bytesWritten);

        // Has user cancelled?
        Application.ProcessMessages();
        if (progressDlg <> nil) then
          begin
          if progressDlg.Cancel then
            begin
            raise EShredderErrorUserCancel.Create('User cancelled');
            end;
          end;

        end;

      // Ensure that the buffer is flushed to disk (even through disk caching
      // software) [from Borland FAQ]
    // xxx - this line commented out - prevents process from exiting?! It remains
    //       visible in task manager, and can't be killed?!
    //  FlushFileBuffers(fileHandle);

    finally
      CloseHandle(fileHandle);
    end;

    end;  // ELSE PART -  if (fileHandle = INVALID_HANDLE_VALUE) then
    

  // Determine return value...
  if (failure) then
    begin
    retVal := -1;
    end
  else if (userCancel) then
    begin
    retVal := -2;
    end
  else
    begin
    // Everything OK...
    retVal := 1;
    end;


  Result := retVal;
end;


function TShredder.GetTempFilename(driveDir: string; serialNo: integer): string;
begin
  Result := driveDir + '\~STUtmp.'+inttostr(serialNo);
end;



// Returns: -1 on error
//          -2 on user cancel
//           1 on success
function TShredder.OverwriteAllFileSlacks(driveLetter: char; silent: boolean): integer;
var
  progressDlg: TSDUProgress_F;
  rootDir: string;
  problemFiles: TStringList;
  reportDlg: TFileList_F;
  drive: string;
begin
  drive := uppercase(driveLetter);
  driveLetter := drive[1];
  progressDlg := TSDUProgress_F.Create(nil);
  problemFiles:= TStringList.create();
  try
    rootDir:= driveLetter+':\';

    progressDlg.Caption := 'Shredding file slack on drive '+driveLetter+':';
    progressDlg.i64Max := CountFiles(rootDir);
    progressDlg.i64Min := 0;
    progressDlg.i64Position := 0;
    if not(silent) then
      begin
      progressDlg.Show();
      end;

    Result := WipeFileSlacksInDir(rootDir, progressDlg, problemFiles);

    if not(silent) AND (problemFiles.count>0) then
      begin
      reportDlg := TFileList_F.Create(nil);
      try
        reportDlg.lbFiles.visible := TRUE;
        reportDlg.lblTitle.caption := 'The following files could not have their slack space shredded:';
        reportDlg.lbFiles.items.assign(problemFiles);
        reportDlg.showmodal;
      finally
        reportDlg.Free();
      end;
      end;
  finally
    problemFiles.Free();
    progressDlg.Free();
  end;

end;


// Perform file slack shredding on all files in specified dir
// Returns: -1 on error
//          -2 on user cancel
//           1 on success
function TShredder.WipeFileSlacksInDir(dirName: string; progressDlg: TSDUProgress_F; problemFiles: TStringList): integer;
var
  slackFile: string;
  fileIterator: TSDUFileIterator;
  currFile: string;
  retval: integer;
begin
  retval := 1;
  fileIterator := TSDUFileIterator.Create(nil);
  try
    fileIterator.Directory := dirName;
    fileIterator.RecurseSubDirs := TRUE;
    fileIterator.IncludeDirNames := FALSE;
    fileIterator.OmitStartDirPrefix := FALSE;

    fileIterator.Reset();

    currFile := fileIterator.Next();
    while (currFile<>'') do
      begin
      slackFile := SDUConvertLFNToSFN(currFile);
      if not(OverwriteFileSlack(slackFile)) then
        begin
        problemFiles.add(slackFile);
        end;

      progressDlg.i64IncPosition();
      if ProgressDlg.Cancel then
        begin
        retval := -2;
        break;
        end;
      currFile := fileIterator.Next();
      end;

  finally
    fileIterator.Free();
  end;

  Result := retval;
end;


function TShredder.OverwriteFileSlack(filename: string): boolean;
var
  fileHandle: THandle;
  fileLengthLo: DWORD;
  fileLengthHi: DWORD;
  numPasses: integer;
  i: integer;
  blankingBytes: TShredBlock;
  bytesWritten: DWORD;
  fileDateStamps: integer;
  fileAttributes : integer;
  retVal: boolean;
  slackSize: DWORD;
  writeFailed: boolean;
  bpc: DWORD;
  bpcMod: DWORD;
begin
  // Record any file attributes and remove any hidden, system or readonly file
  // attribs in order to put them back later
{$IFDEF MSWINDOWS}
{$WARN SYMBOL_PLATFORM OFF}  // Useless warning about platform - we're already 
                             // protecting against that!
  fileAttributes := FileGetAttr(filename);
  // If got attributes OK, continue...
  retVal := (fileAttributes <> -1);
  if (fileAttributes <> -1) then
{$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$IFDEF LINUX}
   xxx - to be implemented: get all file attributes
{$ENDIF}
    begin
{$IFDEF MSWINDOWS}
{$WARN SYMBOL_PLATFORM OFF}  // Useless warning about platform - we're already 
                             // protecting against that!
    if (FileSetAttr(filename, faArchive) = 0) then
{$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$IFDEF LINUX}
   xxx - to be implemented: set file attributes to ensure that we can
{$ENDIF}
      begin
      fileHandle := CreateFile(PChar(filename),
                               GENERIC_READ or GENERIC_WRITE,
                               0,
                               nil,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL or FILE_FLAG_WRITE_THROUGH,
                               0);
      if (fileHandle<>INVALID_HANDLE_VALUE) then
        begin
        // Get the date/timestamps before we start changing the file...
        fileDateStamps := FileGetDate(fileHandle);

        fileLengthLo := GetFileSize(fileHandle, @fileLengthHi);
        // Check that the GetFileSize call was successful...
        if not( (fileLengthLo = $FFFFFFFF) AND (GetLastError()<>NO_ERROR) ) then
          begin
          bpc := BytesPerCluster(filename);
          // If the bytes per sector was reported as "1", as assume that it
          // couldn't be determined, and assume a much larger sector size
          if (bpc=1) then
            begin
            bpc := 128*1024;  // 128K - In practice, it's likely to be much less than this
            end;

          // Reserve block for random garbage
          SetLength(blankingBytes, bpc);

          // Determine the amount of slack space beyond the file to fill the
          // sector
          // Note: This makes the reasonable assumption that ($FFFFFFFF + 1)
          //       will always be a multiple of "bpc"; so we don't need to
          //       bother involving the high DWORD of the file's size in this
          //       calculation
          slackSize := 0;
          bpcMod := (fileLengthLo mod bpc);
          if (bpcMod>0) then
            begin
            slackSize := (bpc - bpcMod);
            end;

          // Determine the number of passes...
          numPasses := FIntPasses;
          if FIntGutmann then
            begin
            numPasses := 35;
            end;

          // ...and finally, perform the slack space wipe
          writeFailed := FALSE;
          for i:=1 to numPasses do
            begin
            // Fill a block with random garbage
            GetRandomDataBlock(i, blankingBytes);

            // Set the file pointer to the end of the file
            SetFilePointer(fileHandle, fileLengthLo, @fileLengthHi, FILE_BEGIN);

            // Write from there, pass data (e.g. a block of Gutmann)
            WriteFile(fileHandle, blankingBytes[0], slackSize, bytesWritten, nil);
            if (bytesWritten<>slackSize) then
              begin
              writeFailed := TRUE;
              break;
              end;

            // Ensure that the buffer is flushed to disk (even through disk caching
            // software) [from Borland FAQ]
// xxx - this line commented out - prevents process from exiting?! It remains
//       visible in task manager, and can't be killed?!
//            FlushFileBuffers(fileHandle);
            end; // for each pass

          // Truncate the file back down to the correct length
          SetFilePointer(fileHandle, fileLengthLo, @fileLengthHi, FILE_BEGIN);
          SetEndOfFile(fileHandle);

          retVal := not(writeFailed);
          end;


        // Reset the date/timestamps
{$WARN SYMBOL_PLATFORM OFF} // Don't care that this is platform specific - if anyone needs this under Kylix, let me know though!
        FileSetDate(fileHandle, fileDateStamps);
{$WARN SYMBOL_PLATFORM ON}

        // Flush and close
// xxx - this line commented out - prevents process from exiting?! It remains
//       visible in task manager, and can't be killed?!
//        FlushFileBuffers(fileHandle);
        CloseHandle(fileHandle);

        end;  // if (fileHandle<>INVALID_HANDLE_VALUE) then

      // Reset the file attributes
{$IFDEF MSWINDOWS}
{$WARN SYMBOL_PLATFORM OFF}  // Useless warning about platform - we're already 
                             // protecting against that!
      FileSetAttr(filename, fileAttributes);
{$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$IFDEF LINUX}
   xxx - to be implemented: reset file attributes to those stored previously
{$ENDIF}
      end;  // if (FileSetAttr(filename, faArchive) = 0) then

    end;  // if (fileAttributes <> -1) then

  Result := retVal;
end;


// Perform dir shredding, using the method specified in the INI file
// filename - the file to be destroyed
procedure TShredder.ShredDir(dirname: string; silent: boolean);
var
  dirToDestroy: string;
  shredderCommandLine: string;
  fileIterator: TSDUFileIterator;
  dirIterator: TSDUDirIterator;
  currFile: string;
  currDir: string;
begin
  if (not(FFileDirUseInt))         AND
     (not(FExtShredFilesThenDir))  AND
     (FExtDirExe<>'')              then
    begin
    // i.e. we using an external shredder which doesn't need all files to be
    // removed before it can be used
    dirname := SDUConvertLFNToSFN(dirname);
    shredderCommandLine := format(FExtDirExe, [dirname]);
    winexec(PChar(shredderCommandLine), SW_MINIMIZE);
    end
  else
    begin
    // Shred all files in the current directory
    fileIterator := TSDUFileIterator.Create(nil);
    try
      fileIterator.Directory := dirname;
      fileIterator.RecurseSubDirs := TRUE;
      fileIterator.IncludeDirNames := FALSE;
      fileIterator.OmitStartDirPrefix := FALSE;
      fileIterator.Reset();

      currFile := fileIterator.Next();
      while (currFile<>'') do
        begin
        DestroyFileOrDir(currFile,
                         FALSE,
                         silent);
        currFile := fileIterator.Next();
        end;
    finally
      fileIterator.Free();
    end;


    dirIterator := TSDUDirIterator.Create(nil);
    try
      dirIterator.Directory := dirName;
      dirIterator.ReverseFormat := TRUE;
      dirIterator.IncludeStartDir := TRUE;
      dirIterator.Reset();

      // Now do the dir structure
      currDir := dirIterator.Next();
      while currDir<>'' do
        begin
        // And finally, remove the current dir...
        //   if external shredder handles dirs, use it
        //   else pass the dirname to the internal shredder for shredding
        if (not(FFileDirUseInt)) AND
           (FExtDirExe<>'')      then
          begin
          // i.e. we using an external shredder which doesn't need all files to be
          // removed before it can be used
          dirToDestroy := SDUConvertLFNToSFN(currDir);
          shredderCommandLine := format(FExtDirExe, [dirToDestroy]);
          WinExec(PChar(shredderCommandLine), SW_MINIMIZE);
          end
        else
          begin
          // Fallback to simply removing the dir using internal method
          DeleteFileOrDir(currDir);
          end;

        currDir := dirIterator.Next();
        end;
    finally
      dirIterator.Free();
    end;

    end; // External shredder not being used/needs files deleted first

end;

procedure TShredder.DestroyRegKey(key: string);
var
  registry: TRegistry;
  NTsubkeys: TStrings;
  keyValues: TStrings;
  valueInfo: TRegDataInfo;
  rootStr: string;
  i: integer;
  j: integer;
  buffer: array of byte;
begin
  registry := TRegistry.create();
  try
    registry.LazyWrite := FALSE;
    
    if Pos('HKCR\', key)=1 then
      begin
      registry.RootKey := HKEY_CLASSES_ROOT;
      end
    else if Pos('HKCU\', key)=1 then
      begin
      registry.RootKey := HKEY_CURRENT_USER;
      end
    else if Pos('HKLM\', key)=1 then
      begin
      registry.RootKey := HKEY_LOCAL_MACHINE;
      end
    else if ( (Pos('HKU \', key)=1) or (Pos('HKU\', key)=1) ) then
      begin
      registry.RootKey := HKEY_USERS;
      end
    else if Pos('HKCC\', key)=1 then
      begin
      registry.RootKey := HKEY_CURRENT_CONFIG;
      end
    else if Pos('HKDD\', key)=1 then
      begin
      registry.RootKey := HKEY_DYN_DATA;
      end;

    rootStr := Copy(key, 1, 5);
    Delete(key, 1, 5);

    if (Win32Platform=VER_PLATFORM_WIN32_NT) then
      begin
      NTsubkeys:=TStringList.Create();
      try
        if registry.OpenKey(key, FALSE) then
          begin
          if registry.HasSubkeys() then
            begin
            registry.GetKeyNames(NTsubkeys);
            end;

          keyValues := TStringList.Create();
          try
            registry.GetValueNames(keyValues);
            for i:=0 to (keyValues.count-1) do
              begin
              registry.GetDataInfo(keyValues[i], valueInfo);
              case valueInfo.RegData of
              rdString:
                begin
                registry.WriteString(keyValues[i],
                                      Format('%-'+inttostr(valueInfo.DataSize-1)+'.'+inttostr(valueInfo.DataSize-1)+'s', ['']));
                end;

              rdExpandString:
                begin
                registry.WriteExpandString(keyValues[i],
                                      Format('%-'+inttostr(valueInfo.DataSize-1)+'.'+inttostr(valueInfo.DataSize-1)+'s', ['']));
                end;

              rdInteger:
                begin
                registry.WriteInteger(keyValues[i], 0);
                end;

              rdBinary:
                begin
                setlength(buffer, valueInfo.DataSize);
                for j:=0 to (valueInfo.DataSize-1) do
                  begin
                  buffer[j] := $FF;
                  end;
                registry.WriteBinaryData(keyValues[i],
                                         buffer[0],
                                         valueInfo.DataSize);
                end;

              rdUnknown:
                begin
                // Nada - don't know how to overwrite!
                end;

              else
                begin
                // Nada - don't know how to overwrite!
                end;

              end;
              end;

          finally
            keyValues.Free();
          end;

          registry.CloseKey();

          for i:=0 to (NTsubkeys.count-1) do
            begin
            DestroyRegKey(rootStr+'\'+NTsubkeys[i]);
            end;
          end;
      finally
        NTsubkeys.Free();
      end;
      end;

    registry.DeleteKey(key);
  finally
    registry.Free();
  end;

end;

// Generate a random filename of the same length as the one supplied, but
// preserving the last "." in the filename
// Give it 5 tries to find a filename that doesn't already exist, if we don't
// find one, just return ''
function TShredder.GenerateRndDotFilename(path: string; origFilename: string): string;
var
  i: integer;
  fndLastDot: boolean;
  finished: boolean;
  count: integer;
begin
  count := 0;
  finished:= FALSE;
  while not(finished) do
    begin
    fndLastDot := FALSE;
    for i:=length(origFilename) downto 1 do
      begin
      if fndLastDot then
        begin
{$WARNINGS OFF}  // Disable useless warning
        origFilename[i] := char(ord('A')+random(26));
{$WARNINGS ON}
        end
      else
        begin
        if origFilename[i]='.' then
          begin
          fndLastDot := TRUE;
          end
        else
          begin
{$WARNINGS OFF}  // Disable useless warning
          origFilename[i] := char(ord('A')+random(26));
{$WARNINGS ON}
          end;
        end;
      end; // for i:=length(origFilename) downto 1 do

    finished := not(FileExists(path+origFilename));
    if not(finished) then
      begin
      inc(count);
      if count=5 then
        begin
        origFilename := '';
        end;
      end;
    end; // while not(finished) do

  Result := origFilename;

end;

function TShredder.CountFiles(dirName: string): integer;
var
  fileIterator: TSDUFileIterator;
  cnt: integer;
begin
  fileIterator:= TSDUFileIterator.Create(nil);
  try
    fileIterator.Directory := dirName;
    fileIterator.RecurseSubDirs := TRUE;
    fileIterator.IncludeDirNames := FALSE;
    fileIterator.Reset();

    cnt := fileIterator.Count();
    
  finally
    fileIterator.Free();
  end;

  Result := cnt;

end;


END.

