unit OTFEFreeOTFE_U;
// Description: Delphi FreeOTFE Component
// Description:
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.SDean12.org/
//
// -----------------------------------------------------------------------------
//


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  OTFE_U,
  dialogs,
  OTFEFreeOTFE_DriverCommon,
  OTFEFreeOTFE_DriverAPI,
  OTFEFreeOTFE_DriverHashAPI,
  OTFEFreeOTFE_DriverCypherAPI,
  OTFEConsts_U,
  OTFEFreeOTFE_VolumeFileAPI,
  OTFEFreeOTFE_LUKSAPI;


const
  MIN_PDA_VOLUME_SIZE = 2;
  BYTES_IN_MEGABYTE = 1024 * 1024;

  // Fairly arbitary number of CDROMs/HDDs/Partitions per HDD to check for
  MAX_CDROMS     = 8;
  MAX_HDDS       = 16;
  MAX_PARTITIONS = 16;

  PARTITION_DEVICE = '\Device\Harddisk%d\Partition%d';
  HDD_DEVICE = '\Device\Harddisk%d';
  CDROM_DEVICE = '\Device\CdRom%d';

  VOL_FILE_EXTN                        = 'vol';

  // Open/Save file filters...
  FILE_FILTER_DFLT_NONE                = '';

  FILE_FILTER_FLT_VOLUMES              = 'Volume files (*.vol)|*.vol|All files|*.*';
  FILE_FILTER_DFLT_VOLUMES             = 'vol';

  FILE_FILTER_FLT_KEYFILES             = 'Keyfiles (*.cdb)|*.cdb|All files|*.*';
  FILE_FILTER_DFLT_KEYFILES            = 'cdb';

  FILE_FILTER_FLT_VOLUMESANDKEYFILES   = 'Volume files and keyfiles (*.vol;*.cdb)|*.vol;*.cdb|Volume files (*.vol)|*.vol|Keyfiles (*.cdb)|*.cdb|All files|*.*';
  FILE_FILTER_DFLT_VOLUMESANDKEYFILES  = FILE_FILTER_DFLT_NONE;

  FILE_FILTER_FLT_TEXTFILES            = 'Text files (*.txt)|*.txt|All files|*.*';
  FILE_FILTER_DFLT_TEXTFILES           = 'txt';

  FILE_FILTER_FLT_CDBBACKUPS           = 'CDB backups (*.cdbBackup)|*.cdbBackup|All files|*.*';
  FILE_FILTER_DFLT_CDBBACKUPS          = 'cdbBackup';

  FILE_FILTER_FLT_EXECUTABLES          = 'Executable files (*.exe)|*.exe|All files|*.*';
  FILE_FILTER_DFLT_EXECUTABLES         = 'exe';

  FILE_FILTER_DFLT_LINUX_SETTINGS      = 'les';
  FILE_FILTER_FLT_LINUX_SETTINGS       = 'Linux encryption settings (*.'+FILE_FILTER_DFLT_LINUX_SETTINGS+')|*.'+FILE_FILTER_DFLT_LINUX_SETTINGS+'|All files|*.*';



  // xxx - This shouldn't really be used; we should be getting the actual
  // host disk's sector size and using that, instead of just *assuming* 512
  // bytes, as we do here
  // xxx - change this to get the host disk's geometry
  ASSUMED_HOST_SECTOR_SIZE = 512;

const
  // Only used by the GUI
  DEFAULT_SALT_LENGTH = 256;  // In bits

  DEFAULT_KEY_ITERATIONS = 2048;
  DEFAULT_KEY_ITERATIONS_INCREMENT = 512;

  // Default key length - only used if user's selected encryption algorithm
  // doesn't have a set keylength
  DEFAULT_CYPHER_KEY_LENGTH = 512;  // In bits

  // Default hash length - only used if user's selected hash algorithm
  // doesn't have a set hash length
  DEFAULT_HASH_LENGTH_LENGTH = 512;  // In bits




type
  // Exceptions...
  EFreeOTFEError = class(Exception);
  EFreeOTFEInvalidParameter = class(EFreeOTFEError);
  EFreeOTFENeedAdminPrivs = class(EFreeOTFEError);
  EFreeOTFEConnectFailure = class(EFreeOTFEError);
  EFreeOTFEObsoleteDriver = class(EFreeOTFEError);
  EFreeOTFEInvalidDerivedKeyLength = class(EFreeOTFEError);
  EFreeOTFEInternalError = class(EFreeOTFEError);  // Critical failure within this software
  EFreeOTFEDriverError = class(EFreeOTFEError);  // Critical failure within the driver


type
  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  TFreeOTFECypherMode = (focmNone, focmECB, focmCBC, focmUnknown);

  TFreeOTFEMountAs = (fomaFixedDisk, fomaRemovableDisk, fomaCD, fomaDVD, fomaUnknown);

const
  FreeOTFEMountAsTitle : array [TFreeOTFEMountAs] of string = (
                                                               'Fixed disk',
                                                               'Removable disk',
                                                               'CD',
                                                               'DVD',
                                                               'Unknown'
                                                              );

  // Indicate which of the above emulated devices are writable
  FreeOTFEMountAsCanWrite : array [TFreeOTFEMountAs] of boolean = (
                                                               TRUE,
                                                               TRUE,
                                                               FALSE,
                                                               FALSE,
                                                               FALSE
                                                              );

  // Decode the above into device types/media types
  FreeOTFEMountAsDeviceType : array [TFreeOTFEMountAs] of DWORD = (
                                                              FILE_DEVICE_DISK,
                                                              FILE_DEVICE_DISK,
                                                              FILE_DEVICE_CD_ROM,
                                                              FILE_DEVICE_DVD,
                                                              FILE_DEVICE_UNKNOWN
                                                              );

  // Decode the above into device types/media types
  FreeOTFEMountAsStorageMediaType : array [TFreeOTFEMountAs] of TFreeOTFEStorageMediaType = (
                                                              mtFixedMedia,
                                                              mtRemovableMedia,
                                                              mtCD_ROM,
                                                              mtDVD_ROM,
                                                              mtUnknown
                                                              );



  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  FreeOTFECypherModeTitle : array [TFreeOTFECypherMode] of string = (
                                                                     'None',
                                                                     'ECB',
                                                                     'CBC',
                                                                     'Unknown'
                                                                    );

  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  // These are the numerical IDs that are to be passed to/from the FreeOTFE
  // encryption drivers
  FreeOTFECypherModeID : array [TFreeOTFECypherMode] of integer = (
                                                       0,  // CYPHER_MODE_NONE
                                                       1,  // CYPHER_MODE_ECB
                                                       2,  // CYPHER_MODE_CBC
                                                    9999   // CYPHER_MODE_UNKNOWN
                                                                  );

  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  FreeOTFEMACTitle : array [TFreeOTFEMACAlgorithm] of string = (
                                                                'Hash',
                                                                'HMAC',
                                                                'Unknown'
                                                               );

  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  // These are the numerical IDs that are to be passed to/from the FreeOTFE
  // encryption drivers
  FreeOTFEMACID : array [TFreeOTFEMACAlgorithm] of integer = (
                                                       1,  // MAC_HASH
                                                       2,  // MAC_HMAC
                                                    9999   // MAC_UNKNOWN
                                                             );

  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  FreeOTFEKDFTitle : array [TFreeOTFEKDFAlgorithm] of string = (
                                                                'Hash with salt',
                                                                'PKCS #5 PBKDF2 (HMAC)',
                                                                'Unknown'
                                                               );

  // This definition is stored here, and not in OTFEFreeOTFE_DriverCypherAPI, so
  // that only this unit need be "used" by user applications
  // These are the numerical IDs that are to be passed to/from the FreeOTFE
  // encryption drivers
  FreeOTFEKDFID : array [TFreeOTFEKDFAlgorithm] of integer = (
                                                       1,  // KDF_HASH
                                                       2,  // KDF_PBKDF2
                                                    9999   // KDF_UNKNOWN
                                                             );


type

  // ------------------------
  // DELPHI FACING STRUCTURES
  // ------------------------

  TOTFEFreeOTFEVolumeInfo = packed record
    DriveLetter: char;
    Filename: string;
    DeviceName: string;

    IVHashDevice: string;
    IVHashGUID: TGUID;
    IVCypherDevice: string;
    IVCypherGUID: TGUID;

    MainCypherDevice: string;
    MainCypherGUID: TGUID;

    Mounted: boolean;
    ReadOnly: boolean;

    VolumeFlags: DWORD;
    SectorIVGenMethod: TFreeOTFESectorIVGenMethod;

    MetaData: string;
  end;



type
  TFreeOTFECypher = packed record
    CypherGUID: TGUID;

    Title: string;
    Mode: TFreeOTFECypherMode;
    KeySize: integer;  // In bits
    BlockSize: integer;  // In bits
    VersionID: DWORD;
  end;


  PFreeOTFECypherDriver = ^TFreeOTFECypherDriver;
  TFreeOTFECypherDriver = packed record
    DriverGUID: TGUID;

    // The device name
    DeviceName: string;
    // The MSDOS device name (to be used when connecting directly from userspace)
    DeviceUserModeName: string;
    // The MSDOS device name (to be passed to the FreeOTFE driver when mounting)
    DeviceKernelModeName: string;
    // The name of the driver, as should be displayed to the user
    Title: string;
    VersionID: DWORD;
    CypherCount: integer;
    Cyphers: array of TFreeOTFECypher;
  end;


  TFreeOTFEHash = packed record
    HashGUID: TGUID;

    Title: string;
    VersionID: DWORD;
    Length: integer;  // In bits
    BlockSize: integer;  // In bits
  end;


  PFreeOTFEHashDriver = ^TFreeOTFEHashDriver;
  TFreeOTFEHashDriver = packed record
    DriverGUID: TGUID;

    // The device name
    DeviceName: string;
    // The MSDOS device name (to be used when connecting directly from userspace)
    DeviceUserModeName: string;
    // The MSDOS device name (to be passed to the FreeOTFE driver when mounting)
    DeviceKernelModeName: string;
    // The name of the driver, as should be displayed to the user
    Title: string;
    VersionID: DWORD;
    HashCount: integer;
    Hashes: array of TFreeOTFEHash;
  end;


  // This definition is needed because we cannot pass an "array of Txyz" to a
  // function/procedure a "var" parameter, and the use SetLength(...) on that
  // parameter within the function/parameter
  TFreeOTFECypherDriverArray = array of TFreeOTFECypherDriver;
  TFreeOTFEHashDriverArray = array of TFreeOTFEHashDriver;


  TOTFEFreeOTFE = class(TOTFE)
  private
    // This is a bit hacky, but the Dump functionality does things that you
    // wouldn't normally do (i.e. it's a high level function that wants to
    // know about the inner workings of lower-level functionality; CDB
    // processing)
    DumpFlag: boolean;
    DumpCriticalDataKey: string;
    DumpCheckMAC: string;
    DumpPlaintextEncryptedBlock: string;
    DumpVolumeDetailsBlock: string;
  protected
    DriverHandle: THandle;
    fCachedVersionID: DWORD;

    // Cached information: The strings in the list are kernel mode device
    // names, the objects are TFreeOTFEHashDriver/TFreeOTFECypherDriver
    fCachedHashDriverDetails: TStringList;
    fCachedCypherDriverDetails: TStringList;

    // Set the component active/inactive
    procedure SetActive(status: Boolean); override;

    // Connect/disconnect to the main FreeOTFE device driver
    function  Connect(): boolean;
    function  Disconnect(): boolean;


    // ---------
    // FreeOTFE *disk* *device* management functions
    // These talk directly to the disk devices to carrry out operations
    // deviceType - Set to FILE_DEVICE_DISK, etc
    function  CreateDiskDevice(deviceType: DWORD = FILE_DEVICE_DISK): string;

    // volFilename - This is the filename of the file/partition as seen by the
    //               user mode software
    function  MountDiskDevice(
                              deviceName: string;
                              volFilename: string;
                              volumeKey: string;
                              sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
                              volumeIV: string;
                              readonly: boolean;
                              IVHashDriver: string;
                              IVHashGUID: TGUID;
                              IVCypherDriver: string;
                              IVCypherGUID: TGUID;
                              mainCypherDriver: string;
                              mainCypherGUID: TGUID;
                              VolumeFlags: integer;
                              metaData: string;
                              offset: int64 = 0;
                              size: int64 = 0;
                              storageMediaType: TFreeOTFEStorageMediaType = mtFixedMedia
                             ): boolean;

    function  DismountDiskDevice(deviceName: string; emergency: boolean): boolean;
    function  DestroyDiskDevice(deviceName: string): boolean;

    function  LDREUDriver(
                          driveLetter: char;
                          deviceName: string;
                          emergency: boolean
                         ): boolean;
    function  LDREUUserApp(
                          driveLetter: char;
                          deviceName: string;
                          emergency: boolean;
                          var unableToOpenDrive: boolean
                         ): boolean;

    // ---------
    // Status functions

    // Get the number of disk devices the driver currently has (including both
    // mounted and unmounted disks)
    function  GetDiskDeviceCount(): cardinal;

    // Get the kernel mode device name of all disk devices
    function  GetDiskDeviceList(deviceNames: TStringList): boolean;


    // Get a list of all FreeOTFE devices which are currently mounted
    // Note: Unmounted devices will have '' in the "driveLetters" item
    //       corresponding to the unmounted device listed in the same item
    //       position in "deviceNames"
    function GetMountedDevices(driveLetters: TStringList; deviceNames: TStringList): boolean;

    // Given the specified drive letter, return the FreeOTFE device for that
    // mounted drive
    function GetDriveDeviceName(driveLetter: char): string;
    // Given the specified device name, return the drive mounted on that device
    function GetDriveDeviceLetter(deviceName: string): char;


    // ---------
    // Raw volume access functions
    // Note: These all transfer RAW data to/from the volume - no
    //       encryption/decryption takes place

    // Get the FreeOTFE driver to read a critical data block from the named file
    // starting from the specified offset
    function  ReadRawVolumeCriticalData(
                             filename: string;
                             offsetWithinFile: int64;
                             var criticalData: string
                            ): boolean; overload;

    // Get the FreeOTFE driver to write a critical data block from the named file
    // starting from the specified offset
    function  WriteRawVolumeCriticalData(
                             filename: string;
                             offsetWithinFile: int64;
                             criticalData: string
                            ): boolean; overload;

    function ReadRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;

    // This executes a sipmle call to the driver to read dataLength bytes from
    // offsetWithinFile
    // Note: Will probably not work when reading directly from partitions;
    //       they typically need read/writes carried out in sector sized blocks
    //        - see ReadRawVolumeDataBounded for this.
    function ReadRawVolumeDataSimple(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;

    // This function is the same as ReadRawVolumeDataSimple(...), but will round
    // the offsetWithinFile down to the nearest sector offset, read in complete
    // sectors, and *only* *return "datalength" *bytes*
    // This function is included as reading from partitions must be carried out
    // in sector size blocks
    function ReadRawVolumeDataBounded(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;

    function WriteRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         data: string
                        ): boolean;

    function ReadWritePlaintextToVolume(
                    readNotWrite: boolean;

                    volFilename: string;
                    volumeKey: string;
                    sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
                    volumeIV: string;
                    IVHashDriver: string;
                    IVHashGUID: TGUID;
                    IVCypherDriver: string;
                    IVCypherGUID: TGUID;
                    mainCypherDriver: string;
                    mainCypherGUID: TGUID;
                    VolumeFlags: integer;
                    mountMountAs: TFreeOTFEMountAs;

                    dataOffset: int64;  // Offset from within mounted volume from where to read/write data
                    dataLength: integer;  // Length of data to read/write. In bytes
                    var data: string;  // Data to read/write

                    offset: int64 = 0;
                    size: int64 = 0;
                    storageMediaType: TFreeOTFEStorageMediaType = mtFixedMedia
                  ): boolean;

    // ---------
    // Misc functions

    // Generate a Linux volume's volume key
    function GenerateLinuxVolumeKey(
                                    hashDriver: string;
                                    hashGUID: TGUID;
                                    userKey: string;
                                    seed: string;
                                    hashWithAs: boolean;
                                    targetKeylengthBits: integer;
                                    var volumeKey: string
                                   ): boolean;

    // Call QueryDosDevice, and return it's output
    function DoQueryDosDevice(deviceName: string; queryResults: TStringList): boolean;
    // Given an MSDOS device name, return the device name of that device
    // (Uses DoQueryDosDevice)
    function GetDeviceName(MSDOSKernelModeDeviceName: string): string;

    // Create/delete DosDevices symlink; effectivly wraps DefineDosDevice(...) with
    // additional support required for Vista changes
    function DosDeviceSymlink(
                              CreateNotDelete: boolean;
                              DeviceName: string;
                              DriveLetter: char;
                              Global: boolean
                             ): boolean;

    // Delete both local and global DosDevices symlinks, as possible
    function DeleteDosDeviceSymlink(DeviceName: string; DriveLetter: char): boolean;

    // Connect to the specified user mode named device
    function ConnectDevice(devicename: string): THandle;
    // Disconnect to the specified device
    function DisconnectDevice(deviceHandle: THandle): boolean;

    // Convert kernel mode device name to user mode device name
    function  GetHashDeviceUserModeDeviceName(hashKernelModeDeviceName: string): string;
    // Convert kernel mode device name to user mode device name
    function  GetCypherDeviceUserModeDeviceName(cypherKernelModeDeviceName: string): string;


    function  GetNextDriveLetter(userDriveLetter, requiredDriveLetter: char): char;


    // ---------
    // Convert critical data area to/from a string representation

    // Note: This function does *not* append random padding data
    function  BuildVolumeDetailsBlock(volumeDetailsBlock: TVolumeDetailsBlock; var stringRep: string): boolean;
    // Note: Ignores any random padding data
    function  ParseVolumeDetailsBlock(stringRep: string; var volumeDetailsBlock: TVolumeDetailsBlock): boolean;


    // Return a nicely formatted string, decoding one bit of a bitmapped value
    function DumpCriticalDataToFileBitmap(
                                   value: DWORD;
                                   bit: integer;
                                   unsetMeaning: string;
                                   setMeaning: string
                                  ): string;

    // Convert a user-space volume filename to a format the kernel mode driver
    // can understand
    function GetKernelModeVolumeFilename(userModeFilename: string): string;
    // Convert a kernel mode driver volume filename to format user-space
    // filename
    function GetUserModeVolumeFilename(kernelModeFilename: string): string;

    // Return TRUE/FALSE, depending on whether the specified KERNEL MODE volume
    // filename refers to a partition/file
    function IsPartition(kernelModeFilename: string): boolean;


    // expectedLength - Set to the amount of metadata to retrieve
    function  GetVolumeMetaData(driveLetter: char; expectedLength: integer; var metaData: string): boolean;

    // ---------
    // Internal caching functions
    procedure CachesCreate();
    procedure CachesDestroy();
    // Add a hash to the cache...
    procedure CachesAddHashDriver(hashKernelModeDeviceName: string; hashDriverDetails: TFreeOTFEHashDriver);
    // Get a hash from the cache...
    // Note: A cache MISS will cause this to return FALSE
    function  CachesGetHashDriver(hashKernelModeDeviceName: string; var hashDriverDetails: TFreeOTFEHashDriver): boolean;
    // Add a cypher to the cache...
    procedure CachesAddCypherDriver(cypherKernelModeDeviceName: string; cypherDriverDetails: TFreeOTFECypherDriver);
    // Get a cypher from the cache...
    // Note: A cache MISS will cause this to return FALSE
    function  CachesGetCypherDriver(cypherKernelModeDeviceName: string; var cypherDriverDetails: TFreeOTFECypherDriver): boolean;

  public
{$IFDEF FREEOTFE_DEBUG}
DebugStrings: TStrings;
DebugShowMessage: boolean;
procedure DebugClear();
procedure DebugMsg(msg: string);
procedure DebugMsgBinary(msg: string);
procedure DebugMsgBinaryPtr(msg: PChar; len:integer);
{$ENDIF}

{$INCLUDE OTFEFreeOTFE_mntLUKS_Intf.pas}

    // -----------------------------------------------------------------------
    // TOTFE standard API
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy(); override;
    function  Mount(volumeFilename: string; readonly: boolean = FALSE): char; overload; override;
    function  Mount(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean; overload; override;
    function  CanMountDevice(): boolean; override;
    function  MountDevices(): string; override;
    function  Dismount(volumeFilename: string; emergency: boolean = FALSE): boolean; overload; override;
    function  Dismount(driveLetter: char; emergency: boolean = FALSE): boolean; overload; override;
    function  Title(): string; overload; override;
    function  Version(): cardinal; overload; override;
    function  VersionStr(): string; overload; override;
    function  IsEncryptedVolFile(volumeFilename: string): boolean; override;
    function  DrivesMounted(): string; overload; override;
    function  GetVolFileForDrive(driveLetter: char): string; override;
    function  GetDriveForVolFile(volumeFilename: string): char; override;
    function  GetMainExe(): string; override;


    // -----------------------------------------------------------------------
    // Extended FreeOTFE specific functions

    // Mount file(s) assuming specific volume type
    // Note: The volumeFilename(s) passed in can EITHER by files on the local
    //       filesystem (e.g. "C:\myfile.dat"), OR partitions
    //       (e.g. "\Device\Harddisk1\Partition3")
    function  MountLinux(volumeFilename: string; readonly: boolean = FALSE): char; overload;
    function  MountLinux(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean; overload;
    // Note: The MountLUKS(...) functions will be called automatically if
    //       MountLinux(...) is called with a LUKS volume
    function  MountLUKS(volumeFilename: string; readonly: boolean = FALSE): char; overload;
    function  MountLUKS(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean; overload;
    function  MountFreeOTFE(volumeFilename: string; readonly: boolean = FALSE): char; overload;
    function  MountFreeOTFE(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean; overload;


    // See function comment in body of function
{$IFDEF LINUX_DETECT}
    function  DetectLinux(
      volumeFilename: string;
      userKey: string;
      keyProcSeed: string;
      keyProcIterations: integer;
      fileOptOffset: int64;
      fileOptSize: int64;
      mountDriveLetter: char;
      mountMountAs: TFreeOTFEMountAs;
      testFilename: string
      ): string;
{$ENDIF}

    // Get information on a mounted volume
    function  GetVolumeInfo(driveLetter: char; var volumeInfo: TOTFEFreeOTFEVolumeInfo): boolean;

    // Get details of all hash drivers and the hashes they support
    function GetHashDrivers(var hashDrivers: TFreeOTFEHashDriverArray): boolean;
    // Get details of a single hash driver, and all it's hashes
    function GetHashDriverHashes(hashKernelModeDeviceName: string; var hashDriverDetails: TFreeOTFEHashDriver): boolean;

    // Get details of all cypher drivers and the cyphers they support
    function GetCypherDrivers(var cypherDrivers: TFreeOTFECypherDriverArray): boolean;
    // Get details of a single cypher driver, and all it's cyphers
    function GetCypherDriverCyphers(cypherKernelModeDeviceName: string; var cypherDriverDetails: TFreeOTFECypherDriver): boolean;

    // These two additional "helper" functions will *clear* and repopulate the TStringLists supplied with:
    //   A displayable title for each hash/cypher
    //   The kernel mode driver name for each hash/cypher
    //   The GUID for each hash/cypher
    function GetHashList(var hashDispTitles, hashKernelModeDriverNames, hashGUIDs: TStringList): boolean;
    function GetCypherList(var cypherDispTitles, cypherKernelModeDriverNames, cypherGUIDs: TStringList): boolean;

    // Two additional "helper" functions...
    // Generate a prettyprinted title for a specific cypher
    // Returns '' on error
    function GetCypherDisplayTitle(cypher: TFreeOTFECypher): string;
    // Generate a prettyprinted title for a specific hash
    // Returns '' on error
    function GetHashDisplayTitle(hash: TFreeOTFEHash): string;


    // Display a standard dialog with details of the specific hash identified
    procedure ShowHashDetailsDlg(driverName: string; hashGUID: TGUID);
    // Display a standard dialog with details of the specific cypher identified
    procedure ShowCypherDetailsDlg(driverName: string; cypherGUID: TGUID);


    function  GetSpecificHashDetails(hashKernelModeDeviceName: string; hashGUID: TGUID; var hashDetails: TFreeOTFEHash): boolean;
    function  GetSpecificCypherDetails(cypherKernelModeDeviceName: string; cypherGUID: TGUID; var cypherDetails: TFreeOTFECypher): boolean;


    // Display dialog to allow user to select a partition
    // Returns '' if the user cancels/on error, otherwise returns a partition
    // identifier
    function  SelectPartition(): string;
    function  HDDDeviceList(hddDevices: TStringList; title: TStringList): boolean;
    function  CDROMDeviceList(cdromDevices: TStringList; title: TStringList): boolean;


    // Check to see if there's at least one cypher and one hash algorithm
    // available for use.
    // Warn user and return FALSE if not.
    function  WarnIfNoHashOrCypherDrivers(): boolean;

    // Convert a version ID into a prettyprinted version string
    function  VersionIDToStr(versionID: DWORD): string;

    
    // ---------
    // Algorithm driver functions

    // Encrypt data using the cypher/cypher driver identified
    // Encrypted data returned in "cyphertext"
    function EncryptData(
                         cypherKernelModeDeviceName: string;
                         cypherGUID: TGUID;
                         var key: string;
                         var IV: string;
                         var plaintext: string;
                         var cyphertext: string
                        ): boolean;
    // Decrypt data using the cypher/cypher driver identified
    // Decrypted data returned in "plaintext"
    function DecryptData(
                         cypherKernelModeDeviceName: string;
                         cypherGUID: TGUID;
                         var key: string;
                         var IV: string;
                         var cyphertext: string;
                         var plaintext: string
                        ): boolean;

    // Combined...
    function EncryptDecryptData(
                                encryptFlag: boolean;
                                cypherKernelModeDeviceName: string;
                                cypherGUID: TGUID;
                                var key: string;
                                var IV: string;
                                var inData: string;
                                var outData: string
                               ): boolean;

    // Hash data using the hash driver identified
    // Result returned in "hashOut"
    function HashData(
                      hashKernelModeDeviceName: string;
                      hashGUID: TGUID;
                      var data: string;
                      var hashOut: string
                     ): boolean;

    // Generate MAC of data using hash driver/encryption driver identified
    // Note: "tBits" is in *bits* not *bytes*
    // tBits - Set the the number of bits to return; set to -ve value for no
    //         truncation/padding
    //         If tBits is larger than the number of bits the MAC would normally
    //         be, then the MAC will be right-padded with NULLs
    //         If tBits is set to a -ve number, then this function will return
    //         an MAC of variable length
    function MACData(
                      macAlgorithm: TFreeOTFEMACAlgorithm;
                      hashKernelModeDeviceName: string;
                      hashGUID: TGUID;
                      cypherKernelModeDeviceName: string;
                      cypherGUID: TGUID;
                      var key: string;
                      var data: string;
                      var MACOut: string;
                      tBits: integer = -1
                     ): boolean;


    // Key derivation function
    function DeriveKey(
                    kdfAlgorithm: TFreeOTFEKDFAlgorithm;
                    hashKernelModeDeviceName: string;
                    hashGUID: TGUID;
                    cypherKernelModeDeviceName: string;
                    cypherGUID: TGUID;
                    Password: string;
                    Salt: string;
                    Iterations: integer;
                    dkLenBits: integer;  // In *bits*
                    var DK: string
                   ): boolean;


    // ---------
    // Volume creation
    function CreateFreeOTFEVolumeWizard(): boolean;
    function CreateLinuxVolumeWizard(): boolean;
    function CreateLinuxGPGKeyfile(): boolean;

    function WizardChangePassword(): boolean;
    function WizardCreateKeyfile(): boolean;


    // ---------
    // Critical data block handling
    function  ReadVolumeCriticalData(
                                     filename: string;
                                     offsetWithinFile: int64;
                                     userPassword: string;
                                     saltLength: integer;  // In bits
                                     keyIterations: integer;

                                     var volumeDetails: TVolumeDetailsBlock;
                                     var CDBMetaData: TCDBMetaData
                                    ): boolean;

    function  ReadVolumeCriticalData_v1(
                                     filename: string;
                                     offsetWithinFile: int64;
                                     userPassword: string;
                                     saltLength: integer;  // In bits

                                     var volumeDetails: TVolumeDetailsBlock;
                                     var CDBMetaData: TCDBMetaData
                                    ): boolean; overload;

    function  ReadVolumeCriticalData_v2(
                                     filename: string;
                                     offsetWithinFile: int64;
                                     userPassword: string;
                                     saltLength: integer;  // In bits
                                     keyIterations: integer;

                                     var volumeDetails: TVolumeDetailsBlock;
                                     var CDBMetaData: TCDBMetaData
                                    ): boolean; overload;

    function ChooseFromValid(
                             validVolumeDetails: TVolumeDetailsBlockArray;
                             validCDBMetaDatas: TCDBMetaDataArray;
                             var volumeDetails: TVolumeDetailsBlock;
                             var CDBMetaData: TCDBMetaData
                            ): boolean;

    // Write out a critical data block
    // If the file specified doesn't already exist, it will be created
    // ALL members of these two "volumeDetails" and "CDBMetaData" structs MUST
    // be populated
    // *Except* for volumeDetails.CDBFormatID - this is force set to the latest value
    function  WriteVolumeCriticalData(
                                      filename: string;
                                      offsetWithinFile: int64;
                                      userPassword: string;
                                      salt: string;
                                      keyIterations: integer;

                                      volumeDetails: TVolumeDetailsBlock;
                                      CDBMetaData: TCDBMetaData;

                                      randomPadData: string // Note: If insufficient is supplied, the write will
                                                            //       *fail*
                                                            //       The maximum that could possibly be required
                                                            //       is CRITICAL_DATA_LENGTH bits.
                                     ): boolean;  overload;

    function  BackupVolumeCriticalData(
                                       srcFilename: string;
                                       srcOffsetWithinFile: int64;
                                       destFilename: string
                                      ): boolean;

    function  RestoreVolumeCriticalData(
                                        srcFilename: string;
                                        destFilename: string;
                                        destOffsetWithinFile: int64
                                       ): boolean;

    function  DumpCriticalDataToFile(
                                   volFilename: string;
                                   offsetWithinFile: int64;
                                   userPassword: string;
                                   saltLength: integer;  // In bits
                                   keyIterations: integer;
                                   dumpFilename: string
                                  ): boolean;

    function  ChangeVolumePassword(
                                   filename: string;
                                   offsetWithinFile: int64;
                                   oldUserPassword: string;
                                   oldSaltLength: integer;  // In bits
                                   oldKeyIterations: integer;
                                   newUserPassword: string;
                                   newSaltBytes: string;
                                   newKeyIterations: integer;
                                   newRequestedDriveLetter: char;
                                   newRandomPadData: string  // Note: If insufficient is supplied, the write will *fail*
                                                             //       The maximum that could possibly be required is CRITICAL_DATA_LENGTH bits.
                                  ): boolean;

    function  CreateKeyfile(
                               srcFilename: string;
                               srcOffsetWithinFile: int64;
                               srcUserPassword: string;
                               srcSaltLength: integer;  // In bits
                               srcKeyIterations: integer;
                               keyFilename: string;
                               keyfileUserPassword: string;
                               keyfileSaltBytes: string;
                               keyfileKeyIterations: integer;
                               keyfileRequestedDrive: char;
                               keyfileRandomPadData: string  // Note: If insufficient is supplied, the write will *fail*
                                                             //       The maximum that could possibly be required is CRITICAL_DATA_LENGTH bits.
                              ): boolean;

    // ---------
    // FreeOTFE Driver handling...

    // Display control for controlling underlying drivers
    procedure ShowDriverControlDlg();

    // Install and start FreeOTFE drivers in portable mode
    function  PortableStart(driverFilenames: TStringList; showProgress: boolean): boolean;

    // Stop and uninstall any FreeOTFE drivers started in portable mode
    function  PortableStop(): boolean;

    // Identify how many drivers are currently running in portable mode
    function  DriversInPortableMode(): integer;

    function  CanUserManageDrivers(): boolean;

    procedure CachesFlush();

    // ---------
    // Debug functions - expose private members in debug builds for testing...

{$IFDEF FREEOTFE_DEBUG}
    // Expose ReadRawVolumeCriticalData(...) for debug
    function  DEBUGReadRawVolumeCriticalData(
                             filename: string;
                             offsetWithinFile: int64;
                             var criticalData: string
                            ): boolean;

    // Expose WriteRawVolumeCriticalData(...) for debug
    function  DEBUGWriteRawVolumeCriticalData(
                             filename: string;
                             offsetWithinFile: int64;
                             criticalData: string
                            ): boolean;

    // Expose ReadRawVolumeData(...) for debug
    function DEBUGReadRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;

    // Expose WriteRawVolumeData(...) for debug
    function DEBUGWriteRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         data: string
                        ): boolean;

    // Expose ReadWritePlaintextToVolume(...) for debug
    function DEBUGReadWritePlaintextToVolume(
                    readNotWrite: boolean;

                    volFilename: string;
                    volumeKey: string;
                    sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
                    volumeIV: string;
                    IVHashDriver: string;
                    IVHashGUID: TGUID;
                    IVCypherDriver: string;
                    IVCypherGUID: TGUID;
                    mainCypherDriver: string;
                    mainCypherGUID: TGUID;
                    VolumeFlags: integer;
                    mountMountAs: TFreeOTFEMountAs;

                    dataOffset: int64;  // Offset from within mounted volume from where to read/write data
                    dataLength: integer;  // Length of data to read/write. In bytes
                    var data: string;  // Data to read/write

                    offset: int64 = 0;
                    size: int64 = 0;
                    storageMediaType: TFreeOTFEStorageMediaType = mtFixedMedia
                  ): boolean;
{$ENDIF}

  end;


procedure Register;


implementation

uses
{$IFDEF LINUX_DETECT}
  DbugIntf,  // GExperts
{$ENDIF}
{$IFDEF FREEOTFE_DEBUG}
  DbugIntf,  // GExperts
{$ENDIF}
  SDUGeneral,
  SDUProgress_U,
  SDUDialogs,
  SDUEndianIntegers,
  OTFEFreeOTFE_frmVolumeType,
  OTFEFreeOTFE_frmKeyEntryFreeOTFE,
  OTFEFreeOTFE_frmKeyEntryLinux,
  OTFEFreeOTFE_frmKeyEntryLUKS,
  OTFEFreeOTFE_frmHashInfo,
  OTFEFreeOTFE_frmCypherInfo,
  OTFEFreeOTFE_frmWizardCreateVolume,
  OTFEFreeOTFE_frmSelectHashCypher,
  OTFEFreeOTFE_frmNewVolumeSize,
  OTFEFreeOTFE_DriverControl,
  OTFEFreeOTFE_frmDriverControl,
  OTFEFreeOTFE_frmWizardChangePasswordCreateKeyfile,
  OTFEFreeOTFE_frmSelectPartition,
  Math,  // Required for min
  ActiveX,  // Required for IsEqualGUID
  ComObj,  // Required for GUIDToString
  WinSvc;  // Required for SERVICE_ACTIVE

type
  EFreeOTFEReadCDBFailure = EFreeOTFEError;  // Internal error - should never
                                             // be propagated beyond this unit
  EFreeOTFEWriteCDBFailure = EFreeOTFEError;  // Internal error - should never
                                              // be propagated beyond this unit
  PHashIdentify = ^THashIdentify;
  THashIdentify = record
    KernelModeDriverName: string;
    GUID: string;
    Details: TFreeOTFEHash;
  end;

  PCypherIdentify = ^TCypherIdentify;
  TCypherIdentify = record
    KernelModeDriverName: string;
    GUID: string;
    Details: TFreeOTFECypher;
  end;

const
  CRLF = #10+#13;

  VERSION_ID_NOT_CACHED = VERSION_ID_FAILURE;

  // These are artifical maximums; they must be enough to store the results of
  // the DIOC calls when the amount of data output is variable 
  MAX_DERIVED_KEY_LENGTH = 1024*1024;  // In *bits*
  MAX_MAC_LENGTH         = 1024*1024;  // In *bits*

  // This const is used in rounding up DIOC buffers to the nearest DWORD
  DIOC_BOUNDRY = 4;

{$INCLUDE OTFEFreeOTFE_mntLUKS_Impl.pas}

// ----------------------------------------------------------------------------
procedure Register;
begin
  RegisterComponents('OTFE', [TOTFEFreeOTFE]);
end;


// ----------------------------------------------------------------------------
constructor TOTFEFreeOTFE.Create(AOwner : TComponent);
begin
  inherited;

{$IFDEF FREEOTFE_DEBUG}
  DebugStrings := TStringList.Create();
  DebugShowMessage := TRUE;
{$ENDIF}

  DumpFlag := FALSE;

  CachesCreate();

end;


// ----------------------------------------------------------------------------
destructor TOTFEFreeOTFE.Destroy();
begin
  CachesDestroy();

{$IFDEF FREEOTFE_DEBUG}
  DebugStrings.Free();
{$ENDIF}

  inherited;
end;

// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.SetActive(status: boolean);
var
  allOK: boolean;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  allOK := TRUE;

  // Flush caches
  CachesFlush();

{$IFDEF FREEOTFE_DEBUG}
if (status) then
  begin
  DebugMsg(
           '-------------------------------------------------------'+CRLF+
           '-------------------------------------------------------'+CRLF+
           'DEBUG BUILD OF TOTFEFreeOTFE component being set ACTIVE'+CRLF+
           '-------------------------------------------------------'+CRLF+
           '-------------------------------------------------------'
          );
  end;
{$ENDIF}

  if (status <> Active) then
    begin
    if status then
      begin
      allOK := Connect();
      if not(allOK) then
        begin
        raise EFreeOTFEConnectFailure.Create('FreeOTFE driver not installed/not running');
        end;

      end
    else
      begin
      allOK := Disconnect();
      if not(allOK) then
        begin
        raise EFreeOTFEConnectFailure.Create('Could not disconnect from FreeOTFE driver?!');
        end;

      end;

    end;


  // Note: LastErrorCode will have been set by Connect(...)/Dicconnect(...) on
  //       failure.
  if allOK then
    begin
    inherited;
    end;


  if (Active) then
    begin
    // Previous versions which are no longer supported due to changes in the driver API
    if (Version() < FREEOTFE_ID_v00_58_0000) then
      begin
      Active := FALSE;
      raise EFreeOTFEObsoleteDriver.Create(
                     'An old version of the FreeOTFE driver was detected. Please upgrade before using this software'
                     );
      end;

    end;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Connect(): boolean;
var
  allOK: boolean;
begin
  allOK := FALSE;

  DriverHandle := ConnectDevice(DEVICE_SYMLINK_MAIN_NAME);

  if (DriverHandle <> 0) then
    begin
    allOK := TRUE;
    end
  else
    begin
    LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Disconnect(): boolean;
begin
  DisconnectDevice(DriverHandle);
  Result:= TRUE;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.MountFreeOTFE(volumeFilename: string; readonly: boolean = FALSE): char;
var
  tmpStringList: TStringList;
  mountedAs: string;
  retVal: char;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := #0;

  tmpStringList:= TStringList.Create();
  try
    tmpStringList.Add(volumeFilename);
    if MountFreeOTFE(tmpStringList, mountedAs, readonly) then
      begin
      retVal := mountedAs[1];
      end;

  finally
    tmpStringList.Free();
  end;

  Result :=  retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.MountFreeOTFE(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean;
var
  retVal: boolean;
  deviceName: string;
  keyEntryDlg: TfrmKeyEntryFreeOTFE;
  i: integer;
  currMountFilename: string;
  userKey: string;
  fileOptOffset: int64;
  offsetPointsToCDB: boolean;
  fileOptSaltLength: integer;
  volFileOffset: int64;
  mountDriveLetter: char;
  mountReadonly: boolean;
  mountMountAs: TFreeOTFEMountAs;

  volumeDetails: TVolumeDetailsBlock;
  CDBMetaData: TCDBMetaData;

  mr: integer;
  userDriveLetter: char;
  keyfile: string;
  keyIterations: integer;
  criticalDataFile: string;
  criticalDataOffset: int64;
  mountForAllUsers: boolean;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := TRUE;

  CheckActive();

  keyEntryDlg := TfrmKeyEntryFreeOTFE.Create(nil);
  try
    keyEntryDlg.SetReadonly(readonly);

    // Get user password, drive letter to use, etc
    mr := keyEntryDlg.ShowModal();
    if (mr = mrCancel) then
      begin
      LastErrorCode := OTFE_ERR_USER_CANCEL;
      retVal := FALSE;
      end
    else if (mr = mrOK) then
      begin
      // Retrieve all data from the mount dialog...
      if (
          // Key and keyfile...
          (keyEntryDlg.GetKey(userKey)) AND
          (keyEntryDlg.GetKeyfile(keyfile)) AND
          (keyEntryDlg.GetKeyIterations(keyIterations)) AND

          // Mount options...
          (keyEntryDlg.GetDriveLetter(userDriveLetter)) AND
          (keyEntryDlg.GetReadonly(mountReadonly)) AND
          (keyEntryDlg.GetMountAs(mountMountAs)) AND

          // File options...
          (keyEntryDlg.GetOffset(fileOptOffset)) AND
          (keyEntryDlg.GetOffsetPointsToCDB(offsetPointsToCDB)) AND
          (keyEntryDlg.GetSaltLength(fileOptSaltLength))
         ) then
        begin
        mountForAllUsers := keyEntryDlg.GetMountForAllUsers();

        if retVal then
          begin
          // For each volume
          for i:=0 to (volumeFilenames.count-1) do
            begin
            currMountFilename := volumeFilenames[i];

            // Determine where the volume's critical data is read from; the
            // volume file itself (starting from the user specified offset), or
            // a keyfile (starting from zero offset)
            criticalDataFile   := volumeFilenames[i];
            criticalDataOffset := fileOptOffset;
            if (keyfile <> '') then
              begin
              criticalDataFile   := keyfile;
              criticalDataOffset := 0;
              end;

            // Get volume's decrypted critical data block
            // Attempt to obtain the current file's critical data
            if not(ReadVolumeCriticalData(
                                          criticalDataFile,
                                          criticalDataOffset,
                                          userKey,
                                          fileOptSaltLength,  // In *bits*
                                          keyIterations,
                                          volumeDetails,
                                          CDBMetaData
                                         )) then
              begin
              // Wrong password, bad volume file, correct hash/cypher driver not installed
              LastErrorCode := OTFE_ERR_WRONG_PASSWORD;
              retVal := FALSE;
              // Bail out...
              break;
              end;

            // Mount the volume
            if (volumeDetails.RequestedDriveLetter = #0) then
              begin
              // Nudge on to prevent getting drive A: or B:
              volumeDetails.RequestedDriveLetter := 'C';
              end;
            mountDriveLetter := GetNextDriveLetter(userDriveLetter, volumeDetails.RequestedDriveLetter);
            if (mountDriveLetter = #0) then
              begin
              // Skip onto next volume file...
              mountedAs := mountedAs + #0;
              Continue;
              end;

            // Locate where the actual encrypted partition starts within the
            // volume file
            volFileOffset := fileOptOffset;
            if (offsetPointsToCDB) then
              begin
              volFileOffset := volFileOffset + int64((CRITICAL_DATA_LENGTH div 8));
              end;

            // Attempt to create a new device to be used
            deviceName := CreateDiskDevice(FreeOTFEMountAsDeviceType[mountMountAs]);
            if (deviceName <> '') then
              begin
              // Attempt to mount the device
              if MountDiskDevice(
                                 deviceName,
                                 currMountFilename,
                                 volumeDetails.MasterKey,
                                 volumeDetails.SectorIVGenMethod,
                                 volumeDetails.VolumeIV,
                                 mountReadonly,
                                 CDBMetaData.HashDriverKernelModeName,
                                 CDBMetaData.HashGUID,
                                 CDBMetaData.CypherDriverKernelModeName,  // IV cypher
                                 CDBMetaData.CypherGUID,  // IV cypher
                                 CDBMetaData.CypherDriverKernelModeName,  // Main cypher
                                 CDBMetaData.CypherGUID,  // Main cypher
                                 volumeDetails.VolumeFlags,
                                 '',
                                 volFileOffset,
                                 volumeDetails.PartitionLen,
                                 FreeOTFEMountAsStorageMediaType[mountMountAs]
                                ) then
                begin
                if DosDeviceSymlink(
                                    TRUE,
                                    deviceName,
                                    mountDriveLetter,
                                    mountForAllUsers
                                   ) then
                  begin
                  mountedAs := mountedAs + mountDriveLetter;
                  end
                else
                  begin
                  // Cleardown
                  DismountDiskDevice(deviceName, TRUE);
                  DestroyDiskDevice(deviceName);
                  mountedAs := mountedAs + #0;
                  LastErrorCode := OTFE_ERR_MOUNT_FAILURE;
                  retVal := FALSE;
                  end;  // ELSE PART - if DefineDosDevice(

                end
              else
                begin
                // Cleardown
                // LastErrorCode set by MountDiskDevice
                DestroyDiskDevice(deviceName);
                mountedAs := mountedAs + #0;
                retVal := FALSE;
                end;  // ELSE PART - if MountDiskDevice(

              end;  // ELSE PART - if (deviceName <> '') then

            end;  // for i:=0 to (volumeFilenames.count) do

          end;  // if retVal then
        end;

      end;


  finally
    keyEntryDlg.Free()
  end;


  // Unmounted volume files...
  // Yes, it is "+1"; if you have only 1 volume file, you want exactly one #0
  for i:=(length(mountedAs)+1) to (volumeFilenames.count) do
    begin
    mountedAs := mountedAs + #0;

    // This should not be needed; included to ensure sanity...
    retVal := FALSE;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.MountLinux(volumeFilename: string; readonly: boolean = FALSE): char;
var
  tmpStringList: TStringList;
  mountedAs: string;
  retVal: char;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := #0;

  tmpStringList:= TStringList.Create();
  try
    tmpStringList.Add(volumeFilename);
    if MountLinux(tmpStringList, mountedAs, readonly) then
      begin
      retVal := mountedAs[1];
      end;

  finally
    tmpStringList.Free();
  end;

  Result :=  retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.MountLinux(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean;
var
  retVal: boolean;
  deviceName: string;
  keyEntryDlg: TfrmKeyEntryLinux;
  i: integer;
  volumeKey: string;

  userKey: string;
  keyProcSeed: string;
  keyProcHashDriver: string;
  keyProcHashGUID: TGUID;
  keyProcHashWithAs: boolean;
  keyProcCypherDriver: string;
  keyProcCypherGUID: TGUID;
  keyProcIterations: integer;
  fileOptOffset: int64;
  fileOptSize: int64;
  mountDriveLetter: char;
  mountReadonly: boolean;
  mainCypherDriver: string;
  mainCypherGUID: TGUID;
  mainIVHashDriver: string;
  mainIVHashGUID: TGUID;
  mainIVCypherDriver: string;
  mainIVCypherGUID: TGUID;
  currDriveLetter: char;
  sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
  startOfVolFile: boolean;
  startOfEndData: boolean;
  mountMountAs: TFreeOTFEMountAs;
  VolumeFlags: DWORD;
  mr: integer;
  mainCypherDetails: TFreeOTFECypher;
  mountForAllUsers: boolean;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := TRUE;

  CheckActive();

  // Sanity checking
  if (volumeFilenames.count = 0) then
    begin
    mountedAs := '';
    Result := TRUE;
    exit;
    end
  else
    begin
    // Is the user attempting to mount Linux LUKS volumes?
    // Mount as LUKS if they are...
    if (IsLUKSVolume(volumeFilenames[0])) then
      begin
      Result := MountLUKS(volumeFilenames, mountedAs, readonly);
      exit;
      end;
    end;
    

  keyEntryDlg := TfrmKeyEntryLinux.Create(nil);
  try
    keyEntryDlg.OTFEFreeOTFEObj := self;
    keyEntryDlg.SetReadonly(readonly);

    mr := keyEntryDlg.ShowModal();
    // Get user password, drive letter to use, etc
    if (mr = mrCancel) then
      begin
      LastErrorCode := OTFE_ERR_USER_CANCEL;
      end
    else if (mr = mrOK) then
      begin
      // Retrieve all data from the mount dialog...
      if (
          // Key...
          (keyEntryDlg.GetKey(userKey)) AND

          // Key processing...
          (keyEntryDlg.GetKeyProcSeed(keyProcSeed)) AND
          (keyEntryDlg.GetKeyProcHashKernelDeviceName(keyProcHashDriver)) AND
          (keyEntryDlg.GetKeyProcHashGUID(keyProcHashGUID)) AND
          (keyEntryDlg.GetKeyProcHashWithAs(keyProcHashWithAs)) AND
          (keyEntryDlg.GetKeyProcCypherKernelDeviceName(keyProcCypherDriver)) AND
          (keyEntryDlg.GetKeyProcCypherGUID(keyProcCypherGUID)) AND
          (keyEntryDlg.GetKeyProcCypherIterationCount(keyProcIterations)) AND

          // File options...
          (keyEntryDlg.GetOffset(fileOptOffset)) AND
          (keyEntryDlg.GetSizeLimit(fileOptSize)) AND

          // Mount options...
          (keyEntryDlg.GetDriveLetter(mountDriveLetter)) AND
          (keyEntryDlg.GetReadonly(mountReadonly)) AND
          (keyEntryDlg.GetMountAs(mountMountAs)) AND

          // Encryption options...
          (keyEntryDlg.GetMainCypherKernelDeviceName(mainCypherDriver)) AND
          (keyEntryDlg.GetMainCypherGUID(mainCypherGUID)) AND
          (keyEntryDlg.GetMainSectorIVGenMethod(sectorIVGenMethod)) AND
          (keyEntryDlg.GetMainIVSectorZeroPos(startOfVolFile, startOfEndData)) AND
          (keyEntryDlg.GetMainIVHashKernelDeviceName(mainIVHashDriver)) AND
          (keyEntryDlg.GetMainIVHashGUID(mainIVHashGUID)) AND
          (keyEntryDlg.GetMainIVCypherKernelDeviceName(mainIVCypherDriver)) AND
          (keyEntryDlg.GetMainIVCypherGUID(mainIVCypherGUID))
         ) then
        begin
        mountForAllUsers := keyEntryDlg.GetMountForAllUsers();

        // Before we can generate the volume key for encryption/decryption, we
        // need to know the keysize of the cypher
        if not(GetSpecificCypherDetails(mainCypherDriver, mainCypherGUID, mainCypherDetails)) then
          begin
          retVal := FALSE;
          end;


        // Generate volume key for encryption/decryption
        if (retVal) then
          begin
          if not(GenerateLinuxVolumeKey(
                                        keyProcHashDriver,
                                        keyProcHashGUID,
                                        userKey,
                                        keyProcSeed,
                                        keyProcHashWithAs,
                                        mainCypherDetails.KeySize,
                                        volumeKey
                                       )) then
            begin
            LastErrorCode := OTFE_ERR_HASH_FAILURE;
            retVal := FALSE;
            end;
          end;


        // Encode IV generation method to flags...
        VolumeFlags := 0;
        if (startOfVolFile) then
          begin
          VolumeFlags := VolumeFlags OR VOL_FLAGS_SECTOR_ID_ZERO_VOLSTART;
          end;


        if retVal then
          begin
          for i:=0 to (volumeFilenames.count-1) do
            begin
            currDriveLetter := GetNextDriveLetter(mountDriveLetter, #0);
            if (currDriveLetter = #0) then
              begin
              // No more drive letters following the user's specified drive
              // letter - don't mount further drives
              retVal := FALSE;
              // Bail out...
              break;
              end;


            // Attempt to create a new device to be used
            deviceName := CreateDiskDevice(FreeOTFEMountAsDeviceType[mountMountAs]);
            if (deviceName <> '') then
              begin
              // Attempt to mount the device
              if MountDiskDevice(
                                 deviceName,
                                 volumeFilenames[i],
                                 volumeKey,
                                 sectorIVGenMethod,
                                 '',  // Linux volumes don't have per-volume IVs
                                 mountReadonly,
                                 mainIVHashDriver,
                                 mainIVHashGUID,
                                 mainIVCypherDriver,
                                 mainIVCypherGUID,
                                 mainCypherDriver,
                                 mainCypherGUID,
                                 VolumeFlags,
                                 '',
                                 fileOptOffset,
                                 fileOptSize,
                                 FreeOTFEMountAsStorageMediaType[mountMountAs]
                                ) then
                begin
                if DosDeviceSymlink(
                                    TRUE,
                                    deviceName,
                                    currDriveLetter,
                                    mountForAllUsers
                                   ) then
                  begin
                  mountedAs := mountedAs + currDriveLetter;
                  end
                else
                  begin
                  // Cleardown
                  DismountDiskDevice(deviceName, TRUE);
                  DestroyDiskDevice(deviceName);
                  mountedAs := mountedAs + #0;
                  retVal := FALSE;
                  end;  // ELSE PART - if DefineDosDevice(

                end
              else
                begin
                // Cleardown
                DestroyDiskDevice(deviceName);
                mountedAs := mountedAs + #0;
                retVal := FALSE;
                end;  // ELSE PART - if MountDiskDevice(

              end;  // ELSE PART - if (deviceName <> '') then

            end;  // for i:=0 to (volumeFilenames.count) do

          end;  // if retVal then
        end;

      end;


  finally
    keyEntryDlg.Free()
  end;


  // Unmounted volume files...
  // Yes, it is "+1"; if you have only 1 volume file, you want exactly one #0
  for i:=(length(mountedAs)+1) to (volumeFilenames.count) do
    begin
    mountedAs := mountedAs + #0;

    // This should not be needed; included to ensure sanity...
    retVal := FALSE;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
{$IFDEF LINUX_DETECT}
// This function was included to allow the detection of Linux encrypted
// volume's settings
// This function is *not* normally built, and it's inclusion is only for
// development/diagnostic purposes; not for general use
// It operates by being given the volume's password, and the filename of a file
// stored on the encrypted volume.
// This function cycles through all possible combinations of Linux settings for
// that password, attempting to mount the volume using the given password.
// It does not terminate on a successful mount, as there may be more than one
// set of settings which mount correctly
// IMPORTANT: The encrypted volume *must* be formatted with a filesystem
//            MS Windows understands (e.g. FAT)
// WARNING: This procedure can take some time to complete(!)
// WARNING: Some parts of this function may be commented out in order to
//          increase it's speed, but at a cost of not checking all combinations
// NOTE: Makes use of GExpert's (3rd party) debugging tool to pump out details
//       of successful mounts before this function returns
function TOTFEFreeOTFE.DetectLinux(
  volumeFilename: string;
  userKey: string;
  keyProcSeed: string;
  keyProcIterations: integer;
  fileOptOffset: int64;
  fileOptSize: int64;
  mountDriveLetter: char;
  mountMountAs: TFreeOTFEMountAs;
  testFilename: string
  ): string;
var
  retVal: string;
  deviceName: string;
  keyEntryDlg: TfrmKeyEntryLinux;
  i: integer;
  volumeKey: string;

  keyProcHashDriver: string;
  keyProcHashGUID: TGUID;
  keyProcHashWithAs: boolean;
  keyProcCypherDriver: string;
  keyProcCypherGUID: TGUID;
  mainCypherDriver: string;
  mainCypherGUID: TGUID;
  mainIVHashDriver: string;
  mainIVHashGUID: TGUID;
  mainIVCypherDriver: string;
  mainIVCypherGUID: TGUID;
  currDriveLetter: char;
  startOfVolFile: boolean;
  startOfEndData: boolean;
  VolumeFlags: DWORD;
  mr: integer;
  mainCypherDetails: TFreeOTFECypher;

  hashDrivers: array of TFreeOTFEHashDriver;
  cypherDrivers: array of TFreeOTFECypherDriver;

  kph_i, kph_j: integer;
  kpc_i, kpc_j: integer;
  mc_i, mc_j: integer;
  hashWithAsInt: integer;
  ivh_i, ivh_j: integer;
  ivc_i, ivc_j: integer;
  sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
  ivStartSectorInt: integer;

  kph_currDriver: TFreeOTFEHashDriver;
  kph_currImpl: TFreeOTFEHash;
  kpc_currDriver: TFreeOTFECypherDriver;
  kpc_currImpl: TFreeOTFECypher;
  mc_currDriver: TFreeOTFECypherDriver;
  mc_currImpl: TFreeOTFECypher;
  ivh_currDriver: TFreeOTFEHashDriver;
  ivh_currImpl: TFreeOTFEHash;
  ivc_currDriver: TFreeOTFECypherDriver;
  ivc_currImpl: TFreeOTFECypher;

begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := '';

  CheckActive();

  // Sanity checking
  // Is the user attempting to mount Linux LUKS volumes?
  // Mount as LUKS if they are...
  if (IsLUKSVolume(volumeFilename)) then
    begin
    retVal := retVal + 'LUKS volume; no need to do exhaustive search';
    exit;
    end;


  if (retVal = '') then
    begin
    // Obtain details of all hashes...
    SetLength(hashDrivers, 0);
    if not(GetHashDrivers(TFreeOTFEHashDriverArray(hashDrivers))) then
      begin
      retVal := retVal + 'Unable to get list of hash drivers.';
      end;
    end;

  if (retVal = '') then
    begin
    // Obtain details of all cyphers...
    SetLength(cypherDrivers, 0);
    if not(GetCypherDrivers(TFreeOTFECypherDriverArray(cypherDrivers))) then
      begin
      retVal := retVal + 'Unable to get list of cypher drivers.';
      end;
    end;


  if (retVal = '') then
    begin
    currDriveLetter := GetNextDriveLetter(mountDriveLetter, #0);
    if (currDriveLetter = #0) then
      begin
      // No more drive letters following the user's specified drive
      // letter - don't mount further drives
      retVal := retVal + 'Out of drive letters.';
      end;
    end;


  if (retVal = '') then
    begin
    // KEY PROCESSING
    // FOR ALL HASH DRIVERS...
    SendDateTime('START', now);
    for kph_i:=low(hashDrivers) to high(hashDrivers) do
      begin
      SendDebug('-kph_i: '+inttostr(kph_i)+'/'+inttostr(high(hashDrivers) - low(hashDrivers)));
      kph_currDriver := hashDrivers[kph_i];
      // FOR ALL HASHES SUPPORTED BY THE CURRENT DRIVER...
      for kph_j:=low(kph_currDriver.Hashes) to high(kph_currDriver.Hashes) do
        begin
        SendDebug('--kph_j: '+inttostr(kph_j)+'/'+inttostr(high(kph_currDriver.Hashes) - low(kph_currDriver.Hashes)));
        kph_currImpl := kph_currDriver.Hashes[kph_j];

        keyProcHashDriver := kph_currDriver.DeviceKernelModeName;
        keyProcHashGUID := kph_currImpl.HashGUID;

        // KEY PROCESSING
        // FOR ALL CYPHER DRIVERS...
        for kpc_i:=low(cypherDrivers) to high(cypherDrivers) do
          begin
          SendDebug('---kpc_i: '+inttostr(kpc_i)+'/'+inttostr(high(cypherDrivers) - low(cypherDrivers)));
          kpc_currDriver := cypherDrivers[kpc_i];
          // FOR ALL CYPHERS SUPPORTED BY THE CURRENT DRIVER...
          for kpc_j:=low(kpc_currDriver.Cyphers) to high(kpc_currDriver.Cyphers) do
            begin
            SendDebug('----kpc_j: '+inttostr(kpc_j)+'/'+inttostr(high(kpc_currDriver.Cyphers) - low(kpc_currDriver.Cyphers)));
            kpc_currImpl := kpc_currDriver.Cyphers[kpc_j];

            keyProcCypherDriver := kpc_currDriver.DeviceKernelModeName;
            keyProcCypherGUID := kpc_currImpl.CypherGUID;

//            // MAIN CYPHER
//            // FOR ALL CYPHER DRIVERS...
//            for mc_i:=low(cypherDrivers) to high(cypherDrivers) do
//              begin
//              mc_currDriver := cypherDrivers[mc_i];
//              // FOR ALL CYPHERS SUPPORTED BY THE CURRENT DRIVER...
//              for mc_j:=low(mc_currDriver.Cyphers) to high(mc_currDriver.Cyphers) do
//                begin
//                mc_currImpl := kpc_currDriver.Cyphers[mc_j];

// xxx
// xxx
mc_currDriver := kpc_currDriver;
mc_currImpl := kpc_currImpl;

                mainCypherDriver := mc_currDriver.DeviceKernelModeName;
                mainCypherGUID := mc_currImpl.CypherGUID;
                mainCypherDetails := mc_currImpl;


                for hashWithAsInt:=0 to 1 do
                  begin
                  SendDebug('-----hashWithAsInt: '+inttostr(hashWithAsInt)+'/(0-1)');

                  // Generate volume key for encryption/decryption
                  if not(GenerateLinuxVolumeKey(
                                                keyProcHashDriver,
                                                keyProcHashGUID,
                                                userKey,
                                                keyProcSeed,
                                                (hashWithAsInt = 1),
                                                mainCypherDetails.KeySize,
                                                volumeKey
                                               )) then
                    begin
                    retVal := retVal + 'HASH FAILURE: '+keyProcHashDriver+':'+GUIDToString(keyProcHashGUID);
                    continue;
                    end;


                  // IV GENERATION
                  // FOR ALL HASH DRIVERS...
                  for ivh_i:=low(hashDrivers) to high(hashDrivers) do
                    begin
                    SendDebug('------ivh_i: '+inttostr(ivh_i)+'/'+inttostr(high(hashDrivers) - low(hashDrivers)));
                    ivh_currDriver := hashDrivers[ivh_i];
                    // FOR ALL HASHES SUPPORTED BY THE CURRENT DRIVER...
                    for ivh_j:=low(ivh_currDriver.Hashes) to high(ivh_currDriver.Hashes) do
                      begin
                      SendDebug('-------ivh_j: '+inttostr(ivh_j)+'/'+inttostr(high(ivh_currDriver.Hashes) - low(ivh_currDriver.Hashes)));
                      ivh_currImpl := ivh_currDriver.Hashes[ivh_j];

                      mainIVHashDriver := ivh_currDriver.DeviceKernelModeName;
                      mainIVHashGUID := ivh_currImpl.HashGUID;

//                      // IV GENERATION
//                      // FOR ALL CYPHER DRIVERS...
//                      for ivc_i:=low(cypherDrivers) to high(cypherDrivers) do
//                        begin
//                        ivc_currDriver := cypherDrivers[ivc_i];
//                        // FOR ALL CYPHERS SUPPORTED BY THE CURRENT DRIVER...
//                        for ivc_j:=low(ivc_currDriver.Cyphers) to high(ivc_currDriver.Cyphers) do
//                          begin
//                          ivc_currImpl := ivc_currDriver.Cyphers[ivc_j];

// xxx
ivc_currDriver := kpc_currDriver;
ivc_currImpl := kpc_currImpl;

                          mainIVCypherDriver := ivc_currDriver.DeviceKernelModeName;
                          mainIVCypherGUID := ivc_currImpl.CypherGUID;

                          // IV GENERATION
//                          for sectorIVGenMethod := low(sectorIVGenMethod) to high(sectorIVGenMethod) do
sectorIVGenMethod := foivg32BitSectorID;
                            begin
//                            SendDebug('--------sectorIVGenMethod: '+inttostr(ord(sectorIVGenMethod))+'/'+inttostr(ord(high(sectorIVGenMethod)) - loword((sectorIVGenMethod))));

//                            for ivStartSectorInt:=0 to 1 do
                              begin
ivStartSectorInt := 1;
//                              SendDebug('---------ivStartSectorInt: '+inttostr(ivStartSectorInt)+'/(0-1)');

                              startOfVolFile := (ivStartSectorInt = 0);
                              startOfEndData := (ivStartSectorInt = 1);

                              // Encode IV generation method to flags...
                              VolumeFlags := 0;
                              if (startOfVolFile) then
                                begin
                                VolumeFlags := VolumeFlags OR VOL_FLAGS_SECTOR_ID_ZERO_VOLSTART;
                                end;


                              // Attempt to create a new device to be used
                              deviceName := CreateDiskDevice(FreeOTFEMountAsDeviceType[mountMountAs]);
                              if (deviceName <> '') then
                                begin
                                // Attempt to mount the device
                                if MountDiskDevice(
                                                   deviceName,
                                                   volumeFilename,
                                                   volumeKey,
                                                   sectorIVGenMethod,
                                                   '',  // Linux volumes don't have per-volume IVs
                                                   TRUE,
                                                   mainIVHashDriver,
                                                   mainIVHashGUID,
                                                   mainIVCypherDriver,
                                                   mainIVCypherGUID,
                                                   mainCypherDriver,
                                                   mainCypherGUID,
                                                   VolumeFlags,
                                                   '',
                                                   fileOptOffset,
                                                   fileOptSize,
                                                   FreeOTFEMountAsStorageMediaType[mountMountAs]
                                                  ) then
                                  begin
                                  if DosDeviceSymlink(
                                                      TRUE,
                                                      deviceName,
                                                      currDriveLetter,
                                                      TRUE
                                                     ) then
                                    begin
                                    // Test if mounted OK
                                    if FileExists(currDriveLetter+':'+testFilename) then
                                      begin
                                      retVal := retVal + '---------------------------------------------------------------';
                                      retVal := retVal + 'OK!';
                                      retVal := retVal + 'mountedAs: '+currDriveLetter;
                                      retVal := retVal + 'kph: '+keyProcHashDriver +':'+GUIDToString(keyProcHashGUID );
                                      retVal := retVal + 'kpc: '+keyProcCypherDriver +':'+GUIDToString(keyProcCypherGUID );
                                      retVal := retVal + 'mc: '+mainCypherDriver +':'+GUIDToString(mainCypherGUID);
                                      retVal := retVal + 'hashWithAsInt: '+inttostr(hashWithAsInt);
                                      retVal := retVal + 'ivh: '+mainIVHashDriver +':'+GUIDToString(mainIVHashGUID );
                                      retVal := retVal + 'ivc: '+mainIVCypherDriver +':'+GUIDToString(mainIVCypherGUID );
                                      retVal := retVal + 'sectorIVGenMethod: '+inttostr(ord(sectorIVGenMethod));
                                      retVal := retVal + 'ivStartSectorInt: '+inttostr(ivStartSectorInt);
                                      retVal := retVal + '---------------------------------------------------------------';
SendDateTime('HIT', now);
SendDebug('retVal: '+retVal);
                                      end;

                                    Dismount(currDriveLetter);
                                    end
                                  else
                                    begin
                                    // Cleardown
                                    DismountDiskDevice(deviceName, TRUE);
                                    DestroyDiskDevice(deviceName);
                                    end;  // ELSE PART - if DefineDosDevice(

                                  end
                                else
                                  begin
                                  // Cleardown
                                  DestroyDiskDevice(deviceName);
                                  end;  // ELSE PART - if MountDiskDevice(

                                end;  // ELSE PART - if (deviceName <> '') then


                              end;  // for ivStartSectorInt:=0 to 1 do
                            end;  // for sectorIVGenMethod := low(sectorIVGenMethod) to high(sectorIVGenMethod) do
//                          end;  // for ivc_j:=low(ivc_currDriver.Cyphers) to high(ivc_currDriver.Cyphers) do
//                        end;  // for ivc_i:=low(cypherDrivers) to high(cypherDrivers) do
                      end;  // for ivh_j:=low(ivh_currDriver.Hashes) to high(ivh_currDriver.Hashes) do
                    end;  // for ivh_i:=low(hashDrivers) to high(hashDrivers) do
                  end;  // for hashWithAsInt:=0 to 1 do
//                end;  // for mc_j:=low(mc_currDriver.Cyphers) to high(mc_currDriver.Cyphers) do
//              end;  // for mc_i:=low(cypherDrivers) to high(cypherDrivers) do
            end;  // for kpc_j:=low(kpc_currDriver.Cyphers) to high(kpc_currDriver.Cyphers) do
          end;  // for kpc_i:=low(cypherDrivers) to high(cypherDrivers) do
        end;  // for kph_j:=low(kph_currDriver.Hashes) to high(kph_currDriver.Hashes) do
      end;  // for kph_i:=low(hashDrivers) to high(hashDrivers) do
    end;  //  if (retVal = '') then

SendDateTime('FINISHED', now);

  Result := retVal;
end;
{$ENDIF}

// ----------------------------------------------------------------------------
// Generate a Linux volume's volume key
// hashWithAs - If set to FALSE, the user's key will be seeded and hashed once
//              the result will be right padded with 0x00 chars/truncated as
//              appropriate
//              If set to TRUE, then:
//                If "targetKeylengthBits" is > -1, the user's key will be
//                seeded and hashed once
//                Otherwise, the volume key will be created by concatenating
//                  Hashing the user's seeded key
//                  Hashing 'A' + the user's seeded key
//                  Hashing 'AA' + the user's seeded key
//                  Hashing 'AAA' + the user's seeded key
//                etc - until the required number of bits is reached
function TOTFEFreeOTFE.GenerateLinuxVolumeKey(
                                              hashDriver: string;
                                              hashGUID: TGUID;
                                              userKey: string;
                                              seed: string;
                                              hashWithAs: boolean;
                                              targetKeylengthBits: integer;
                                              var volumeKey: string
                                             ): boolean;
var
  seededKey: string;
  hashOutput: string;
  allOK: boolean;
begin
  allOK:= TRUE;

  // Salt the user's password
  seededKey := userKey + seed;


  // If the target key length is any size key, or we don't hash with A's, then
  // we simply hash the user's key once with the specified hash
  if (
      (targetKeylengthBits = -1) or
      not(hashWithAs)
     ) then
    begin
    allOK := HashData(hashDriver, hashGUID, seededKey, volumeKey);
    end
  else
    begin
    // We're been asked to supply a volume key with a fixed, defined length
    while (
           allOK and
           (Length(volumeKey) < (targetKeylengthBits div 8))
          ) do
      begin
      // Note: We'll truncate later - don't worry if volumeKey is set to too
      //       many bits at this stage...
      allOK := allOK and HashData(hashDriver, hashGUID, seededKey, hashOutput);
      volumeKey := volumeKey + hashOutput;
      // Prepare for next...
      seededKey := 'A'+seededKey; 
      end;

    end;


  // If the target key length is > -1, truncate or we right-pad with 0x00
  // as appropriate
  if (allOK) then
    begin
    if (targetKeylengthBits > -1) then
      begin
      // Copy as much of the hash value as possible to match the key length
      volumeKey := Copy(volumeKey, 1, min((targetKeylengthBits div 8), length(volumeKey)));
      // Right-pad if needed
      volumeKey := volumeKey + StringOfChar(#0, ((targetKeylengthBits div 8) - Length(volumeKey)));
      end;

    end;


  if not(allOK) then
    begin
    LastErrorCode := OTFE_ERR_HASH_FAILURE;
    end;

  Result := allOK;
end;



// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.MountLUKS(volumeFilename: string; readonly: boolean = FALSE): char;
var
  tmpStringList: TStringList;
  mountedAs: string;
  retVal: char;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := #0;

  tmpStringList:= TStringList.Create();
  try
    tmpStringList.Add(volumeFilename);
    if MountLUKS(tmpStringList, mountedAs, readonly) then
      begin
      retVal := mountedAs[1];
      end;

  finally
    tmpStringList.Free();
  end;

  Result :=  retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.MountLUKS(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean;
var
  retVal: boolean;
  deviceName: string;
  keyEntryDlg: TfrmKeyEntryLUKS;
  i: integer;
  volumeKey: string;

  userKey: string;
  fileOptSize: int64;
  mountDriveLetter: char;
  mountReadonly: boolean;
  currDriveLetter: char;
  mountMountAs: TFreeOTFEMountAs;
  VolumeFlags: DWORD;
  mr: integer;
  LUKSHeader: TLUKSHeader;
  keySlot: integer;
  baseIVCypherOnHashLength: boolean;
  mountForAllUsers: boolean;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retVal := TRUE;

  CheckActive();

  keyEntryDlg := TfrmKeyEntryLUKS.Create(nil);
  try
    keyEntryDlg.OTFEFreeOTFEObj := self;
    keyEntryDlg.SetReadonly(readonly);

    mr := keyEntryDlg.ShowModal();
    // Get user password, drive letter to use, etc
    if (mr = mrCancel) then
      begin
      LastErrorCode := OTFE_ERR_USER_CANCEL;
      end
    else if (mr = mrOK) then
      begin
      // Retrieve all data from the mount dialog...
      if (
          // Key...
          (keyEntryDlg.GetKey(userKey)) AND
          (keyEntryDlg.GetIVCypherBase(baseIVCypherOnHashLength)) AND

          // File options...
          (keyEntryDlg.GetSizeLimit(fileOptSize)) AND

          // Mount options...
          (keyEntryDlg.GetDriveLetter(mountDriveLetter)) AND
          (keyEntryDlg.GetReadonly(mountReadonly)) AND
          (keyEntryDlg.GetMountAs(mountMountAs))
         ) then
        begin
        mountForAllUsers := keyEntryDlg.GetMountForAllUsers();

        for i:=0 to (volumeFilenames.count-1) do
          begin
          currDriveLetter := GetNextDriveLetter(mountDriveLetter, #0);
          if (currDriveLetter = #0) then
            begin
            // No more drive letters following the user's specified drive
            // letter - don't mount further drives
            retVal := FALSE;
            // Bail out...
            break;
            end;

          if not(RecoverMasterKey(
                              volumeFilenames[i],
                              userKey,
                              baseIVCypherOnHashLength,
                              LUKSHeader,
                              keySlot,
                              volumeKey
                             )) then
            begin
            retVal := FALSE;
            // Bail out...
            break;
            end;

{$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Recovered key OK');
  DebugMsg('key slot '+inttostr(keySlot)+' unlocked.');
  DebugMsg('Master key: '+PrettyHex(
                                    PChar(volumeKey),
                                    length(volumeKey)
                                   ));
{$ENDIF}

          // Attempt to create a new device to be used
          deviceName := CreateDiskDevice(FreeOTFEMountAsDeviceType[mountMountAs]);
          if (deviceName <> '') then
            begin
            // Attempt to mount the device
            VolumeFlags := 0;
            if MountDiskDevice(
                               deviceName,
                               volumeFilenames[i],
                               volumeKey,
                               LUKSHeader.sectorIVGenMethod,
                               '',  // Linux volumes don't have per-volume IVs
                               mountReadonly,
                               LUKSHeader.IVHashKernelModeDeviceName,
                               LUKSHeader.IVHashGUID,
                               LUKSHeader.IVCypherKernelModeDeviceName,
                               LUKSHeader.IVCypherGUID,
                               LUKSHeader.cypherKernelModeDeviceName,
                               LUKSHeader.cypherGUID,
                               VolumeFlags,
                               '',
                               (LUKSHeader.payload_offset * LUKS_SECTOR_SIZE),
                               fileOptSize,
                               FreeOTFEMountAsStorageMediaType[mountMountAs]
                              ) then
              begin
              if DosDeviceSymlink(
                                    TRUE,
                                    deviceName,
                                    currDriveLetter,
                                    mountForAllUsers
                                   ) then
                begin
                mountedAs := mountedAs + currDriveLetter;
                end
              else
                begin
                // Cleardown
                DismountDiskDevice(deviceName, TRUE);
                DestroyDiskDevice(deviceName);
                mountedAs := mountedAs + #0;
                retVal := FALSE;
                end;  // ELSE PART - if DefineDosDevice(

              end
            else
              begin
              // Cleardown
              DestroyDiskDevice(deviceName);
              mountedAs := mountedAs + #0;
              retVal := FALSE;
              end;  // ELSE PART - if MountDiskDevice(

            end;  // ELSE PART - if (deviceName <> '') then

          end;  // for i:=0 to (volumeFilenames.count) do

        end;

      end;


  finally
    keyEntryDlg.Free()
  end;


  // Unmounted volume files...
  // Yes, it is "+1"; if you have only 1 volume file, you want exactly one #0
  for i:=(length(mountedAs)+1) to (volumeFilenames.count) do
    begin
    mountedAs := mountedAs + #0;

    // This should not be needed; included to ensure sanity...
    retVal := FALSE;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Attempt to create a new device to be used
function TOTFEFreeOTFE.CreateDiskDevice(deviceType: DWORD): string;
var
  retVal: string;
  DIOCBufferIn: TDIOC_DISK_DEVICE_CREATE;
  DIOCBufferOut: TDIOC_DEVICE_NAME;
  bytesReturned: DWORD;
begin
  retVal := '';

  CheckActive();

  // !!! WARNING !!!
  // We use FILE_DEVICE_DISK as FILE_DEVICE_VIRTUAL_DISK can
  // make NTFS-only Windows XP systems unstable?
  // See MS Knowledgebase article: Q257405 for further details
  // http://support.microsoft.com/default.aspx?scid=kb;en-us;257405

  DIOCBufferIn.DeviceType := deviceType;

  if (DeviceIoControl(
                      DriverHandle,
                      IOCTL_FREEOTFE_CREATE,
                      @DIOCBufferIn,
                      sizeof(DIOCBufferIn),
                      @DIOCBufferOut,
                      sizeof(DIOCBufferOut),
                      bytesReturned,
                      nil
                     )) then
    begin
    retVal := copy(DIOCBufferOut.DeviceName, 1, StrLen(DIOCBufferOut.DeviceName));
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Disk device created: '+retVal);
{$ENDIF}
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Attempt to mount on an existing device
function TOTFEFreeOTFE.MountDiskDevice(
                                   deviceName: string;
                                   volFilename: string;
                                   volumeKey: string;
                                   sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
                                   volumeIV: string;
                                   readonly: boolean;
                                   IVHashDriver: string;
                                   IVHashGUID: TGUID;
                                   IVCypherDriver: string;
                                   IVCypherGUID: TGUID;
                                   mainCypherDriver: string;
                                   mainCypherGUID: TGUID;
                                   VolumeFlags: integer;
                                   metaData: string;
                                   offset: int64 = 0;
                                   size: int64 = 0;
                                   storageMediaType: TFreeOTFEStorageMediaType = mtFixedMedia
                                  ): boolean;
var
  retVal: boolean;
  ptrDIOCBuffer: PDIOC_MOUNT;
  bytesReturned: DWORD;
  bufferSize: integer;
  mainCypherDetails: TFreeOTFECypher;
  kmFilename: string;
begin
  retVal := FALSE;

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('In MountDiskDevice');
{$ENDIF}

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('MountDiskDevice called with: ');
DebugMsg('  deviceName: '+deviceName);
DebugMsg('  filename: '+volFilename);
DebugMsg('  volumeKey: ');
DebugMsgBinary(volumeKey);
DebugMsg('  IVHashDriver: '+IVHashDriver);
DebugMsg('  IVHashGUID: '+GUIDToString(IVHashGUID));
DebugMsg('  IVCypherDriver: '+IVCypherDriver);
DebugMsg('  IVCypherGUID: '+GUIDToString(IVCypherGUID));
DebugMsg('  mainCypherDriver: '+mainCypherDriver);
DebugMsg('  mainCypherGUID: '+GUIDToString(mainCypherGUID));
DebugMsg('  sectorIVGenMethod: '+FreeOTFESectorIVGenMethodTitle[sectorIVGenMethod]);
DebugMsg('  VolumeFlags: '+inttostr(VolumeFlags));
DebugMsg('  metaData: ');
DebugMsgBinary(metaData);
DebugMsg('  offset: '+inttostr(offset));
DebugMsg('  size: '+inttostr(size));
{$ENDIF}

  CheckActive();


  if not(GetSpecificCypherDetails(mainCypherDriver, mainCypherGUID, mainCypherDetails)) then
    begin
    Result := retval;
    exit;
    end;

  // Ensure the volumeKey's length matches that of the cypher's keysize, if
  // the cypher's keysize is >= 0
  if (mainCypherDetails.KeySize >= 0) then
    begin
    // THIS IS CORRECT; caters for both volumeKey>keysize and
    // volumeKey<keysize
    // Copy as much of the hash value as possible to match the key length
    volumeKey := Copy(volumeKey, 1, min((mainCypherDetails.KeySize div 8), length(volumeKey)));
    // If the hash wasn't big enough, pad out with zeros
    volumeKey := volumeKey + StringOfChar(#0, ((mainCypherDetails.KeySize div 8) - Length(volumeKey)));
    end;


  // If the sector IV doesn't use hashing, don't pass a hash algorithm in
  if not(SCTRIVGEN_USES_HASH[sectorIVGenMethod]) then
    begin
    IVHashDriver := '';
    IVHashGUID := StringToGUID(NULL_GUID);
    end;

  // If the sector IV doesn't use hashing, don't pass a hash algorithm in
  if not(SCTRIVGEN_USES_CYPHER[sectorIVGenMethod]) then
    begin
    IVCypherDriver := '';
    IVCypherGUID := StringToGUID(NULL_GUID);
    end;


  // Subtract sizeof(char) - once for the volume key, once for the volumeIV
  bufferSize := sizeof(TDIOC_MOUNT)
                - sizeof(char) + (sizeof(char)*Length(volumeKey))
                - sizeof(char) + (sizeof(char)*Length(volumeIV))
                - sizeof(char) + (sizeof(char)*Length(metaData));
  // Round up to next "DIOC boundry". Although this always results in an
  // oversized buffer, it's guaranteed to be big enough.
  bufferSize := bufferSize + (DIOC_BOUNDRY - (bufferSize mod DIOC_BOUNDRY));
  ptrDIOCBuffer := allocmem(bufferSize);
  try
    StrPCopy(ptrDIOCBuffer.DiskDeviceName, deviceName);

    // Determine the filename as the kernel sees it
    kmFilename:= GetKernelModeVolumeFilename(volFilename);

    StrPCopy(ptrDIOCBuffer.Filename, kmFilename);
    ptrDIOCBuffer.DataStart := offset;

    ptrDIOCBuffer.DataEnd := 0;
    if (size > 0) then
      begin
      ptrDIOCBuffer.DataEnd := offset + size;
      end;

    StrPCopy(ptrDIOCBuffer.IVHashDeviceName, IVHashDriver);
    ptrDIOCBuffer.IVHashGUID := IVHashGUID;
    StrPCopy(ptrDIOCBuffer.IVCypherDeviceName, IVCypherDriver);
    ptrDIOCBuffer.IVCypherGUID := IVCypherGUID;

    StrPCopy(ptrDIOCBuffer.MainCypherDeviceName, mainCypherDriver);
    ptrDIOCBuffer.MainCypherGUID := mainCypherGUID;

    ptrDIOCBuffer.ReadOnly := readonly;
    ptrDIOCBuffer.MountSource := FreeOTFEMountSourceID[fomsFile];
    if IsPartition(kmFilename) then
      begin
      ptrDIOCBuffer.MountSource := FreeOTFEMountSourceID[fomsPartition];
      end;

    ptrDIOCBuffer.StorageMediaType := FreeOTFEStorageMediaTypeID[storageMediaType];

    ptrDIOCBuffer.VolumeFlags := VolumeFlags;
    ptrDIOCBuffer.SectorIVGenMethod := FreeOTFESectorIVGenMethodID[sectorIVGenMethod];

    ptrDIOCBuffer.MasterKeyLength := Length(volumeKey) * 8;
    ptrDIOCBuffer.VolumeIVLength := Length(volumeIV) * 8;
    ptrDIOCBuffer.MetaDataLength := Length(metaData);
    // This may seem a little weird, but we do this because the VolumeIV and metaData are
    // immediatly after the master key
    StrMove(@ptrDIOCBuffer.MasterKey, PChar(volumeKey), Length(volumeKey));
    StrMove(((PChar(@ptrDIOCBuffer.MasterKey))+length(volumeKey)), PChar(volumeIV), Length(volumeIV));
    StrMove(((PChar(@ptrDIOCBuffer.MasterKey))+length(volumeKey)+length(volumeIV)), PChar(metaData), Length(metaData));


    if (DeviceIoControl(
                        DriverHandle,
                        IOCTL_FREEOTFE_MOUNT,
                        ptrDIOCBuffer,
                        bufferSize,
                        nil,
                        0,
                        bytesReturned,
                        nil
                       )) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Mounted OK!');
{$ENDIF}
      retVal := TRUE;
      end;

  finally
    FreeMem(ptrDIOCBuffer);
  end;


{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Exiting MountDiskDevice');
{$ENDIF}

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Attempt to dismount a device
function TOTFEFreeOTFE.DismountDiskDevice(deviceName: string; emergency: boolean): boolean;
var
  retVal: boolean;
  DIOCBuffer: TDIOC_DISMOUNT;
  bytesReturned: DWORD;
begin
  retVal := FALSE;

  CheckActive();

  StrPCopy(DIOCBuffer.DiskDeviceName, deviceName);
  DIOCBuffer.Emergency := emergency;

  if (DeviceIoControl(
                      DriverHandle,
                      IOCTL_FREEOTFE_DISMOUNT,
                      @DIOCBuffer,
                      sizeof(DIOCBuffer),
                      nil,
                      0,
                      bytesReturned,
                      nil
                     )) then
    begin
    retVal := TRUE;
    end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// This dismount routine is based on the MS "media eject" routine for NT/2k/XP
// at:
// http://support.microsoft.com/default.aspx?scid=http://support.microsoft.com:80/support/kb/articles/Q165/7/21.asp&NoWebContent=1
function TOTFEFreeOTFE.LDREUUserApp(
                          driveLetter: char;
                          deviceName: string;
                          emergency: boolean;
                          var unableToOpenDrive: boolean
                         ): boolean;
var
  driveDevice: THandle;
  BytesReturned: DWORD;
  driveLetterColon: string;
  driveFile: string;
  pmr: TPREVENT_MEDIA_REMOVAL;
  allOK: boolean;
  volumeInfo: TOTFEFreeOTFEVolumeInfo;
begin
  CheckActive();

  driveLetterColon := upcase(driveLetter)+':';
  driveFile := '\\.\'+driveLetterColon;

  allOK := TRUE;

  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('getVolumeInfo');
{$ENDIF}
    allOK := GetVolumeInfo(driveLetter, volumeInfo);

    // If we couldn't get the drive info, then we can't even do an
    // emergency dismount; we don't know which FreeOTFE device to dismount
    if (not(allOK)) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('getVolumeInfo NOT ALLOK');
{$ENDIF}
      emergency := FALSE;
      end;
    end;



  // We attempt to open the drive, and then carry out various operations to
  // shut the drive down.
  // If were operating in an emergency, then we don't care if some of these
  // operations fail - it's an EMERGENCY! JUST GET RID OF THE %$!# DRIVE!


  // Initialize to flag that we haven't opened it
  driveDevice := INVALID_HANDLE_VALUE;
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('createfile: '+driveFile);
{$ENDIF}
    driveDevice := CreateFile(
                              PChar(driveFile),
                              //GENERIC_READ or GENERIC_WRITE,
                              GENERIC_READ,
                              FILE_SHARE_READ or FILE_SHARE_WRITE,
                              nil,
                              OPEN_EXISTING,
                              FILE_FLAG_NO_BUFFERING,
                              0
                             );
    allOK := (driveDevice <> INVALID_HANDLE_VALUE);
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('driveDevice: '+inttostr(driveDevice)+' (INVALID_HANDLE_VALUE = '+inttostr(INVALID_HANDLE_VALUE));
{$ENDIF}
    end;
    
  unableToOpenDrive := (driveDevice = INVALID_HANDLE_VALUE);

  if ( (allOK) or (emergency and (driveDevice<>INVALID_HANDLE_VALUE)) )then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('lockvolume');
{$ENDIF}
    allOK := DeviceIoControl(
                             driveDevice,
                             FSCTL_LOCK_VOLUME,
                             nil,
                             0,
                             nil,
                             0,
                             bytesReturned,
                             nil
                            );
    end;


  if ( (allOK) or (emergency and (driveDevice<>INVALID_HANDLE_VALUE)) )then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('dismountvolume');
{$ENDIF}
    allOK := DeviceIoControl(
                             driveDevice,
                             FSCTL_DISMOUNT_VOLUME,
                             nil,
                             0,
                             nil,
                             0,
                             bytesReturned,
                             nil
                            );
    end;


  if ( (allOK) or (emergency and (driveDevice<>INVALID_HANDLE_VALUE)) )then
    begin
    pmr.PreventMediaRemoval := FALSE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('mediaremoval');
{$ENDIF}
    allOK := DeviceIoControl(
                             driveDevice,
                             IOCTL_STORAGE_MEDIA_REMOVAL,
                             @pmr,
                             sizeof(pmr),
                             nil,
                             0,
                             bytesReturned,
                             nil
                            );
    end;


  if ( (allOK) or (emergency and (driveDevice<>INVALID_HANDLE_VALUE)) )then
    begin
    pmr.PreventMediaRemoval := FALSE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('ejectmedia');
{$ENDIF}
    allOK := DeviceIoControl(
                             driveDevice,
                             IOCTL_STORAGE_EJECT_MEDIA,
                             nil,
                             0,
                             nil,
                             0,
                             bytesReturned,
                             nil
                            );
    end;


  // An explicit call to IOCTL_FREEOTFE_DISMOUNT should be redundant; the
  // IOCTL_STORAGE_EJECT_MEDIA should have already done this
  if (allOK or emergency) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('dismountdevice');
{$ENDIF}
    allOK := DismountDiskDevice(volumeInfo.deviceName, emergency);
    end;


  if ( (allOK) or (emergency and (driveDevice<>INVALID_HANDLE_VALUE)) )then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('unlock');
{$ENDIF}
    allOK := DeviceIoControl(
                             driveDevice,
                             FSCTL_UNLOCK_VOLUME,
                             nil,
                             0,
                             nil,
                             0,
                             bytesReturned,
                             nil
                            );
    end;


  if (driveDevice<>INVALID_HANDLE_VALUE) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('closehandle');
{$ENDIF}
    // Note that we can't do:
    //   allOK := allOK and CloseHandle(driveDevice);"
    // here because lazy evaluation will prevent CloseHandle(...) from being
    // called if allOK is FALSE - and we *want* CloseHandle(...) called
    // regardless, otherwise we end up with FreeOTFE never closing this file
    // handle, and "leaking" this handle
    if not(CloseHandle(driveDevice)) then
      begin
      allOK := FALSE;
      end;
    end;


  Result := allOK;

end;


// ----------------------------------------------------------------------------// ----------------------------------------------------------------------------
// Send LDREU to get the driver to lock, dismount, remove eject and unlock
// driveFile - Must be of the format "\??\Z:"
// deviceName - Must be of the formst \Device\FreeOTFE\Disks\Disk1
function TOTFEFreeOTFE.LDREUDriver(
                                   driveLetter: char;
                                   deviceName: string;
                                   emergency: boolean
                                 ): boolean;
var
  retVal: boolean;
  DIOCBuffer: TDIOC_LDREU;
  bytesReturned: DWORD;
  kernelDriveFile: string;
begin
  retVal := FALSE;

  CheckActive();

  // This functionality only present in v2.00 of the FreeOTFE driver; previous
  // versions could only carry this out within the userspace app
  if CheckVersionAtLeast(FREEOTFE_ID_v02_00_0000) then
    begin
    kernelDriveFile := '\??\'+uppercase(driveLetter)+':';

    StrPCopy(DIOCBuffer.DriveFile, kernelDriveFile);
    StrPCopy(DIOCBuffer.DiskDeviceName, deviceName);
    DIOCBuffer.Emergency := emergency;

    if (DeviceIoControl(
                        DriverHandle,
                        IOCTL_FREEOTFE_LDREU,
                        @DIOCBuffer,
                        sizeof(DIOCBuffer),
                        nil,
                        0,
                        bytesReturned,
                        nil
                       )) then
      begin
      retVal := TRUE;
      end;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Attempt to destroy a device
function TOTFEFreeOTFE.DestroyDiskDevice(deviceName: string): boolean;
var
  retVal: boolean;
  DIOCBuffer: TDIOC_DEVICE_NAME;
  bytesReturned: DWORD;
begin
  retVal := FALSE;

  CheckActive();

  StrPCopy(DIOCBuffer.DeviceName, deviceName);

  if (DeviceIoControl(
                      DriverHandle,
                      IOCTL_FREEOTFE_DESTROY,
                      @DIOCBuffer,
                      sizeof(DIOCBuffer),
                      nil,
                      0,
                      bytesReturned,
                      nil
                     )) then
    begin
    retVal := TRUE;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Mount(volumeFilename: string; readonly: boolean = FALSE): char;
var
  tmpStringList: TStringList;
  mountedAs: string;
  retVal: char;
begin
  retVal := #0;

  tmpStringList:= TStringList.Create();
  try
    tmpStringList.Add(volumeFilename);
    if Mount(tmpStringList, mountedAs, readonly) then
      begin
      retVal := mountedAs[1];
      end;

  finally
    tmpStringList.Free();
  end;

  Result :=  retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Mount(volumeFilenames: TStringList; var mountedAs: string; readonly: boolean = FALSE): boolean;
var
  frmSelectVolumeType: TfrmSelectVolumeType;
  retVal: boolean;
  mr: integer;
begin
  retVal := FALSE;

  CheckActive();

  frmSelectVolumeType:= TfrmSelectVolumeType.Create(nil);
  try
    mr := frmSelectVolumeType.ShowModal();
    if (mr = mrCancel) then
      begin
      LastErrorCode := OTFE_ERR_USER_CANCEL;
      end
    else if (mr = mrOK) then
      begin
      if frmSelectVolumeType.FreeOTFEVolume then
        begin
        retVal := MountFreeOTFE(volumeFilenames, mountedAs, readonly);
        end
      else
        begin
        retVal := MountLinux(volumeFilenames, mountedAs, readonly);
        end;

      end;

  finally
    frmSelectVolumeType.Free();;
  end;

  Result := retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Dismount(volumeFilename: string; emergency: boolean = FALSE): boolean;
begin
  Result := Dismount(GetDriveForVolFile(volumeFilename), emergency);

end;


// ----------------------------------------------------------------------------
// This dismount routine is based on the MS "media eject" routine for NT/2k/XP
// at:
// http://support.microsoft.com/default.aspx?scid=http://support.microsoft.com:80/support/kb/articles/Q165/7/21.asp&NoWebContent=1
function TOTFEFreeOTFE.Dismount(driveLetter: char; emergency: boolean = FALSE): boolean;
var
  allOK: boolean;
  volumeInfo: TOTFEFreeOTFEVolumeInfo;
  unableToOpenDrive: boolean;
begin
  CheckActive();

  allOK := TRUE;

  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('getVolumeInfo');
{$ENDIF}
    allOK := GetVolumeInfo(driveLetter, volumeInfo);

    // If we couldn't get the drive info, then we can't even do an
    // emergency dismount; we don't know which FreeOTFE device to dismount
    if (not(allOK)) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('getVolumeInfo NOT ALLOK');
{$ENDIF}
      emergency := FALSE;
      end;
    end;



  // We attempt to open the drive, and then carry out various operations to
  // shut the drive down.
  // If were operating in an emergency, then we don't care if some of these
  // operations fail - it's an EMERGENCY! JUST GET RID OF THE %$!# DRIVE!

  if (allOK) then
    begin
    unableToOpenDrive := FALSE;
    allOK := LDREUUserApp(
                          driveLetter,
                          volumeInfo.deviceName,
                          emergency,
                          unableToOpenDrive
                         );
    end;


  // If we were unable to open the drive to carry out the dismount, we're
  // probably on Vista, without having escalated UAC before trying to dismount.
  // Fallback to getting the driver handle the privileged part of the
  // dismount (i.e. opening the volume) and carry out the LDREU
  if (
      not(allOK) and
      unableToOpenDrive
     ) then
    begin
    allOK := LDREUDriver(driveLetter, volumeInfo.deviceName, emergency);
    end;

  if (allOK or emergency) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('remove definition');
{$ENDIF}
    allOK := DeleteDosDeviceSymlink(volumeInfo.deviceName, driveLetter);
    end;


  // And finally... Destroy the device
  if (allOK or emergency) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('destroy device');
{$ENDIF}
    allOK := DestroyDiskDevice(volumeInfo.deviceName);
    end;


  Result := allOK;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Title(): string;
begin
  Result := 'FreeOTFE';
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.Version(): cardinal;
var
  BytesReturned: DWORD;
  retVal: cardinal;
  DIOCBuffer: TDIOC_VERSION;
begin
  retVal := VERSION_ID_FAILURE;

  CheckActive();

  // If we have the version ID cached, use the cached information
  if (fCachedVersionID <> VERSION_ID_NOT_CACHED) then
    begin
    retVal := fCachedVersionID;
    end
  else
    begin
    if (DeviceIoControl(
                        DriverHandle,
                        IOCTL_FREEOTFE_VERSION,
                        nil,
                        0,
                        @DIOCBuffer,
                        sizeof(DIOCBuffer),
                        BytesReturned,
                        nil
                        )) then
      begin
      if BytesReturned=sizeof(DIOCBuffer) then
        begin
        retVal := DIOCBuffer.VersionID;
        fCachedVersionID := retVal;
        end;

      end;

    end;


  Result := retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.VersionStr(): string;
var
  verNo: cardinal;
  retVal: string;
begin
  retval := '';

  CheckActive();

  verNo := Version();
  retVal := VersionIDToStr(verNo);

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Convert a version ID into a prettyprinted version string
function TOTFEFreeOTFE.VersionIDToStr(versionID: DWORD): string;
var
  majorVer: integer;
  minorVer: integer;
  buildVer: integer;
  retval: string;
begin
  retval := '';

  if (versionID <> VERSION_ID_FAILURE) then
    begin
    majorVer := (versionID AND $FF000000) div $00FF0000;
    minorVer := (versionID AND $00FF0000) div $0000FF00;
    buildVer := (versionID AND $0000FFFF);
    retval := Format('v%d.%.2d.%.4d', [majorVer, minorVer, buildVer]);
    end;

  Result := retval;

end;


// ----------------------------------------------------------------------------
// Get the number of disk devices the driver currently has
// Returns $FFFFFFFF on error
function TOTFEFreeOTFE.GetDiskDeviceCount(): cardinal;
var
  BytesReturned: DWORD;
  retVal: cardinal;
  DIOCBuffer: TDIOC_DISK_DEVICE_COUNT;
begin
  retVal := $FFFFFFFF;

  CheckActive();

  if (DeviceIoControl(
                      DriverHandle,
                      IOCTL_FREEOTFE_GET_DISK_DEVICE_COUNT,
                      nil,
                      0,
                      @DIOCBuffer,
                      sizeof(DIOCBuffer),
                      BytesReturned,
                      nil
                      )) then
    begin
    if BytesReturned=sizeof(DIOCBuffer) then
      begin
      retVal := DIOCBuffer.Count;
      end;

    end;


  Result := retVal;

end;


// ----------------------------------------------------------------------------
// Get the kernel mode device name of all disk devices
// WARNING: If devices are mounted while this function is called, this function
//          may fail if the number of devices goes up between the call to
//          GetDiskDevicesCount(), and the DIOC call (the buffer won't be big
//          enough to store all device names)
function TOTFEFreeOTFE.GetDiskDeviceList(deviceNames: TStringList): boolean;
var
  BytesReturned: DWORD;
  retVal: boolean;
  ptrBuffer: Pointer;
  ptrBufferOffset: Pointer;
  bufferSize: cardinal;
  i: integer;
  devName: string;
  deviceCount: integer;
  ptrDIOCDeviceNameList: PDIOC_DEVICE_NAME_LIST;
begin
  retVal := FALSE;

  CheckActive();

  deviceCount := GetDiskDeviceCount();


  bufferSize := sizeof(TDIOC_DEVICE_NAME_LIST) - FREEOTFE_MAX_FILENAME_LENGTH + (FREEOTFE_MAX_FILENAME_LENGTH*deviceCount);
  // Round up to next "DIOC boundry". Although this always results in an
  // oversized buffer, it's guaranteed to be big enough.
  bufferSize := bufferSize + (DIOC_BOUNDRY - (bufferSize mod DIOC_BOUNDRY));
  ptrBuffer := allocmem(bufferSize);

  if (DeviceIoControl(
                      DriverHandle,
                      IOCTL_FREEOTFE_GET_DISK_DEVICE_LIST,
                      nil,
                      0,
                      ptrBuffer,
                      bufferSize,
                      BytesReturned,
                      nil
                     )) then
    begin
    if BytesReturned<=bufferSize then
      begin
      // In case devices were unmounted between the call to
      // GetDiskDeviceCount() and the DIOC call...
      ptrDIOCDeviceNameList := ptrBuffer;
      deviceCount := ptrDIOCDeviceNameList.DeviceCount;

      ptrBufferOffset := Pointer(PChar(ptrBuffer) + sizeof(TDIOC_DEVICE_NAME_LIST) - FREEOTFE_MAX_FILENAME_LENGTH);
      for i:=1 to deviceCount do
        begin
        devName := Copy(PChar(ptrBufferOffset), 1, StrLen(ptrBufferOffset));

        deviceNames.Add(devName);

        // Setup for the next one...
        ptrBufferOffset := Pointer(PChar(ptrBufferOffset) + FREEOTFE_MAX_FILENAME_LENGTH);
        end;

      retval := TRUE;
      end;

    end
  else
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('devicelist DIOC 2 FAIL');
{$ENDIF}
    end;

  FreeMem(ptrBuffer);

  
  Result := retval;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.IsEncryptedVolFile(volumeFilename: string): boolean;
begin
  CheckActive();

  // We can't tell! Return TRUE...
  Result := TRUE;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.DrivesMounted(): string;
var
  retVal: string;
  i: integer;
  driveLetters: TStringList;
  deviceNames: TStringList;
begin
  retVal := '';

  CheckActive();

  driveLetters:= TStringList.Create();
  try
    deviceNames:= TStringList.Create();
    try
      if GetMountedDevices(driveLetters, deviceNames) then
        begin
        for i:=0 to (driveLetters.count-1) do
          begin
          if (driveLetters[i]<>'') then
            begin
            retVal := retVal + driveLetters[i];
            end;

          end;

        retVal := SortString(retVal);
        end;

    finally
      deviceNames.Free();
    end;
  finally
    driveLetters.Free();
  end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetVolFileForDrive(driveLetter: char): string;
var
  retVal: string;
  volumeInfo: TOTFEFreeOTFEVolumeInfo;
begin
  retVal := '';

  CheckActive();

  if (GetVolumeInfo(driveLetter, volumeInfo)) then
    begin
    retVal := volumeInfo.Filename;
    end;

  Result := retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetDriveForVolFile(volumeFilename: string): char;
var
  mounted: string;
  volumeInfo: TOTFEFreeOTFEVolumeInfo;
  retVal: char;
  i: integer;
begin
  retVal := #0;

  CheckActive();

  mounted := DrivesMounted();
  for i:=1 to length(mounted) do
    begin
    if GetVolumeInfo(mounted[i], volumeInfo) then
      begin
      if (uppercase(volumeInfo.Filename) = uppercase(volumeFilename)) then
        begin
        retVal := volumeInfo.DriveLetter;
        break;
        end;

      
      end;
    end;

  Result := retVal;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetMainExe(): string;
begin
  // This is not meaningful for FreeOTFE
  FLastErrCode:= OTFE_ERR_UNABLE_TO_LOCATE_FILE;
  Result := '';

end;


// ----------------------------------------------------------------------------
// Get a list of all FreeOTFE devices which are currently mounted
// Drive letters will be added to "driveLetters", and the corresponding device
// in "deviceNames".
// If the FreeOTFE driver reports that a device is mounted, but there is no
// corresponding drive letter, then the relevant entry in "driveLetter" will
// be an empty string, instead of a drive letter.
// NOTE: THIS FUNCTION *CANNOT* CALL "GetVolumeInfo(...)"!!
// Returns TRUE/FALSE on success/failure
function TOTFEFreeOTFE.GetMountedDevices(driveLetters: TStringList; deviceNames: TStringList): boolean;
const
  // 5 is the arbitary number of times we'll try this until we just give up on
  // it...
  MAX_ATTEMPTS = 5;
var
  retVal: boolean;
  i: integer;
  driveLetter: char;
  driveColon: string;
  foundDriveLetter: boolean;
  attempt: integer;
  flagTryAgain: boolean;
  currDeviceName: string;
  queryResults: TStringList;
begin
  retVal := FALSE;

  CheckActive();

  // flagTryAgain indicates whether the devices have changed during this
  // operation; if they have, it tries it again
  flagTryAgain := TRUE;
  // attempt counts the number of times the this loop was attempted
  attempt := 1;
  while ((flagTryAgain) and (attempt<=MAX_ATTEMPTS)) do
    begin
    flagTryAgain := FALSE;
    driveLetters.Clear();
    deviceNames.Clear();

    if not(GetDiskDeviceList(deviceNames)) then
      begin
      inc(attempt);
      flagTryAgain := TRUE;
      end;

    end;  // while

    

  if (attempt <= MAX_ATTEMPTS) then
    begin
    for i:=0 to (deviceNames.count-1) do
      begin
      currDeviceName := deviceNames[i];

      foundDriveLetter:= FALSE;

      for driveLetter:='A' to 'Z' do
        begin
        driveColon := driveLetter+':';

        queryResults:= TStringList.Create();
        try
          if (DoQueryDosDevice(driveColon, queryResults)) then
            begin
            if (queryResults.count>=1) then
              begin
              if (queryResults[0] = currDeviceName) then
                begin
                foundDriveLetter := TRUE;
                driveLetters.Add(driveLetter);
                break;
                end;
              end;
            end;


        finally
          queryResults.Free();
        end;

        end;

      if (not(foundDriveLetter)) then
        begin
        driveLetters.Add('');
        end;

      end;

    retVal := TRUE;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetVolumeInfo(driveLetter: char; var volumeInfo: TOTFEFreeOTFEVolumeInfo): boolean;
var
  retVal: boolean;
  deviceName: string;
  BytesReturned: DWORD;
  inBuffer: TDIOC_DEVICE_NAME;
  outBuffer: TDIOC_DISK_DEVICE_STATUS;
  tmpSectorIVGenMethod: TFreeOTFESectorIVGenMethod;
begin
  retVal := FALSE;

  CheckActive();

  driveLetter := upcase(driveLetter);

  deviceName := GetDriveDeviceName(driveLetter);
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('deviceName: '+deviceName);
{$ENDIF}
  if (deviceName <> '') then
    begin
    StrPCopy(inBuffer.DeviceName, deviceName);

    if (DeviceIoControl(
                        DriverHandle,
                        IOCTL_FREEOTFE_GET_DISK_DEVICE_STATUS,
                        @inBuffer,
                        sizeof(inBuffer),
                        @outBuffer,
                        sizeof(outBuffer),
                        BytesReturned,
                        nil
                        )) then
      begin
      if (BytesReturned <= sizeof(outBuffer)) then
        begin
        volumeInfo.Filename         := Copy(outBuffer.Filename, 1, StrLen(outBuffer.Filename));
        volumeInfo.DeviceName       := Copy(outBuffer.DiskDeviceName, 1, StrLen(outBuffer.DiskDeviceName));
        volumeInfo.IVHashDevice     := Copy(outBuffer.IVHashDeviceName, 1, StrLen(outBuffer.IVHashDeviceName));
        volumeInfo.IVHashGUID       := outBuffer.IVHashGUID;
        volumeInfo.IVCypherDevice   := Copy(outBuffer.IVCypherDeviceName, 1, StrLen(outBuffer.IVCypherDeviceName));
        volumeInfo.IVCypherGUID     := outBuffer.IVCypherGUID;
        volumeInfo.MainCypherDevice := Copy(outBuffer.MainCypherDeviceName, 1, StrLen(outBuffer.MainCypherDeviceName));
        volumeInfo.MainCypherGUID   := outBuffer.MainCypherGUID;
        volumeInfo.Mounted          := outBuffer.Mounted;
        volumeInfo.ReadOnly         := outBuffer.ReadOnly;
        volumeInfo.VolumeFlags      := outBuffer.VolumeFlags;

        volumeInfo.SectorIVGenMethod:= foivgUnknown;
        for tmpSectorIVGenMethod := low(TFreeOTFESectorIVGenMethod) to high(TFreeOTFESectorIVGenMethod) do
          begin
          if (FreeOTFESectorIVGenMethodID[tmpSectorIVGenMethod] = outBuffer.SectorIVGenMethod) then
            begin
            volumeInfo.SectorIVGenMethod:= tmpSectorIVGenMethod;
            end;
          end;

        volumeInfo.DriveLetter := driveLetter;

        volumeInfo.Filename := GetUserModeVolumeFilename(volumeInfo.Filename);

        retVal := GetVolumeMetaData(driveLetter, outBuffer.MetaDataLength, volumeInfo.MetaData);
        end;

      end;

    end;

  Result := retVal;

end;

// ----------------------------------------------------------------------------
// expectedLength - Set to the amount of metadata to retrieve
function TOTFEFreeOTFE.GetVolumeMetaData(driveLetter: char; expectedLength: integer; var metaData: string): boolean;
var
  retVal: boolean;
  deviceName: string;
  bufferSize: DWORD;
  BytesReturned: DWORD;
  inBuffer: TDIOC_DEVICE_NAME;
  ptrDIOCBuffer: PDIOC_DISK_DEVICE_METADATA;
begin
  retVal := FALSE;

  CheckActive();

  driveLetter := upcase(driveLetter);

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('expectedLength: '+inttostr(expectedLength));
{$ENDIF}

  deviceName := GetDriveDeviceName(driveLetter);
  if (deviceName <> '') then
    begin
    StrPCopy(inBuffer.DeviceName, deviceName);

    bufferSize := sizeof(TDIOC_DISK_DEVICE_METADATA)
                  - sizeof(char) + (sizeof(char)*expectedLength);
    // Round up to next "DIOC boundry". Although this always results in an
    // oversized buffer, it's guaranteed to be big enough.
    bufferSize := bufferSize + (DIOC_BOUNDRY - (bufferSize mod DIOC_BOUNDRY));
    ptrDIOCBuffer := allocmem(bufferSize);
    try
      StrPCopy(inBuffer.DeviceName, deviceName);

      if (DeviceIoControl(
                          DriverHandle,
                          IOCTL_FREEOTFE_GET_DISK_DEVICE_METADATA,
                          @inBuffer,
                          sizeof(inBuffer),
                          ptrDIOCBuffer,
                          bufferSize,
                          BytesReturned,
                          nil
                          )) then
        begin
        if (BytesReturned <= bufferSize) then
          begin
          // Set metaData so that it has enough characters which can be
          // overwritten with StrMove
          metaData := StringOfChar(#0, ptrDIOCBuffer.MetaDataLength);
          StrMove(PChar(metaData), @ptrDIOCBuffer.MetaData, ptrDIOCBuffer.MetaDataLength);

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Metadata length: '+inttostr(ptrDIOCBuffer.MetaDataLength));
DebugMsgBinary(metaData);
{$ENDIF}

          retVal := TRUE;
          end;

        end;

      finally
        FreeMem(ptrDIOCBuffer);
      end;

    end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Determine the FreeOTFE device for the specified drive letter, this may then
// be passed back to the driver
// NOTE: THIS FUNCTION *CANNOT* CALL "GetVolumeInfo(...)"!!
// Returns '' on failure
function TOTFEFreeOTFE.GetDriveDeviceName(driveLetter: char): string;
var
  retVal: string;
  mountedDriveLetters: TStringList;
  deviceNames: TStringList;
  i: integer;
begin
  retVal := '';

  CheckActive();

  mountedDriveLetters:= TStringList.Create();
  try
    deviceNames:= TStringList.Create();
    try
      if GetMountedDevices(mountedDriveLetters, deviceNames) then
        begin
        for i:=0 to (mountedDriveLetters.count-1) do
          begin
          if (mountedDriveLetters[i] = driveLetter) then
            begin
            retVal := deviceNames[i];
            break;
            end;

          end;

        end;

    finally
      deviceNames.Free();
    end;
  finally
    mountedDriveLetters.Free();
  end;

  Result := retVal;

end;


// ----------------------------------------------------------------------------
// Given the specified device name, return the drive mounted on that device
// NOTE: THIS FUNCTION *CANNOT* CALL "GetVolumeInfo(...)"!!
// Returns #0 on failure
function TOTFEFreeOTFE.GetDriveDeviceLetter(deviceName: string): char;
var
  retVal: char;
  mountedDriveLetters: TStringList;
  deviceNames: TStringList;
  i: integer;
begin
  retVal := #0;

  CheckActive();

  mountedDriveLetters:= TStringList.Create();
  try
    deviceNames:= TStringList.Create();
    try
      if GetMountedDevices(mountedDriveLetters, deviceNames) then
        begin
        for i:=0 to (mountedDriveLetters.count-1) do
          begin
          if (deviceNames[i] = deviceName) then
            begin
            retVal := mountedDriveLetters[i][1];
            break;
            end;

          end;

        end;

    finally
      deviceNames.Free();
    end;
  finally
    mountedDriveLetters.Free();
  end;

  Result := retVal;

end;


// ----------------------------------------------------------------------------
// Connects to the named device, returns a device handle, or 0 on error
function TOTFEFreeOTFE.ConnectDevice(devicename: string): THandle;
var
  deviceHandle: THandle;
begin
  deviceHandle := CreateFile(
                             PChar(devicename),
                             GENERIC_READ or GENERIC_WRITE,
                             FILE_SHARE_READ or FILE_SHARE_WRITE,
                             nil,
                             OPEN_EXISTING,
                             FILE_FLAG_NO_BUFFERING,
                             0
                            );

  if (deviceHandle = INVALID_HANDLE_VALUE) then
    begin
    deviceHandle := 0;
    end;

  Result := deviceHandle;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.DisconnectDevice(deviceHandle: THandle): boolean;
begin
  CloseHandle(deviceHandle);
  Result:= TRUE;
end;


// ----------------------------------------------------------------------------
// Create/delete DosDevices symlink; effectivly wraps DefineDosDevice(...) with
// additional support required for Vista changes
// createNotDelete - Set to TRUE to create, FALSE to delete
function TOTFEFreeOTFE.DosDeviceSymlink(
                                        CreateNotDelete: boolean;
                                        DeviceName: string;
                                        DriveLetter: char;
                                        Global: boolean
                                       ): boolean;
var
  retVal: boolean;
  DIOCBuffer: TDIOC_DOS_MOUNTPOINT;
  bytesReturned: DWORD;
  mountpoint: string;
  DIOCCode: DWORD;
  driveLetterColon: string;
begin
  retVal := FALSE;

  CheckActive();

  DriveLetter := upcase(DriveLetter);

  // v2.00 of the FreeOTFE driver use DIOC call to create global/local version
  if not(CheckVersionAtLeast(FREEOTFE_ID_v02_00_0000)) then
    begin
    driveLetterColon := DriveLetter+':';

    if CreateNotDelete then
      begin
      retval := DefineDosDevice(
                                DDD_RAW_TARGET_PATH,
                                PChar(driveLetterColon),
                                PChar(deviceName)
                               );
      end
    else
      begin
      retval := DefineDosDevice(
                                DDD_REMOVE_DEFINITION,
                                PChar(driveLetterColon),
                                nil
                               );
      end;

    end
  else
    begin
    mountpoint := DriveLetter + ':';

    StrPCopy(DIOCBuffer.DiskDeviceName, DeviceName);
    DIOCBuffer.Global := Global;
    StrPCopy(DIOCBuffer.Mountpoint, mountpoint);

    if CreateNotDelete then
      begin
      DIOCCode := IOCTL_FREEOTFE_CREATE_DOS_MOUNTPOINT;
      end
    else
      begin
      DIOCCode := IOCTL_FREEOTFE_DELETE_DOS_MOUNTPOINT;
      end;

    if (DeviceIoControl(
                        DriverHandle,
                        DIOCCode,
                        @DIOCBuffer,
                        sizeof(DIOCBuffer),
                        nil,
                        0,
                        bytesReturned,
                        nil
                       )) then
      begin
      retVal := TRUE;

      BroadcastDriveChangeMessage(CreateNotDelete, DriveLetter);
      end;
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Delete both local and global DosDevices symlinks, as possible
function TOTFEFreeOTFE.DeleteDosDeviceSymlink(DeviceName: string; DriveLetter: char): boolean;
var
  retval: boolean;
begin
  retval := FALSE;

  // Note: We do ***NOT*** use
  //         retval := (
  //                    DosDeviceSymlink(..., TRUE) or
  //                    DosDeviceSymlink(..., FALSE)
  //                   );
  //       here, as lazy evaluation will prevent one or the other from being
  //       evaluated - we want *both* to always be executed!
  if DosDeviceSymlink(
                      FALSE,
                      DeviceName,
                      DriveLetter,
                      TRUE
                     ) then
    begin
    retval := TRUE;
    end;

  if DosDeviceSymlink(
                      FALSE,
                      DeviceName,
                      DriveLetter,
                      FALSE
                     ) then
    begin
    retval := TRUE;
    end;


  Result := retval;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.DoQueryDosDevice(deviceName: string; queryResults: TStringList): boolean;
const
  buffSizeIncrement: integer = 1024;
var
  ptrBuffer: Pointer;
  ptrBufferOffset: Pointer;
  dev: string;
  offset: integer;
  z: integer;
  pcharDeviceName: PChar;
  bufferSize: integer;
  bufferUsed: integer;
  finished: boolean;
  retval: boolean;
begin
  retval := FALSE;

  bufferUsed := 0;

  pcharDeviceName := PChar(deviceName);
  if (deviceName = '') then
    begin
    pcharDeviceName := nil;
    end;

  finished := FALSE;
  bufferSize := buffSizeIncrement;
  ptrBuffer := AllocMem(bufferSize);
  FillChar(ptrBuffer^, bufferSize, 'Z');
  try
    while (not(finished)) do
      begin
      // Cleardown any previous error
      SetLastError($FFFFFFFF);

      // Here, we dereference ptrBuffer to zero the buffer's contents
      bufferUsed := QueryDosDevice(pcharDeviceName, ptrBuffer, bufferSize);

      // We need this "if" statement because Windows 2000 *will* return a partial
      // buffer (set bufferUsed > 0) if the buffer is large enough to store some,
      // but not all, of the information
      // Windows XP will only return bufferUsed > 0 if *all* information could
      // be returned (it seems)
      if (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
        begin
        FreeMem(ptrBuffer);
        bufferSize := bufferSize + 1024; // Increment buffer size by 10K
        ptrBuffer := AllocMem(bufferSize);
        continue;
        end
      else
        begin
        finished := TRUE;
        end;

      end;


    if (bufferUsed > 0) then
      begin
      queryResults.Clear();

      offset := 0;
      while (offset<=bufferUsed) do
        begin
        ptrBufferOffset := PChar(ptrBuffer) + offset;

        z := StrLen(ptrBufferOffset);

        // From the help file on QueryDosDevice:
        //   "The final null-terminated string is followed by an additional NULL."
        // i.e. if z=0, then we've done the last one; quit the loop
        if (z = 0) then
          begin
          break;
          end;

        dev := Copy(PChar(ptrBufferOffset), 1, z);
        queryResults.add(dev);

        offset := offset + length(dev) + 1;
        end;

      retval := TRUE;
      end;

  finally
    FreeMem(ptrBuffer);
  end;


  Result := retval;
end;


// ----------------------------------------------------------------------------
// Given an MSDOS device name, return the device name of that device
// Returns '' on error
function TOTFEFreeOTFE.GetDeviceName(MSDOSKernelModeDeviceName: string): string;
var
  deviceNames: TStringList;
  deviceNameMapping: TStringList;
  i, j: integer;
  retval: string;
begin
  retval := '';

  deviceNames:= TStringList.Create();
  try
    if DoQueryDosDevice('', deviceNames) then
      begin
      for i:=0 to (deviceNames.count-1) do
        begin
        deviceNameMapping:= TStringList.Create();
        try

          if DoQueryDosDevice(deviceNames[i], deviceNameMapping) then
            begin
            for j:=0 to (deviceNameMapping.count-1) do
              begin
              if (deviceNameMapping[j] = MSDOSKernelModeDeviceName) then
                begin
                retval := deviceNames[i];
                break;
                end;
              end;
              
            // If we've found the name, exit the outer loop as well
            if (retval<>'') then
              begin
              break;
              end;

            end;

        finally
          deviceNameMapping.Free();
        end;

        end;

      end;

  finally
    deviceNames.Free();
  end;

  Result := retval;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetCypherDrivers(var cypherDrivers: TFreeOTFECypherDriverArray): boolean;
var
  retVal: boolean;
  deviceNames: TStringList;
  deviceNameMapping: TStringList;
  i, j: integer;
begin
  retVal := TRUE;

  CheckActive();

  SetLength(cypherDrivers, 0);

  deviceNames:= TStringList.Create();
  try
    if DoQueryDosDevice('', deviceNames) then
      begin
      for i:=0 to (deviceNames.count-1) do
        begin
        deviceNameMapping:= TStringList.Create();
        try
          if DoQueryDosDevice(deviceNames[i], deviceNameMapping) then
            begin
            for j:=0 to (deviceNameMapping.count-1) do
              begin
              if pos(DEVICE_CYPHER_DIR_NAME, deviceNameMapping[j])>0 then
                begin
                // Get the details for the device
                SetLength(cypherDrivers, (Length(cypherDrivers)+1));
                retval := GetCypherDriverCyphers(deviceNameMapping[j], cypherDrivers[high(cypherDrivers)]);

                // If we got the details, bang on to the next device
                // Note: If we couldn't get the details, this will have the
                //       effect of allowing the loop to go round again, but
                //       this time retval will still be FALSE; if the loop
                //       eventually exits, then retval will *still* be false,
                //       signalling an error
                if (retval) then
                  begin
                  break;
                  end;
                end;

              end;  // for j:=1 to (deviceNameMapping.count-1) do

            end;

        finally
          deviceNameMapping.Free();
        end;

        // Propogate breakout, if retval is FALSE
        if (not(retval)) then
          begin
          break;
          end;

        end;  // for i:=1 to (deviceNames.count-1) do

      end;

  finally
    deviceNames.Free();
  end;

  Result := retval;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetCypherDriverCyphers(cypherKernelModeDeviceName: string; var cypherDriverDetails: TFreeOTFECypherDriver): boolean;
var
  cypherHandle: THandle;
  BytesReturned: DWORD;
  retval: boolean;
  DIOCBuffer: TDIOC_CYPHER_IDENTIFYDRIVER;
  ptrBuffer: Pointer;
  ptrBufferOffset: Pointer;
{$IFDEF FREEOTFE_DEBUG}
DIOCBufferCyphers: PDIOC_CYPHER_IDENTIFYSUPPORTED;
{$ENDIF}
  bufferSize: cardinal;
  cypherDetails: PCYPHER;
  i: integer;
  currCypherMode: TFreeOTFECypherMode;
  arrIdx: integer;
  deviceUserModeName: string;
begin
  retval := FALSE;

  CheckActive();

  if CachesGetCypherDriver(cypherKernelModeDeviceName, cypherDriverDetails) then
    begin
    retVal := TRUE;
    end
  else
    begin
    // Determine the user mode MSDOS device name from the kernel mode device name
    deviceUserModeName := GetCypherDeviceUserModeDeviceName(cypherKernelModeDeviceName);


{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Connecting to: '+deviceUserModeName);
{$ENDIF}
    cypherHandle := ConnectDevice(deviceUserModeName);
    if (cypherHandle = 0) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Couldn''t connect to: '+deviceUserModeName);
{$ENDIF}
      end
    else
      begin
      if (DeviceIoControl(
                          cypherHandle,
                          IOCTL_FREEOTFECYPHER_IDENTIFYDRIVER,
                          nil,
                          0,
                          @DIOCBuffer,
                          sizeof(DIOCBuffer),
                          BytesReturned,
                          nil
                         )) then
        begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('_1_ supplied: '+inttostr(sizeof(DIOCBuffer))+' used: '+inttostr(BytesReturned));
  {$ENDIF}
        if BytesReturned=sizeof(DIOCBuffer) then
          begin
          cypherDriverDetails.DriverGUID  := DIOCBuffer.DriverGUID;
          cypherDriverDetails.DeviceName  := GetDeviceName(cypherKernelModeDeviceName);
          cypherDriverDetails.DeviceKernelModeName := cypherKernelModeDeviceName;
          cypherDriverDetails.DeviceUserModeName   := DeviceUserModeName;
          cypherDriverDetails.Title       := Copy(DIOCBuffer.Title, 1, StrLen(DIOCBuffer.Title));
          cypherDriverDetails.VersionID   := DIOCBuffer.VersionID;
          cypherDriverDetails.CypherCount := DIOCBuffer.CypherCount;


          bufferSize := sizeof(TDIOC_CYPHER_IDENTIFYSUPPORTED) - sizeof(TCYPHER) + (sizeof(TCYPHER)*cypherDriverDetails.CypherCount);
          // Round up to next "DIOC boundry". Although this always results in an
          // oversized buffer, it's guaranteed to be big enough.
          bufferSize := bufferSize + (DIOC_BOUNDRY - (bufferSize mod DIOC_BOUNDRY));
          ptrBuffer := allocmem(bufferSize);

          SetLength(cypherDriverDetails.Cyphers, cypherDriverDetails.CypherCount);

          if (DeviceIoControl(
                              cypherHandle,
                              IOCTL_FREEOTFECYPHER_IDENTIFYSUPPORTED,
                              nil,
                              0,
                              ptrBuffer,
                              bufferSize,
                              BytesReturned,
                              nil
                             )) then
            begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('_2_ supplied: '+inttostr(bufferSize)+' used: '+inttostr(BytesReturned));
  {$ENDIF}
            if BytesReturned<=bufferSize then
              begin
  {$IFDEF FREEOTFE_DEBUG}
  DIOCBufferCyphers := (PDIOC_CYPHER_IDENTIFYSUPPORTED(ptrBuffer));
  DebugMsg('cypher details count: '+inttostr(DIOCBufferCyphers.BufCount));
  {$ENDIF}
              ptrBufferOffset := Pointer(PChar(ptrBuffer) + sizeof(TDIOC_CYPHER_IDENTIFYSUPPORTED) - sizeof(TCYPHER));
              arrIdx := low(cypherDriverDetails.Cyphers);
              for i:=1 to cypherDriverDetails.CypherCount do
                begin
                cypherDetails := (PCYPHER(ptrBufferOffset));
                cypherDriverDetails.Cyphers[arrIdx].CypherGUID := cypherDetails.CypherGUID;
                cypherDriverDetails.Cyphers[arrIdx].Title := Copy(cypherDetails.Title, 1, StrLen(cypherDetails.Title));

                // Failsafe to "unknown"
                cypherDriverDetails.Cyphers[arrIdx].Mode := focmUnknown;
                for currCypherMode:=low(FreeOTFECypherModeID) to high(FreeOTFECypherModeID) do
                  begin
                  if (cypherDetails.Mode = FreeOTFECypherModeID[currCypherMode]) then
                    begin
                    cypherDriverDetails.Cyphers[arrIdx].Mode := currCypherMode;
                    end;

                  end;

                cypherDriverDetails.Cyphers[arrIdx].KeySize := cypherDetails.KeySize;
                cypherDriverDetails.Cyphers[arrIdx].BlockSize := cypherDetails.BlockSize;
                cypherDriverDetails.Cyphers[arrIdx].VersionID := cypherDetails.VersionID;

                // Setup for the next one...
                ptrBufferOffset := Pointer(PChar(ptrBufferOffset) + sizeof(TCYPHER));
                inc(arrIdx);
                end;

              // Cache the information retrieved
              CachesAddCypherDriver(cypherKernelModeDeviceName, cypherDriverDetails);

              retval := TRUE;
              end;

            end
          else
            begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('getcypherdrivers DIOC 2 FAIL');
  {$ENDIF}
            end;

          FreeMem(ptrBuffer);
          end;

        end;

      DisconnectDevice(cypherHandle);
      end;

    end;  // if CachesGetCypherDriver(cypherKernelModeDeviceName, cypherDriverDetails) then


{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Exiting function...');
{$ENDIF}
  Result := retval;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetHashDrivers(var hashDrivers: TFreeOTFEHashDriverArray): boolean;
var
  retVal: boolean;
  deviceNames: TStringList;
  deviceNameMapping: TStringList;
  i, j: integer;
begin
  retVal := TRUE;

  CheckActive();

  SetLength(hashDrivers, 0);

  deviceNames:= TStringList.Create();
  try
    if DoQueryDosDevice('', deviceNames) then
      begin
      for i:=0 to (deviceNames.count-1) do
        begin
        deviceNameMapping:= TStringList.Create();
        try
          if DoQueryDosDevice(deviceNames[i], deviceNameMapping) then
            begin
            for j:=0 to (deviceNameMapping.count-1) do
              begin
              if pos(DEVICE_HASH_DIR_NAME, deviceNameMapping[j])>0 then
                begin
                // Get the details for the device
                SetLength(hashDrivers, (Length(hashDrivers)+1));
                retval := GetHashDriverHashes(deviceNameMapping[j], hashDrivers[high(hashDrivers)]);

                // If we got the details, bang on to the next device
                // Note: If we couldn't get the details, this will have the
                //       effect of allowing the loop to go round again, but
                //       this time retval will still be FALSE; if the loop
                //       eventually exits, then retval will *still* be false,
                //       signalling an error
                if (retval) then
                  begin
                  break;
                  end;
                end;

              end;  // for j:=1 to (deviceNameMapping.count-1) do

            end;

        finally
          deviceNameMapping.Free();
        end;

        // Propogate breakout, if retval is FALSE
        if (not(retval)) then
          begin
          break;
          end;

        end;  // for i:=1 to (deviceNames.count-1) do

      end;

  finally
    deviceNames.Free();
  end;

  Result := retval;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetHashDriverHashes(hashKernelModeDeviceName: string; var hashDriverDetails: TFreeOTFEHashDriver): boolean;
var
  hashHandle: THandle;
  BytesReturned: DWORD;
  retval: boolean;
  DIOCBuffer: TDIOC_HASH_IDENTIFYDRIVER;
  ptrBuffer: Pointer;
  ptrBufferOffset: Pointer;
{$IFDEF FREEOTFE_DEBUG}
  DIOCBufferHashes: PDIOC_HASH_IDENTIFYSUPPORTED;
{$ENDIF}
  bufferSize: cardinal;
  hashDetails: PHASH;
  i: integer;
  arrIdx: integer;
  deviceUserModeName: string;
begin
  retval := FALSE;

  CheckActive();


  if CachesGetHashDriver(hashKernelModeDeviceName, hashDriverDetails) then
    begin
    retVal := TRUE;
    end
  else
    begin
    // Determine the user mode MSDOS device name from the kernel mode device name
    deviceUserModeName := GetHashDeviceUserModeDeviceName(hashKernelModeDeviceName);


{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Connecting to: '+deviceUserModeName);
{$ENDIF}
    hashHandle := ConnectDevice(deviceUserModeName);
    if (hashHandle = 0) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Couldn''t connect to: '+deviceUserModeName);
{$ENDIF}
      end
    else
      begin
      if (DeviceIoControl(
                          hashHandle,
                          IOCTL_FREEOTFEHASH_IDENTIFYDRIVER,
                          nil,
                          0,
                          @DIOCBuffer,
                          sizeof(DIOCBuffer),
                          BytesReturned,
                          nil
                         )) then
        begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('_3_ supplied: '+inttostr(sizeof(DIOCBuffer))+' used: '+inttostr(BytesReturned));
{$ENDIF}
        if BytesReturned=sizeof(DIOCBuffer) then
          begin
          hashDriverDetails.DriverGUID := DIOCBuffer.DriverGUID;
          hashDriverDetails.DeviceName := GetDeviceName(hashKernelModeDeviceName);
          hashDriverDetails.DeviceKernelModeName := hashKernelModeDeviceName;
          hashDriverDetails.DeviceUserModeName := DeviceUserModeName;
          hashDriverDetails.Title := Copy(DIOCBuffer.Title, 1, StrLen(DIOCBuffer.Title));
          hashDriverDetails.VersionID := DIOCBuffer.VersionID;
          hashDriverDetails.HashCount := DIOCBuffer.HashCount;


          bufferSize := sizeof(TDIOC_HASH_IDENTIFYSUPPORTED) - sizeof(THASH) + (sizeof(THASH)*hashDriverDetails.HashCount);
          // Round up to next "DIOC boundry". Although this always results in an
          // oversized buffer, it's guaranteed to be big enough.
          bufferSize := bufferSize + (DIOC_BOUNDRY - (bufferSize mod DIOC_BOUNDRY));
          ptrBuffer := allocmem(bufferSize);

          SetLength(hashDriverDetails.Hashes, hashDriverDetails.HashCount);

          if (DeviceIoControl(
                              hashHandle,
                              IOCTL_FREEOTFEHASH_IDENTIFYSUPPORTED,
                              nil,
                              0,
                              ptrBuffer,
                              bufferSize,
                              BytesReturned,
                              nil
                             )) then
            begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('_4_ supplied: '+inttostr(bufferSize)+' used: '+inttostr(BytesReturned));
{$ENDIF}
            if (BytesReturned<=bufferSize) then
              begin
{$IFDEF FREEOTFE_DEBUG}
DIOCBufferHashes := (PDIOC_HASH_IDENTIFYSUPPORTED(ptrBuffer));
DebugMsg('hash details count: '+inttostr(DIOCBufferHashes.BufCount));
{$ENDIF}
              ptrBufferOffset := Pointer(PChar(ptrBuffer) + sizeof(TDIOC_HASH_IDENTIFYSUPPORTED) - sizeof(THASH));
              arrIdx := low(hashDriverDetails.Hashes);
              for i:=1 to hashDriverDetails.HashCount do
                begin
                hashDetails := (PHASH(ptrBufferOffset));
                hashDriverDetails.Hashes[arrIdx].HashGUID  := hashDetails.HashGUID;
                hashDriverDetails.Hashes[arrIdx].Title     := Copy(hashDetails.Title, 1, StrLen(hashDetails.Title));
                hashDriverDetails.Hashes[arrIdx].VersionID := hashDetails.VersionID;
                hashDriverDetails.Hashes[arrIdx].Length    := hashDetails.Length;
                hashDriverDetails.Hashes[arrIdx].BlockSize := hashDetails.BlockSize;

                // Setup for the next one...
                ptrBufferOffset := Pointer(PChar(ptrBufferOffset) + sizeof(THASH));
                inc(arrIdx);
                end;

              // Cache the information retrieved
              CachesAddHashDriver(hashKernelModeDeviceName, hashDriverDetails);
              
              retval := TRUE;
              end;

            end
          else
            begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('gethashdrivers DIOC 2 FAIL');
{$ENDIF}
            end;

          FreeMem(ptrBuffer);
          end;

        end;

      DisconnectDevice(hashHandle);
      end;

    end;  // if CachesGetHashDriver(hashKernelModeDeviceName, hashDriverDetails) then


  Result := retval;

end;


// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.ShowHashDetailsDlg(driverName: string; hashGUID: TGUID);
var
  detailsDlg: TfrmHashInfo;
begin
  detailsDlg:= TfrmHashInfo.Create(nil);
  try
    detailsDlg.OTFEFreeOTFEObj := self;
    detailsDlg.ShowDriverName:= driverName;
    detailsDlg.ShowGUID:= hashGUID;

    detailsDlg.ShowModal();
  finally
    detailsDlg.Free();
  end;

end;


// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.ShowCypherDetailsDlg(driverName: string; cypherGUID: TGUID);
var
  detailsDlg: TfrmCypherInfo;
begin
  detailsDlg:= TfrmCypherInfo.Create(nil);
  try
    detailsDlg.OTFEFreeOTFEObj := self;
    detailsDlg.ShowDriverName:= driverName;
    detailsDlg.ShowGUID:= cypherGUID;

    detailsDlg.ShowModal();
  finally
    detailsDlg.Free();
  end;

end;


// ----------------------------------------------------------------------------
// Convert kernel mode device name to user mode device name
function TOTFEFreeOTFE.GetHashDeviceUserModeDeviceName(hashKernelModeDeviceName: string): string;
var
  deviceUserModeName: string;
begin
  // Determine the user mode MSDOS device name from the kernel mode device name
  deviceUserModeName := hashKernelModeDeviceName;
  // Lose the prefix
  delete(deviceUserModeName, 1, length(DEVICE_HASH_DIR_NAME));
  // Trim off the leading "\"
  delete(deviceUserModeName, 1, 1);
  // Tack on the dosdevice root
  deviceUserModeName := DEVICE_HASH_SYMLINK_PREFIX + deviceUserModeName;

  Result := deviceUserModeName;
end;


// ----------------------------------------------------------------------------
// Convert kernel mode device name to user mode device name
function TOTFEFreeOTFE.GetCypherDeviceUserModeDeviceName(cypherKernelModeDeviceName: string): string;
var
  deviceUserModeName: string;
begin
  // Determine the user mode MSDOS device name from the kernel mode device name
  deviceUserModeName := cypherKernelModeDeviceName;
  // Lose the prefix
  delete(deviceUserModeName, 1, length(DEVICE_CYPHER_DIR_NAME));
  // Trim off the leading "\"
  delete(deviceUserModeName, 1, 1);
  // Tack on the dosdevice root
  deviceUserModeName := DEVICE_CYPHER_SYMLINK_PREFIX + deviceUserModeName;

  Result := deviceUserModeName;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetSpecificHashDetails(hashKernelModeDeviceName: string; hashGUID: TGUID; var hashDetails: TFreeOTFEHash): boolean;
var
  i: integer;
  retval: boolean;
  hashDriverDetails: TFreeOTFEHashDriver;
begin
  retval := FALSE;

  if GetHashDriverHashes(hashKernelModeDeviceName, hashDriverDetails) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('low hashDriverDetails.Hashes: '+inttostr(low(hashDriverDetails.Hashes)));
DebugMsg('high hashDriverDetails.Hashes: '+inttostr(high(hashDriverDetails.Hashes)));
{$ENDIF}
    for i:=low(hashDriverDetails.Hashes) to high(hashDriverDetails.Hashes) do
      begin
      if (IsEqualGUID(hashDriverDetails.Hashes[i].HashGUID, hashGUID)) then
        begin
        hashDetails := hashDriverDetails.Hashes[i];
        retval := TRUE;
        break;
        end;

      end;

    end;

  Result := retval;

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetSpecificCypherDetails(cypherKernelModeDeviceName: string; cypherGUID: TGUID; var cypherDetails: TFreeOTFECypher): boolean;
var
  i: integer;
  retval: boolean;
  cypherDriverDetails: TFreeOTFECypherDriver;
begin
  retval := FALSE;

  if GetCypherDriverCyphers(cypherKernelModeDeviceName, cypherDriverDetails) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('low cypherDriverDetails.Cyphers: '+inttostr(low(cypherDriverDetails.Cyphers)));
DebugMsg('high cypherDriverDetails.Cyphers: '+inttostr(high(cypherDriverDetails.Cyphers)));
{$ENDIF}
    for i:=low(cypherDriverDetails.Cyphers) to high(cypherDriverDetails.Cyphers) do
      begin
      if (IsEqualGUID(cypherDriverDetails.Cyphers[i].CypherGUID, cypherGUID)) then
        begin
        cypherDetails := cypherDriverDetails.Cyphers[i];
        retval := TRUE;
        break;
        end;

      end;

    end;

  Result := retval;

end;


// ----------------------------------------------------------------------------
// Hash the specified data with the supplied hash device/hash GUID
function TOTFEFreeOTFE.HashData(
                      hashKernelModeDeviceName: string;
                      hashGUID: TGUID;
                      var data: string;
                      var hashOut: string
                     ): boolean;
var
  hashHandle: THandle;
  BytesReturned: DWORD;
  retval: boolean;
  ptrDIOCBufferIn: PDIOC_HASH_DATA_IN;
  bufferSizeIn: DWORD;
  ptrDIOCBufferOut: PDIOC_HASH_DATA_OUT;
  bufferSizeOut: DWORD;
  hashDetails: TFreeOTFEHash;
  deviceUserModeName: string;
  hashByteCount: integer;
  expectedHashSizeBits: integer;  // In *bits*
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retval := FALSE;

  CheckActive();

  // Determine the user mode MSDOS device name from the kernel mode device name
  deviceUserModeName := GetHashDeviceUserModeDeviceName(hashKernelModeDeviceName);

  if GetSpecificHashDetails(hashKernelModeDeviceName, hashGUID, hashDetails) then
    begin
    hashHandle := ConnectDevice(deviceUserModeName);
    if (hashHandle = 0) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('couldn''t connect to: '+deviceUserModeName);
{$ENDIF}
      LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
      end
    else
      begin
      bufferSizeIn := sizeof(TDIOC_HASH_DATA_IN) - sizeof(byte) + Length(Data);
      // Round up to next "DIOC boundry". Although this always results in an
      // oversized buffer, it's guaranteed to be big enough.
      bufferSizeIn := bufferSizeIn + (DIOC_BOUNDRY - (bufferSizeIn mod DIOC_BOUNDRY));
      ptrDIOCBufferIn := allocmem(bufferSizeIn);
      try
        // Just make sure its big enough
        expectedHashSizeBits := hashDetails.Length;
        if (hashDetails.Length = -1) then
          begin
          expectedHashSizeBits := (Length(Data) * 8);
          end;
        bufferSizeOut := sizeof(TDIOC_HASH_DATA_OUT) - sizeof(byte) + (expectedHashSizeBits div 8);
        ptrDIOCBufferOut := allocmem(bufferSizeOut);
        try
          ptrDIOCBufferIn.HashGUID := hashGUID;

          ptrDIOCBufferIn.DataLength := Length(Data) * 8;
          StrMove(@ptrDIOCBufferIn.Data, PChar(Data), Length(Data));

          if (DeviceIoControl(
                              hashHandle,
                              IOCTL_FREEOTFEHASH_HASHDATA,
                              ptrDIOCBufferIn,
                              bufferSizeIn,
                              ptrDIOCBufferOut,
                              bufferSizeOut,
                              BytesReturned,
                              nil
                             )) then
            begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('_5_ supplied: '+inttostr(bufferSizeOut)+' used: '+inttostr(BytesReturned));
{$ENDIF}
            if (BytesReturned <= bufferSizeOut) then
              begin
              hashByteCount := (ptrDIOCBufferOut.HashLength div 8);
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('hashByteCount: '+inttostr(hashByteCount));
{$ENDIF}

              // Set hashOut so that it has enough characters which can be
              // overwritten with StrMove
              hashOut := StringOfChar(#0, hashByteCount);
              StrMove(PChar(hashOut), @ptrDIOCBufferOut.Hash, hashByteCount);

              retval := TRUE;
              end
            else
              begin
              LastErrorCode := OTFE_ERR_HASH_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('incorrect bytecount returned');
{$ENDIF}
              end;

            end
          else
            begin
            LastErrorCode := OTFE_ERR_HASH_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('hash DIOC 2 FAIL');
{$ENDIF}
            end;

        finally
          FillChar(ptrDIOCBufferOut^, bufferSizeOut, 0);
          FreeMem(ptrDIOCBufferOut);
        end;

      finally
        FillChar(ptrDIOCBufferIn^, bufferSizeIn, 0);
        FreeMem(ptrDIOCBufferIn);
      end;

      DisconnectDevice(hashHandle);
      end;

    end
  else
    begin
    LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Unable to GetSpecificHashDetails');
{$ENDIF}
    end;  // ELSE PART - if GetSpecificHashDetails(hashKernelModeDeviceName, hashGUID, hashDetails) then

  Result := retval;

end;


// ----------------------------------------------------------------------------
// Generate MAC of data using hash driver/encryption driver identified
// Note: "tBits" is in *bits* not *bytes*
// tBits - Set the the number of bits to return; set to -ve value for no
//         truncation/padding
//         If tBits is larger than the number of bits the MAC would normally
//         be, then the MAC will be right-padded with NULLs
//         If tBits is set to a -ve number, then this function will return
//         an MAC of variable length
function TOTFEFreeOTFE.MACData(
                      macAlgorithm: TFreeOTFEMACAlgorithm;
                      hashKernelModeDeviceName: string;
                      hashGUID: TGUID;
                      cypherKernelModeDeviceName: string;
                      cypherGUID: TGUID;
                      var key: string;
                      var data: string;
                      var MACOut: string;
                      tBits: integer = -1
                     ): boolean;
var
  BytesReturned: DWORD;
  retval: boolean;
  ptrDIOCBufferIn: PDIOC_GENERATE_MAC_IN;
  bufferSizeIn: DWORD;
  ptrDIOCBufferOut: PDIOC_GENERATE_MAC_OUT;
  bufferSizeOut: DWORD;
  outputByteCount: integer;
  tmpSizeBytes: integer;
  minBufferSizeOut: integer;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retval := FALSE;

  CheckActive();

  bufferSizeIn := sizeof(TDIOC_GENERATE_MAC_IN) -
                  sizeof(ptrDIOCBufferIn^.Key) +
                  Length(key) -
                  sizeof(ptrDIOCBufferIn^.Data) +
                  Length(data);
  // Round up to next "DIOC boundry". Although this always results in an
  // oversized buffer, it's guaranteed to be big enough.
  bufferSizeIn := bufferSizeIn + (DIOC_BOUNDRY - (bufferSizeIn mod DIOC_BOUNDRY));
  ptrDIOCBufferIn := allocmem(bufferSizeIn);
  try
    // We use the maximum buffer possible, as the full DK length depends on
    // the algorithm, hash, etc
    tmpSizeBytes := (max(MAX_MAC_LENGTH, tBits) div 8);  // Convert to bytes


    minBufferSizeOut := sizeof(TDIOC_HASH_DATA_OUT) - sizeof(ptrDIOCBufferOut^.MAC);
    bufferSizeOut := minBufferSizeOut + tmpSizeBytes;
    // Round up to next "DIOC boundry". Although this always results in an
    // oversized buffer, it's guaranteed to be big enough.
    bufferSizeOut := bufferSizeOut + (DIOC_BOUNDRY - (bufferSizeOut mod DIOC_BOUNDRY));
    ptrDIOCBufferOut := allocmem(bufferSizeOut);
    try
      ptrDIOCBufferIn.MACAlgorithm := FreeOTFEMACID[macAlgorithm];

      StrPCopy(ptrDIOCBufferIn.HashDeviceName, hashKernelModeDeviceName);
      ptrDIOCBufferIn.HashGUID := hashGUID;

      StrPCopy(ptrDIOCBufferIn.CypherDeviceName, cypherKernelModeDeviceName);
      ptrDIOCBufferIn.CypherGUID := cypherGUID;

      ptrDIOCBufferIn.LengthWanted := tBits;

      ptrDIOCBufferIn.KeyLength := Length(key) * 8;
      ptrDIOCBufferIn.DataLength := Length(data) * 8;
      // This may seem a little weird, but we do this because the data is
      // immediatly after the key
      StrMove(@ptrDIOCBufferIn.Key, PChar(key), Length(key));
      StrMove(((PChar(@ptrDIOCBufferIn.Key))+length(key)), PChar(data), Length(data));


      if (DeviceIoControl(
                          DriverHandle,
                          IOCTL_FREEOTFE_GENERATE_MAC,
                          ptrDIOCBufferIn,
                          bufferSizeIn,
                          ptrDIOCBufferOut,
                          bufferSizeOut,
                          BytesReturned,
                          nil
                         )) then
        begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('MAC out buffer supplied: '+inttostr(bufferSizeOut)+'; DIOC call used: '+inttostr(BytesReturned)+'; min: '+inttostr(minBufferSizeOut));
{$ENDIF}
        if (
            (BytesReturned >= DWORD(minBufferSizeOut)) AND
            (BytesReturned <= DWORD(bufferSizeOut))
           ) then
          begin
          outputByteCount := (ptrDIOCBufferOut.MACLength div 8);
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('outputByteCount: '+inttostr(outputByteCount));
{$ENDIF}

          // Set MACOut so that it has enough characters which can be
          // overwritten with StrMove
          MACOut := StringOfChar(#0, outputByteCount);
          StrMove(PChar(MACOut), @ptrDIOCBufferOut.MAC, outputByteCount);

          retval := TRUE;
          end
        else
          begin
          LastErrorCode := OTFE_ERR_MAC_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('incorrect bytecount returned');
{$ENDIF}
          end;

        end
      else
        begin
        LastErrorCode := OTFE_ERR_MAC_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('MAC DIOC 2 FAIL');
{$ENDIF}
        end;

    finally
      FillChar(ptrDIOCBufferOut^, bufferSizeOut, 0);
      FreeMem(ptrDIOCBufferOut);
    end;

  finally
    FillChar(ptrDIOCBufferIn^, bufferSizeIn, 0);
    FreeMem(ptrDIOCBufferIn);
  end;


  Result := retval;

end;


// ----------------------------------------------------------------------------
// dkLenBits - This must *always* be >= 0
function TOTFEFreeOTFE.DeriveKey(
                    kdfAlgorithm: TFreeOTFEKDFAlgorithm;
                    hashKernelModeDeviceName: string;
                    hashGUID: TGUID;
                    cypherKernelModeDeviceName: string;
                    cypherGUID: TGUID;
                    Password: string;
                    Salt: string;
                    Iterations: integer;
                    dkLenBits: integer;  // In *bits*
                    var DK: string
                   ): boolean;
var
  BytesReturned: DWORD;
  retval: boolean;
  ptrDIOCBufferIn: PDIOC_DERIVE_KEY_IN;
  bufferSizeIn: DWORD;
  ptrDIOCBufferOut: PDIOC_DERIVE_KEY_OUT;
  bufferSizeOut: DWORD;
  outputByteCount: integer;
  tmpSizeBytes: integer;
  minBufferSizeOut: integer;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retval := FALSE;

  CheckActive();


  bufferSizeIn := sizeof(TDIOC_DERIVE_KEY_IN) -
                  sizeof(ptrDIOCBufferIn^.Password) +
                  Length(Password) -
                  sizeof(ptrDIOCBufferIn^.Salt) +
                  Length(Salt);
  // Round up to next "DIOC boundry". Although this always results in an
  // oversized buffer, it's guaranteed to be big enough.
  bufferSizeIn := bufferSizeIn + (DIOC_BOUNDRY - (bufferSizeIn mod DIOC_BOUNDRY));
  ptrDIOCBufferIn := allocmem(bufferSizeIn);
  try
    // We use the maximum buffer possible, as the full MAC length depends on
    // the MAC, hash, etc
    tmpSizeBytes := (max(MAX_DERIVED_KEY_LENGTH, dkLenBits) div 8);  // Convert to bytes

    minBufferSizeOut := sizeof(TDIOC_HASH_DATA_OUT) - sizeof(ptrDIOCBufferOut^.DerivedKey);
    bufferSizeOut := minBufferSizeOut + tmpSizeBytes;
    // Round up to next "DIOC boundry". Although this always results in an
    // oversized buffer, it's guaranteed to be big enough.
    bufferSizeOut := bufferSizeOut + (DIOC_BOUNDRY - (bufferSizeOut mod DIOC_BOUNDRY));
    ptrDIOCBufferOut := allocmem(bufferSizeOut);
    try
     ptrDIOCBufferIn.KDFAlgorithm := FreeOTFEKDFID[kdfAlgorithm];

      StrPCopy(ptrDIOCBufferIn.HashDeviceName, hashKernelModeDeviceName);
      ptrDIOCBufferIn.HashGUID := hashGUID;

      StrPCopy(ptrDIOCBufferIn.CypherDeviceName, cypherKernelModeDeviceName);
      ptrDIOCBufferIn.CypherGUID := cypherGUID;

      ptrDIOCBufferIn.Iterations   := Iterations;
      ptrDIOCBufferIn.LengthWanted := dkLenBits;

      ptrDIOCBufferIn.PasswordLength := (Length(Password) * 8);
      ptrDIOCBufferIn.SaltLength     := (Length(Salt) * 8);
      // This may seem a little weird, but we do this because the salt is
      // immediatly after the password
      StrMove(@ptrDIOCBufferIn.Password, PChar(Password), Length(Password));
      StrMove(((PChar(@ptrDIOCBufferIn.Password))+length(Password)), PChar(Salt), Length(Salt));


      if (DeviceIoControl(
                          DriverHandle,
                          IOCTL_FREEOTFE_DERIVE_KEY,
                          ptrDIOCBufferIn,
                          bufferSizeIn,
                          ptrDIOCBufferOut,
                          bufferSizeOut,
                          BytesReturned,
                          nil
                         )) then
        begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('MAC out buffer supplied: '+inttostr(bufferSizeOut)+'; DIOC call used: '+inttostr(BytesReturned)+'; min: '+inttostr(minBufferSizeOut));
{$ENDIF}
        if (
            (BytesReturned >= DWORD(minBufferSizeOut)) AND
            (BytesReturned <= DWORD(bufferSizeOut))
           ) then
          begin
          outputByteCount := (ptrDIOCBufferOut.DerivedKeyLength div 8);
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('outputByteCount: '+inttostr(outputByteCount));
{$ENDIF}

          // Set MACOut so that it has enough characters which can be
          // overwritten with StrMove
          DK := StringOfChar(#0, outputByteCount);
          StrMove(PChar(DK), @ptrDIOCBufferOut.DerivedKey, outputByteCount);

          retval := TRUE;
          end
        else
          begin
          LastErrorCode := OTFE_ERR_KDF_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('incorrect bytecount returned');
{$ENDIF}
          end;

        end
      else
        begin
        LastErrorCode := OTFE_ERR_KDF_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('MAC DIOC 2 FAIL');
{$ENDIF}
        end;

    finally
      FillChar(ptrDIOCBufferOut^, bufferSizeOut, 0);
      FreeMem(ptrDIOCBufferOut);
    end;

  finally
    FillChar(ptrDIOCBufferIn^, bufferSizeIn, 0);
    FreeMem(ptrDIOCBufferIn);
  end;


  Result := retval;

end;


// ----------------------------------------------------------------------------
// encryptFlag - set to TRUE to encrypt, FALSE to decrypt
function TOTFEFreeOTFE.EncryptDecryptData(
                                   encryptFlag: boolean;
                                   cypherKernelModeDeviceName: string;
                                   cypherGUID: TGUID;
                                   var key: string;
                                   var IV: string;
                                   var inData: string;
                                   var outData: string
                                  ): boolean;
var
  cypherHandle: THandle;
  BytesReturned: DWORD;
  retval: boolean;
  ptrDIOCBufferIn: PDIOC_CYPHER_DATA_IN;
  bufferSizeIn: DWORD;
  ptrDIOCBufferOut: PDIOC_CYPHER_DATA_OUT;
  bufferSizeOut: DWORD;
  cypherDetails: TFreeOTFECypher;
  deviceUserModeName: string;
  dwIoControlCode: DWORD;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;
  retval := FALSE;

  CheckActive();

  // Determine the user mode MSDOS device name from the kernel mode device name
  deviceUserModeName := GetCypherDeviceUserModeDeviceName(cypherKernelModeDeviceName);

  if GetSpecificCypherDetails(cypherKernelModeDeviceName, cypherGUID, cypherDetails) then
    begin
    cypherHandle := ConnectDevice(deviceUserModeName);
    if (cypherHandle = 0) then
      begin
      LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('couldn''t connect to: '+deviceUserModeName);
{$ENDIF}
      end
    else
      begin
      dwIoControlCode := IOCTL_FREEOTFECYPHER_DECRYPT;
      if (encryptFlag=TRUE) then
        begin
        dwIoControlCode := IOCTL_FREEOTFECYPHER_ENCRYPT;
        end;


      bufferSizeIn := sizeof(TDIOC_CYPHER_DATA_IN) -
                      sizeof(ptrDIOCBufferIn^.Key) +
                      Length(key) -
                      sizeof(ptrDIOCBufferIn^.IV) +
                      Length(IV) -
                      sizeof(ptrDIOCBufferIn^.Data) +
                      Length(inData);
      // Round up to next "DIOC boundry". Although this always results in an
      // oversized buffer, it's guaranteed to be big enough.
      bufferSizeIn := bufferSizeIn + (DIOC_BOUNDRY - (bufferSizeIn mod DIOC_BOUNDRY));
      ptrDIOCBufferIn := allocmem(bufferSizeIn);
      try
        // Just make sure its big enough
        bufferSizeOut := sizeof(TDIOC_CYPHER_DATA_OUT) - sizeof(ptrDIOCBufferOut^.Data) + Length(inData);
        ptrDIOCBufferOut := allocmem(bufferSizeOut);
        try
          ptrDIOCBufferIn.CypherGUID := cypherGUID;

          ptrDIOCBufferIn.KeyLength  := (Length(key) * 8);  //  In *bits*
          ptrDIOCBufferIn.IVLength   := (Length(IV) * 8);  //  In *bits*
          ptrDIOCBufferIn.DataLength := Length(inData);  //  In *bytes*
          // This may seem a little weird, but we do this because the data is
          // immediatly after the IV, which is immediatly after the key
          StrMove(@ptrDIOCBufferIn.Key,                                   PChar(key),    Length(key));
          StrMove(((PChar(@ptrDIOCBufferIn.Key))+length(key)),            PChar(IV),     Length(IV));
          StrMove(((PChar(@ptrDIOCBufferIn.Key))+length(key)+length(IV)), PChar(inData), Length(inData));

          if (DeviceIoControl(
                              cypherHandle,
                              dwIoControlCode,
                              ptrDIOCBufferIn,
                              bufferSizeIn,
                              ptrDIOCBufferOut,
                              bufferSizeOut,
                              BytesReturned,
                              nil
                             )) then
            begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('_6_ supplied: '+inttostr(bufferSizeOut)+' used: '+inttostr(BytesReturned));
{$ENDIF}
            retval := TRUE;
            if (BytesReturned <= bufferSizeOut) then
              begin
              // Set hashOut so that it has enough characters which can be
              // overwritten with StrMove
              outData := StringOfChar(#0, BytesReturned);

              StrMove(PChar(outData), @ptrDIOCBufferOut.Data, BytesReturned);

              retval := TRUE;
              end
            else
              begin
              LastErrorCode := OTFE_ERR_CYPHER_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('incorrect bytecount returned');
{$ENDIF}
              end;

            end
          else
            begin
            LastErrorCode := OTFE_ERR_CYPHER_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('cypher DIOC 2 FAIL');
{$ENDIF}
            end;


        finally
          FillChar(ptrDIOCBufferOut^, bufferSizeOut, 0);
          FreeMem(ptrDIOCBufferOut);
        end;

      finally
        FillChar(ptrDIOCBufferIn^, bufferSizeIn, 0);
        FreeMem(ptrDIOCBufferIn);
      end;


      DisconnectDevice(cypherHandle);
      end;


    end
  else
    begin
    LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Unable to EncryptDecryptData');
{$ENDIF}
    end;

  Result := retval;

end;


// ----------------------------------------------------------------------------
// Encrypt data using the cypher/cypher driver identified
// Encrypted data returned in "plaintext"
function TOTFEFreeOTFE.EncryptData(
                     cypherKernelModeDeviceName: string;
                     cypherGUID: TGUID;
                     var key: string;
                     var IV: string;
                     var plaintext: string;
                     var cyphertext: string
                    ): boolean;
begin
  // Call generic function...
  Result := EncryptDecryptData(
                               TRUE,
                               cypherKernelModeDeviceName,
                               cypherGUID,
                               key,
                               IV,
                               plaintext,
                               cyphertext
                              );
end;


// ----------------------------------------------------------------------------
// Decrypt data using the cypher/cypher driver identified
// Decrypted data returned in "plaintext"
function TOTFEFreeOTFE.DecryptData(
                     cypherKernelModeDeviceName: string;
                     cypherGUID: TGUID;
                     var key: string;
                     var IV: string;
                     var cyphertext: string;
                     var plaintext: string
                    ): boolean;
begin
  // Call generic function...
  Result := EncryptDecryptData(
                               FALSE,
                               cypherKernelModeDeviceName,
                               cypherGUID,
                               key,
                               IV,
                               cyphertext,
                               plaintext
                              );
end;


// ----------------------------------------------------------------------------
// Note: Ignores any random padding data
function TOTFEFreeOTFE.ParseVolumeDetailsBlock(stringRep: string; var volumeDetailsBlock: TVolumeDetailsBlock): boolean;
const
  // CDB format 1 and 2 volume flags were used to identify sector IV generation
  // method
  VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV        = 1;  // Bit 0
  VOL_FLAGS_PRE_CDB_3__SECTOR_ID_ZERO_VOLSTART    = 2;  // Bit 1
  // Bit 2 unused
  VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE  = 8;  // Bit 3
var
  tmpByte: byte;
  tmpInt64: int64;
  tmpDWORD: DWORD;
  allOK: boolean;
  i: DWORD;
  idx: DWORD;
  tmpSectorIVGenMethod: TFreeOTFESectorIVGenMethod;
begin
  allOK := TRUE;

  // The first byte in the string is at index 1
  idx := 1;

  // 8 bits: CDB Format ID...
  volumeDetailsBlock.CDBFormatID:= ord(stringRep[idx]);
  inc(idx, 1);


  // 32 bits: Volume flags...
  tmpDWORD := 0;
  for i:=1 to 4 do
    begin
    tmpDWORD := tmpDWORD shl 8;  // Shift 8 bits to the *left*
    tmpByte := ord(stringRep[idx]);
    inc(idx);
    tmpDWORD := tmpDWORD + tmpByte;
    end;
  // CDBFormatID 1 CDBs (FreeOTFE v00.00.01 and v00.00.02) had a ***BUG*** that
  // caused the wrong value to be read back as the VolumeFlags
  if (volumeDetailsBlock.CDBFormatID < 2) then
    begin
    volumeDetailsBlock.VolumeFlags:= ord(stringRep[10]);
    end
  else
    begin
    volumeDetailsBlock.VolumeFlags:= tmpDWORD;
    end;


  // For CDB Format IDs less than 2, determine the sector IV generation method
  // based on the volume flags, and switch off the volume flags used
  if (volumeDetailsBlock.CDBFormatID < 3) then
    begin
    // VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV and VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE -> foivgHash64BitSectorID
    // 0                                        and VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE -> foivgUnknown (can't hash sector ID if it's not being used!)
    // VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV and 0                                              -> foivg64BitSectorID
    // 0                                        and 0                                              -> foivgNone

    volumeDetailsBlock.SectorIVGenMethod:= foivgUnknown;
    if (
         (
            volumeDetailsBlock.VolumeFlags AND
            (VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV OR VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE)
         ) =
            (VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV OR VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE)
       ) then
      begin
      volumeDetailsBlock.SectorIVGenMethod:= foivgHash64BitSectorID;
      end
    else if (
         (
            volumeDetailsBlock.VolumeFlags AND
            (0                                        OR VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE)
         ) =
            (0                                        OR VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE)
       ) then
      begin
      volumeDetailsBlock.SectorIVGenMethod:= foivgUnknown;
      end
    else if (
         (
            volumeDetailsBlock.VolumeFlags AND
            (VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV OR 0                                             )
         ) =
            (VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV OR 0                                             )
       ) then
      begin
      volumeDetailsBlock.SectorIVGenMethod:= foivg64BitSectorID;
      end
    else if (
         (
            volumeDetailsBlock.VolumeFlags AND
            (0                                        OR 0                                             )
         ) =
            (0                                        OR 0                                             )
       ) then
      begin
      volumeDetailsBlock.SectorIVGenMethod:= foivgNone;
      end;

    volumeDetailsBlock.VolumeFlags := volumeDetailsBlock.VolumeFlags AND
                                      not (VOL_FLAGS_PRE_CDB_3__USE_SECTOR_ID_AS_IV OR VOL_FLAGS_PRE_CDB_3__HASH_SECTOR_ID_BEFORE_USE);
    end;


  // 64 bits: Partition length...
  tmpInt64 := 0;
  for i:=1 to 8 do
    begin
    tmpInt64 := tmpInt64 shl 8;  // Shift 8 bits to the *left*
    tmpByte := ord(stringRep[idx]);
    inc(idx);
    tmpInt64 := tmpInt64 + tmpByte;
    end;
  volumeDetailsBlock.PartitionLen := tmpInt64;


  // 32 bits: Master key length...
  tmpDWORD := 0;
  for i:=1 to 4 do
    begin
    tmpDWORD := tmpDWORD shl 8;  // Shift 8 bits to the *left*
    tmpByte := ord(stringRep[idx]);
    inc(idx);
    tmpDWORD := tmpDWORD + tmpByte;
    end;
  volumeDetailsBlock.MasterKeyLength := tmpDWORD;
  // Note: If decrypted badly, can't rely on: criticalData.MasterKeyLen to be
  //       valid
  //       We check if the volume details block's master key length implies
  //       would put idx beyond the amount of data we have
  if ((idx + (volumeDetailsBlock.MasterKeyLength div 8)) > DWORD(length(stringRep))) then
    begin
    allOK := FALSE;
    end;

  // Variable bits: Master key...
  if (allOK) then
    begin
    volumeDetailsBlock.MasterKey:= Copy(stringRep, idx, (volumeDetailsBlock.MasterKeyLength div 8));
    inc(idx, (volumeDetailsBlock.MasterKeyLength div 8));
    end;


  // 8 bits: Requested drive letter...
  if (allOK) then
    begin
    volumeDetailsBlock.RequestedDriveLetter := stringRep[idx];
    inc(idx);
    end;


  // If the CDB Format ID is 2 or more, the CDB has a volume IV block
  // For CDB Format IDs less than 2, use a NULL volume IV
  if (volumeDetailsBlock.CDBFormatID < 2) then
    begin
    volumeDetailsBlock.VolumeIVLength := 0;
    volumeDetailsBlock.VolumeIV := '';
    end
  else
    begin
    // 32 bits: Volume IV length...
    tmpDWORD := 0;
    for i:=1 to 4 do
      begin
      tmpDWORD := tmpDWORD shl 8;  // Shift 8 bits to the *left*
      tmpByte := ord(stringRep[idx]);
      inc(idx);
      tmpDWORD := tmpDWORD + tmpByte;
      end;
    volumeDetailsBlock.VolumeIVLength := tmpDWORD;
    // Note: If decrypted badly, can't rely on: criticalData.MasterKeyLen to be
    //       valid
    //       We check if the volume details block's master key length implies
    //       would put idx beyond the amount of data we have
    if ((idx + (volumeDetailsBlock.VolumeIVLength div 8)) > DWORD(length(stringRep))) then
      begin
      allOK := FALSE;
      end;

    // Variable bits: Master key...
    if (allOK) then
      begin
      volumeDetailsBlock.VolumeIV:= Copy(stringRep, idx, (volumeDetailsBlock.VolumeIVLength div 8));
      inc(idx, (volumeDetailsBlock.VolumeIVLength div 8));
      end;

    end;


  // For CDB Format IDs less than 2, determine the sector IV generation method
  // based on the volume flags, and reset the volume flags
  if (volumeDetailsBlock.CDBFormatID < 3) then
    begin
    // volumeDetailsBlock.SectorIVGenMethod already set above when by processing VolumeFlags
    end
  else
    begin
    // 8 bits: Sector IV generation method...
    tmpByte:= ord(stringRep[idx]);
// NEXT LINE COMMENTED OUT TO PREVENT COMPILER WARNING
//    inc(idx, 1);
    volumeDetailsBlock.SectorIVGenMethod := foivgUnknown;
    for tmpSectorIVGenMethod := low(TFreeOTFESectorIVGenMethod) to high(TFreeOTFESectorIVGenMethod) do
      begin
      if (FreeOTFESectorIVGenMethodID[tmpSectorIVGenMethod] = tmpByte) then
        begin
        volumeDetailsBlock.SectorIVGenMethod:= tmpSectorIVGenMethod;
        end;
      end;

    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Note: This function does *not* append random padding data
// Note that volumeDetails.CDBFormatID is ignored - it is force set to the latest value
// !!!!!!!!!!!!!!!!
// !!!  WARNING !!!
// !!!!!!!!!!!!!!!!
// IF YOU UPDATE THIS, YOU SHOULD PROBABLY ALSO CHANGE THE DEFINITION OF:
//   function BuildVolumeDetailsBlock(...)
//   function ParseVolumeDetailsBlock(...)
// !!!!!!!!!!!!!!!!
// !!!  WARNING !!!
// !!!!!!!!!!!!!!!!
function TOTFEFreeOTFE.BuildVolumeDetailsBlock(volumeDetailsBlock: TVolumeDetailsBlock; var stringRep: string): boolean;
var
  tmpDWORD: DWORD;
  tmpInt64: int64;
  tmpByte: byte;
  i: integer;
  allOK: boolean;
  tmpString: string;
begin
  allOK := TRUE;

  volumeDetailsBlock.CDBFormatID := CDB_FORMAT_ID;

  stringRep := '';

  // 8 bits: CDB Format ID...
  stringRep := stringRep + char(volumeDetailsBlock.CDBFormatID);


  // 32bits: Volume flags...
  tmpDWORD := volumeDetailsBlock.VolumeFlags;
  tmpString:= '';
  for i:=1 to 4 do
    begin
    tmpByte := (tmpDWORD AND $FF);
    tmpDWORD := tmpDWORD shr 8;  // Shift 8 bits to the *right*
    tmpString := char(tmpByte) + tmpString;
    end;
  stringRep := stringRep + tmpString;


  // 64 bits: Partition length...
  tmpInt64 := volumeDetailsBlock.PartitionLen;
  tmpString:= '';
  for i:=1 to 8 do
    begin
    tmpByte := (tmpInt64 AND $FF);
    tmpInt64 := tmpInt64 shr 8;  // Shift 8 bits to the *right*
    tmpString := char(tmpByte) + tmpString;
    end;
  stringRep := stringRep + tmpString;


  // 32 bits: Master key length...
  tmpDWORD := volumeDetailsBlock.MasterKeyLength;
  tmpString:= '';
  for i:=1 to 4 do
    begin
    tmpByte := (tmpDWORD AND $FF);
    tmpDWORD := tmpDWORD shr 8;  // Shift 8 bits to the *right*
    tmpString := char(tmpByte) + tmpString;
    end;
  stringRep := stringRep + tmpString;


  // Variable bits: Master key...
  stringRep := stringRep + Copy(volumeDetailsBlock.MasterKey, 1, (volumeDetailsBlock.MasterKeyLength div 8));


  // Requested drive letter...
  stringRep := stringRep + volumeDetailsBlock.RequestedDriveLetter;


  if (volumeDetailsBlock.CDBFormatID >= 2) then
    begin
    // 32 bits: Volume IV length...
    tmpDWORD := volumeDetailsBlock.VolumeIVLength;
    tmpString:= '';
    for i:=1 to 4 do
      begin
      tmpByte := (tmpDWORD AND $FF);
      tmpDWORD := tmpDWORD shr 8;  // Shift 8 bits to the *right*
      tmpString := char(tmpByte) + tmpString;
      end;
    stringRep := stringRep + tmpString;


    // Variable bits: Volume IV ...
    stringRep := stringRep + Copy(volumeDetailsBlock.VolumeIV, 1, (volumeDetailsBlock.VolumeIVLength div 8));

    
    if (volumeDetailsBlock.CDBFormatID >= 3) then
      begin
      // 8 bits: Sector IV generation method...
      stringRep := stringRep + char(FreeOTFESectorIVGenMethodID[volumeDetailsBlock.SectorIVGenMethod]);
      end;

    end;


  Result := allOK;
end;


// ----------------------------------------------------------------------------
// !!! WARNING !!!
// THIS SOFTWARE ASSUMES THAT THE FOLLOWING ARE ALL MULTIPLES OF 8:
//   Cypher keysizes (in bits)
//   Cypher blocksizes (in bits)
//   Hash lengths (in bits)
// !!!!!!!!!!!!!!!!
// !!!  WARNING !!!
// !!!!!!!!!!!!!!!!
// IF YOU UPDATE THIS, YOU SHOULD PROBABLY ALSO CHANGE THE DEFINITION OF:
//   function BuildVolumeDetailsBlock(...)
//   function ParseVolumeDetailsBlock(...)
// !!!!!!!!!!!!!!!!
// !!!  WARNING !!!
// !!!!!!!!!!!!!!!!
function TOTFEFreeOTFE.ReadVolumeCriticalData(
    filename: string;
    offsetWithinFile: int64;
    userPassword: string;
    saltLength: integer;  // In bits
    keyIterations: integer;
    
    var volumeDetails: TVolumeDetailsBlock;
    var CDBMetaData: TCDBMetaData
    ): boolean;
var
  allOK: boolean;
begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Attempting v1 format CDB decode failed...');
{$ENDIF}

  allOK := ReadVolumeCriticalData_v2(
                                    filename,
                                    offsetWithinFile,
                                    userPassword,
                                    saltLength,
                                    keyIterations,
                                    volumeDetails,
                                    CDBMetaData
                                    );

  if not(allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('v2 format CDB decode failed - attempting v1...');
{$ENDIF}
    allOK := ReadVolumeCriticalData_v1(
                                    filename,
                                    offsetWithinFile,
                                    userPassword,
                                    saltLength,
                                    volumeDetails,
                                    CDBMetaData
                                    );
    end;


  if (allOK) then
    begin
    // Warn user if the volume appears to be a later version that this software
    // supports
    if (volumeDetails.CDBFormatID > CDB_FORMAT_ID) then
      begin
      SDUMessageDlg(
                 'WARNING: This volume appears to have been created with a '+CRLF+
                 'later version of FreeOTFE than this software supports.'+CRLF+
                 CRLF+
                 'You may continue, however:'+CRLF+
                 '1) Your data may be read incorrectly due to changes in the '+CRLF+
                 'decryption process, and'+CRLF+
                 '2) It is not recommended that you write to this volume using this '+CRLF+
                 'software.',
                 mtWarning, [mbOK], 0);
      end;

    end;

  Result:= allOK;
end;


// ----------------------------------------------------------------------------
// Version 1 volumes:
//   *) Used the hash of the user's password + salt as the CDB
//      encryption/decryption key
//   *) Used the hash of the volume details block as the check data
// !!! WARNING !!!
// THIS SOFTWARE ASSUMES THAT THE FOLLOWING ARE ALL MULTIPLES OF 8:
//   Cypher keysizes (in bits)
//   Cypher blocksizes (in bits)
//   Hash lengths (in bits)
function TOTFEFreeOTFE.ReadVolumeCriticalData_v1(
    filename: string;
    offsetWithinFile: int64;
    userPassword: string;
    saltLength: integer;  // In bits

    var volumeDetails: TVolumeDetailsBlock;
    var CDBMetaData: TCDBMetaData
    ): boolean;
var
  validVolumeDetails: TVolumeDetailsBlockArray;
  validCDBMetaData: TCDBMetaDataArray;

  criticalDataBuffer: string;
  encryptedBlockLen: integer;  // In *bits*
  encryptedBlock: string;
  decryptedBlock: string;
  strVolumeDetailsBlock: string;
  hashDrivers: array of TFreeOTFEHashDriver;
  hi, hj: integer;
  currHashDriver: TFreeOTFEHashDriver;
  currHashImpl: TFreeOTFEHash;
  cypherDrivers: array of TFreeOTFECypherDriver;
  ci, cj: integer;
  currCypherDriver: TFreeOTFECypherDriver;
  currCypherImpl: TFreeOTFECypher;
  allOK: boolean;
  salt: string;
  saltedUserPassword: string;
  derivedKey: string;  // The key to be used in decrypting the CDB, after
                       // salting, hashing, HMACing, etc - but before any
                       // truncating/padding as a result of the cypher's keysize
  hk: integer;  // Hash value length, in *bits*
  ks: integer;  // Cypher keysize, in *bits*
  bs: integer;  // Cypher blocksize, in *bits*
  criticalDataKey: string;
  IV: string;
  checkDataDecrypted: string;
  checkDataGenerated: string;
  strDecryptedVolDetailsBlk: string;

  currVolumeDetails: TVolumeDetailsBlock;
  currCDBMetaData: TCDBMetaData;

  i: integer;
  tgtVolDetailsBlockLen: integer;  // In *bits*
  prevCursor: TCursor;
begin
  SetLength(validVolumeDetails, 0);
  SetLength(validCDBMetaData, 0);


  prevCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try


    // NOTE: "allOK" indicates a *fatal* error in processing; not that an invalid
    //       password was entered


    allOK := ReadRawVolumeCriticalData(
                                    filename,
                                    offsetWithinFile,
                                    criticalDataBuffer
                                   );


    // Append the salt onto the user's password
    if (allOK) then
      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Determining user''s salted password...');
  {$ENDIF}
      salt := Copy(criticalDataBuffer, 1, (saltLength div 8));
      saltedUserPassword := userPassword + salt;
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('User''s salted key:');
  DebugMsg('-- begin --');
  DebugMsgBinary(saltedUserPassword);
  DebugMsg('-- end --');
  {$ENDIF}
      end;

    // Obtain details of all hashes...
    if (allOK) then
      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Getting hash drivers...');
  {$ENDIF}
      SetLength(hashDrivers, 0);
      allOK := GetHashDrivers(TFreeOTFEHashDriverArray(hashDrivers));
      end;

    // Obtain details of all cyphers...
    if (allOK) then
      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Getting cypher drivers...');
  {$ENDIF}
      SetLength(cypherDrivers, 0);
      allOK := GetCypherDrivers(TFreeOTFECypherDriverArray(cypherDrivers));
      end;


    if (allOK) then
      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Starting main hashes loop...');
  {$ENDIF}
      // FOR ALL HASH DRIVERS...
      for hi:=low(hashDrivers) to high(hashDrivers) do
        begin
        currHashDriver := hashDrivers[hi];
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Checking using hash driver: '+currHashDriver.Title+' ('+GUIDToString(currHashDriver.DriverGUID)+')');
  {$ENDIF}
        // FOR ALL HASH ALGORITHMS SUPPORTED BY THE CURRENT DRIVER...
        for hj:=low(hashDrivers[hi].Hashes) to high(hashDrivers[hi].Hashes) do
          begin
          currHashImpl := currHashDriver.Hashes[hj];
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Checking using hash impl: '+GetHashDisplayTitle(currHashImpl));
  {$ENDIF}
          hk := currHashImpl.Length;
          if (hk < 0) then
            begin
            // Assume 512 bits for the hash length, if the hash supports arbitary
            // lengths
            hk := 512;
            end;


  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('hk: '+inttostr(hk));
  DebugMsg('About to derive key from user''s password and salt');
  {$ENDIF}

          // HASH THE USER's PASSWORD+SALT...
          if not(HashData(currHashDriver.DeviceKernelModeName, currHashImpl.HashGUID, saltedUserPassword, derivedKey)) then
            begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('ERROR: Hashing FAILED.');
  {$ENDIF}
            allOK := FALSE;
            end;


          if (allOK) then
            begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Hashed data OK.');
  DebugMsg('Processed user''s password and salt to give:');
  DebugMsgBinary(derivedKey);
  {$ENDIF}
            // FOR ALL CYPHER DRIVERS...
            for ci:=low(cypherDrivers) to high(cypherDrivers) do
              begin
              currCypherDriver := cypherDrivers[ci];
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Checking using cypher driver: '+currCypherDriver.Title+' ('+GUIDToString(currCypherDriver.DriverGUID)+')');
  {$ENDIF}
              // FOR ALL CYPHERS SUPPORTED BY THE CURRENT DRIVER...
              for cj:=low(currCypherDriver.Cyphers) to high(currCypherDriver.Cyphers) do
                begin
                currCypherImpl := currCypherDriver.Cyphers[cj];
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Checking using cypher impl: '+GetCypherDisplayTitle(currCypherImpl));
  {$ENDIF}

                // Determine keysize...
                ks := currCypherImpl.KeySize;  // In *bits*
                // Determine blocksize...
                bs := currCypherImpl.BlockSize;
                if (bs <= 0) then
                  begin
                  // Assume 8 bits for the blocksize, if the cypher supports arbitary
                  // block sizes
                  bs := 8;
                  end;

                // Extract the encrypted block from the critical data block
                encryptedBlockLen := (((CRITICAL_DATA_LENGTH - saltLength) div bs) * bs);
                // +1 because we index from 1, not zero
                encryptedBlock := Copy(criticalDataBuffer, ((saltLength div 8)+1), (encryptedBlockLen div 8));



                // Determine the criticalDataKey...
                // Fallback - assign no key (e.g. if the cypher's keysize = 0 bits)
                criticalDataKey := '';
                if (ks < 0) then
                  begin
                  // If the cypher's keysize is -1, then the critical data block
                  // en/decryption key is the hash value returned
                  criticalDataKey := derivedKey;
                  end
                else
                  begin
                  if ((Length(derivedKey) * 8) < ks) then
                    begin
                    // The hash value isn't long enough; right pad with zero bytes
                    criticalDataKey := derivedKey + StringOfChar(#0, ((ks div 8) - Length(derivedKey)));
                    end
                  else
                    begin
                    // The hash value is too long; truncate
                    criticalDataKey := Copy(derivedKey, 1, (ks div 8));
                    end;

                  end;  // if (ks > 0) then



                // Zero the IV for decryption
                IV := StringOfChar(#0, (bs div 8));

  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('About to decrypt data...');
  {$ENDIF}
                if not(DecryptData(
                                   currCypherDriver.DeviceKernelModeName,
                                   currCypherImpl.CypherGUID,
                                   criticalDataKey,
                                   IV,
                                   encryptedBlock,
                                   decryptedBlock
                                  )) then
                  begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('ERROR: Decryption FAILED.');
  {$ENDIF}
                  allOK := FALSE;
                  end
                else
                  begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Key:');
  DebugMsgBinary(criticalDataKey);
  DebugMsg('Decoded to:');
  DebugMsgBinary(decryptedBlock);
  {$ENDIF}
                  // Extract the decrypted check hash and volume details block
                  // 1 and +1 because we index from 1
                  checkDataDecrypted    := Copy(decryptedBlock, 1, (hk div 8));
                  tgtVolDetailsBlockLen := (((4096 - saltLength) div bs) * bs) - hk;
                  strVolumeDetailsBlock := Copy(decryptedBlock, ((hk div 8)+1), (tgtVolDetailsBlockLen div 8));
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('About to generate hash/HMAC using decoded data...');
  {$ENDIF}


                  // HASH THE USER DECODED DATA...
                  if not(HashData(currHashDriver.DeviceKernelModeName, currHashImpl.HashGUID, strVolumeDetailsBlock, checkDataGenerated)) then
                    begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('ERROR: Hash of decrypted data FAILED.');
  {$ENDIF}
                    allOK := FALSE;
                    end;


                  if (allOK) then
                    begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Decrypted data hashed OK.');
  {$ENDIF}
                    // If the checkdata newly generated is less than hk bits, right pad it with zero bits
                    if ((Length(checkDataGenerated) * 8) < hk) then
                      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Check hash generated hash less than hk bits; padding...');
  {$ENDIF}
                      checkDataGenerated := checkDataGenerated + StringOfChar(#0, (((Length(checkDataGenerated) * 8) - hk) div 8));
                      end;
                    // If the new checkdata is greater than hk bits, truncate it after the first
                    // hk bits
                    if ((Length(checkDataGenerated) * 8) > hk) then
                      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Check hash generated hash more than hk bits; padding...');
  {$ENDIF}
                      // +1 because we index from 1, not zero
                      Delete(checkDataGenerated, (hk div 8)+1, (Length(checkDataGenerated) - (hk div 8)));
                      end;


                    if (checkDataDecrypted = checkDataGenerated) then
                      begin
                      // We have found the hash and cypher to successfully decrypt!

  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('SUCCESS! Found valid hash/cypher combination!');
  DebugMsg('Volume details block is length: '+inttostr(length(strVolumeDetailsBlock)));
  {$ENDIF}
                      if not(ParseVolumeDetailsBlock(strVolumeDetailsBlock, currVolumeDetails)) then
                        begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('ERROR: FAILED to parse plaintext volume details block to structure');
  {$ENDIF}
                        // At this point, there has been some failure (e.g.
                        // critical data block version ID or block structure
                        // incorrect)
                        // Note that we *don't* set allOK to FALSE; this failure
                        // is valid and only indicates that the current
                        // cypher/hash combination are incorrect for the volume;
                        // not that something has gone wrong with the mount
                        // operation
                        end
                      else
                        begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('OK so far...');
  {$ENDIF}
                        // Sanity check...
                        if
                           (
                            (ks >= 0) AND
                            (currVolumeDetails.MasterKeyLength <> DWORD(ks))
                           ) then
                          begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('ERROR: Sanity check failed(!)');
  {$ENDIF}
                          // Skip to next loop iteration...
                          Continue;
                          end;

                        // Add the decrypted critical data area onto an array
                        // containing other successfully decrypted copies,
                        // together with the hash/cypher combination successfully used



                        // v1 volumes only use hashed for MAC and hash of
                        // user's salted password as KDF
                        currCDBMetaData.MACAlgorithm := fomacHash;
                        currCDBMetaData.KDFAlgorithm := fokdfHashWithSalt;

                        currCDBMetaData.CypherDriverKernelModeName := cypherDrivers[ci].DeviceKernelModeName;
                        currCDBMetaData.CypherGUID                 := cypherDrivers[ci].Cyphers[cj].CypherGUID;
                        currCDBMetaData.HashDriverKernelModeName   := hashDrivers[hi].DeviceKernelModeName;
                        currCDBMetaData.HashGUID                   := hashDrivers[hi].Hashes[hj].HashGUID;
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Got valid decryption cypher/hash combination');
  {$ENDIF}
                        SetLength(validVolumeDetails, (Length(validVolumeDetails)+1));
                        SetLength(validCDBMetaData, (Length(validCDBMetaData)+1));
                        validVolumeDetails[Length(validVolumeDetails)-1] := currVolumeDetails;
                        validCDBMetaData[Length(validCDBMetaData)-1] := currCDBMetaData;
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('OK.');
  {$ENDIF}

                        // Additional: Store internal information, if dumping
                        if DumpFlag then
                          begin
                          DumpCriticalDataKey        := criticalDataKey;
                          DumpCheckMAC               := checkDataDecrypted;
                          DumpPlaintextEncryptedBlock:= decryptedBlock;
                          DumpVolumeDetailsBlock     := strVolumeDetailsBlock;
                          end;

                        end;  // ELSE PART - if not(ParseVolumeDetailsBlock(strVolumeDetailsBlock, currVolumeDetails)) then

                      end
                    else
                      begin
                      // Do nothing - just loop around for the next one...
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Check hash and generated hash do not match.');
  {$ENDIF}
                      end;  // ELSE PART - if (checkDataDecrypted = checkDataGenerated) then

                    end;  // ELSE PART - if not(HashData(...

                  end;  // ELSE PART - if not(DecryptData(...

                end;  // for cj:=low(currCypherDriver.Cyphers) to high(currCypherDriver.Cyphers) do
              end;  // for ci:=low(cypherDrivers) to high(cypherDrivers) do
            end;  // ELSE PART - if (HashData(...

          end;  // for hj:=low(hashDrivers[hi].Hashes) to high(hashDrivers[hi].Hashes) do

        end;  // for hi:=low(hashDrivers) to high(hashDrivers) do

      end;  // if (allOK) then


  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Completed checking for all hash/cypher combinations');
  {$ENDIF}


  finally
    Screen.Cursor := prevCursor;
  end;


  // Depending on how many hash/cypher combinations were found, select the
  // appropriate one to use
  if (allOK) then
    begin
    allOK := ChooseFromValid(
                             validVolumeDetails,
                             validCDBMetaData,
                             volumeDetails,
                             CDBMetaData
                            );
    end;

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Finished function.');
{$ENDIF}

  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Version 2 volumes:
//   *) Used the PKCS#5 PBKDF2 key derivation function with the user's password
//      and salt to generate thethe CDB encryption/decryption key
//   *) Used the HMAC of the (derived key, volume details block) as the check data
// !!! WARNING !!!
// THIS SOFTWARE ASSUMES THAT THE FOLLOWING ARE ALL MULTIPLES OF 8:
//   Cypher keysizes (in bits)
//   Cypher blocksizes (in bits)
//   Hash lengths (in bits)
function TOTFEFreeOTFE.ReadVolumeCriticalData_v2(
    filename: string;
    offsetWithinFile: int64;
    userPassword: string;
    saltLength: integer;  // In bits
    keyIterations: integer;

    var volumeDetails: TVolumeDetailsBlock;
    var CDBMetaData: TCDBMetaData
    ): boolean;
var
  validVolumeDetails: TVolumeDetailsBlockArray;
  validCDBMetaData: TCDBMetaDataArray;

  hashDrivers: array of TFreeOTFEHashDriver;
  hi, hj: integer;
  currHashDriver: TFreeOTFEHashDriver;
  currHashImpl: TFreeOTFEHash;

  cypherDrivers: array of TFreeOTFECypherDriver;
  ci, cj: integer;
  currCypherDriver: TFreeOTFECypherDriver;
  currCypherImpl: TFreeOTFECypher;

  allOK: boolean;
  salt: string;
  derivedKey: string;  // The key to be used in decrypting the CDB, after
                       // salting, hashing, HMACing, etc - but before any
                       // truncating/padding as a result of the cypher's keysize
  cdkSize_bits: integer;  // Size of the "critical data key", in *bits*
  maxCDKSize_bits: integer;  // Max size of the "critical data key", in *bits*
  IV: string;

  criticalDataBlock: string;

  currVolumeDetails: TVolumeDetailsBlock;
  currCDBMetaData: TCDBMetaData;

  i: integer;
  leb_bits: integer;  // In *bits*
  criticalDataKey: string;
  maxCriticalDataKey: string;

  paddedEncryptedBlock: string;
  encryptedBlock: string;
  plaintextEncryptedBlock: string;

  paddedCheckMAC: string;
  checkMAC: string;
  generatedMAC: string;

  volumeDetailsBlock: string;

  MACAlgorithm: TFreeOTFEMACAlgorithm;
  KDFAlgorithm: TFreeOTFEKDFAlgorithm;
  prevCursor: TCursor;
begin
  SetLength(validVolumeDetails, 0);
  SetLength(validCDBMetaData, 0);


  prevCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try


    MACAlgorithm:= fomacHMAC;
    KDFAlgorithm:= fokdfPBKDF2;

    // NOTE: "allOK" indicates whether or not a *fatal* error in processing;
    //       NOT that an invalid password was entered
    allOK := TRUE;


    // Obtain details of all hashes...
    if (allOK) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Getting hash drivers...');
{$ENDIF}
      SetLength(hashDrivers, 0);
      allOK := GetHashDrivers(TFreeOTFEHashDriverArray(hashDrivers));
      end;


    // Obtain details of all cyphers...
    if (allOK) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Getting cypher drivers...');
{$ENDIF}
      SetLength(cypherDrivers, 0);
      allOK := GetCypherDrivers(TFreeOTFECypherDriverArray(cypherDrivers));
      end;


    // Read in the appropriate CDB...
    if (allOK) then
      begin
      allOK := ReadRawVolumeCriticalData(
                                      filename,
                                      offsetWithinFile,
                                      criticalDataBlock
                                     );
      end;


    // Strip off salt
    if (allOK) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Extracting salt...');
{$ENDIF}
      salt := Copy(criticalDataBlock, 1, (saltLength div 8));
      paddedEncryptedBlock := Copy(
                                   criticalDataBlock,
                                   ((saltLength div 8)+1),
                                   (Length(criticalDataBlock) - (saltLength div 8))
                                  );
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Salt follows:');
DebugMsgBinary(salt);
{$ENDIF}
      end;



    // !!! OPTIMISATION !!!
    // Determine the longest cypher key
    if (allOK) then
      begin
      maxCDKSize_bits := 0;

      // FOR ALL CYPHER DRIVERS...
      for ci:=low(cypherDrivers) to high(cypherDrivers) do
        begin
        currCypherDriver := cypherDrivers[ci];
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Checking using cypher driver: '+currCypherDriver.Title+' ('+GUIDToString(currCypherDriver.DriverGUID)+')');
{$ENDIF}
        // FOR ALL CYPHERS SUPPORTED BY THE CURRENT DRIVER...
        for cj:=low(currCypherDriver.Cyphers) to high(currCypherDriver.Cyphers) do
          begin
          currCypherImpl := currCypherDriver.Cyphers[cj];
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Checking using cypher impl: '+GetCypherDisplayTitle(currCypherImpl));
{$ENDIF}

          // Determine size of "critical data key"...
          cdkSize_bits := currCypherImpl.KeySize;  // In *bits*
          if (cdkSize_bits < 0) then
            begin
            cdkSize_bits := 512;
            end;

          maxCDKSize_bits := max(maxCDKSize_bits, cdkSize_bits);

          end;
        end;
      end;



    if (allOK) then
      begin
      try
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Starting main hashes loop...');
{$ENDIF}
        // FOR ALL HASH DRIVERS...
        for hi:=low(hashDrivers) to high(hashDrivers) do
          begin
          currHashDriver := hashDrivers[hi];
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Checking using hash driver: '+currHashDriver.Title+' ('+GUIDToString(currHashDriver.DriverGUID)+')');
{$ENDIF}
          // FOR ALL HASH ALGORITHMS SUPPORTED BY THE CURRENT DRIVER...
          for hj:=low(hashDrivers[hi].Hashes) to high(hashDrivers[hi].Hashes) do
            begin
            currHashImpl := currHashDriver.Hashes[hj];
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Checking using hash impl: '+GetHashDisplayTitle(currHashImpl));
{$ENDIF}


            // v2 volumes won't work with hash algorithms with <= 0 length, or
            // blocksize, as HMAC/PBKDF2 require these to be fixed
            //  - skip these, and move onto the next hash algorithm (if any)
            if (
                (currHashImpl.Length <= 0) OR
                (currHashImpl.BlockSize <= 0)
               ) then
              begin
              Continue;
              end;

            // !!! OPTIMISATION !!!
            // Because we only allow PBKDF2, which:
            //   a) Doesn't need a cypher
            //   b) Short keys are identical to truncated longer keys
            // We derive the key *here*, outside the cypher loop
            if (KDFAlgorithm <> fokdfPBKDF2) then
              begin
              raise EFreeOTFEInternalError.Create('PBKDF2 assumed - optimisation failed');
              end;

            allOK := DeriveKey(
                            KDFAlgorithm,
                            currHashDriver.DeviceKernelModeName,
                            currHashImpl.HashGUID,
                            '',
                            StringToGUID(NULL_GUID),
                            userPassword,
                            salt,
                            keyIterations,
                            maxCDKSize_bits,  // cdkSizeBits
                            maxCriticalDataKey
                           );
            if not(allOK) then
              begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg(
     'FAILED TO DERIVE CRITICAL DATA KEY:'+CRLF+
     'Hash:'+CRLF+
     currHashImpl.Title+' ('+GUIDToString(currHashImpl.HashGUID)+')'+CRLF+
     'Cypher:'+CRLF+
     currCypherImpl.Title+' ('+GUIDToString(currCypherImpl.CypherGUID)+')'
    );
{$ENDIF}
              SDUMessageDlg(
                         'Failed to derive critical data key.'+CRLF+
                         CRLF+
                         'Hash:'+CRLF+
                         '  '+currHashDriver.DeviceKernelModeName+CRLF+
                         '  '+GUIDToString(currHashImpl.HashGUID),
                         mtError, [mbOK], 0
                        );
              // Bail out...
              raise EFreeOTFEReadCDBFailure.Create('Failed to derive key');
              end;


            // FOR ALL CYPHER DRIVERS...
            for ci:=low(cypherDrivers) to high(cypherDrivers) do
              begin
              currCypherDriver := cypherDrivers[ci];
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Checking using cypher driver: '+currCypherDriver.Title+' ('+GUIDToString(currCypherDriver.DriverGUID)+')');
{$ENDIF}
              // FOR ALL CYPHERS SUPPORTED BY THE CURRENT DRIVER...
              for cj:=low(currCypherDriver.Cyphers) to high(currCypherDriver.Cyphers) do
                begin
                currCypherImpl := currCypherDriver.Cyphers[cj];
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Checking using cypher impl: '+GetCypherDisplayTitle(currCypherImpl));
{$ENDIF}

                // Determine size of "critical data key"...
                cdkSize_bits := currCypherImpl.KeySize;  // In *bits*
                if (cdkSize_bits < 0) then
                  begin
                  cdkSize_bits := 512;
                  end;


                // !!! OPTIMISATION !!!
                criticalDataKey := Copy(maxCriticalDataKey, 1, (cdkSize_bits div 8));


                // Calculate "leb"
                if (currCypherImpl.BlockSize > 8) then
                  begin
                  leb_bits := ((CRITICAL_DATA_LENGTH - saltLength) div currCypherImpl.BlockSize) * currCypherImpl.BlockSize;
                  end
                else
                  begin
                  leb_bits := (CRITICAL_DATA_LENGTH - saltLength);
                  end;

                // Strip off the padding after the encrypted block
                encryptedBlock := Copy(paddedEncryptedBlock, 1, (leb_bits div 8));

                // Zero the IV for decryption
                IV := '';
                if (currCypherImpl.BlockSize > 0) then
                  begin
                  IV := StringOfChar(#0, (currCypherImpl.BlockSize div 8));
                  end;

                allOK := DecryptData(
                                     currCypherDriver.DeviceKernelModeName,
                                     currCypherImpl.CypherGUID,
                                     criticalDataKey,
                                     IV,
                                     encryptedBlock,
                                     plaintextEncryptedBlock
                                    );
                if not(allOK) then
                  begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg(
         'FAILED TO DECRYPT ENCRYPTED BLOCK:'+CRLF+
         'Hash:'+CRLF+
         currHashImpl.Title+' ('+GUIDToString(currHashImpl.HashGUID)+')'+CRLF+
         'Cypher:'+CRLF+
         currCypherImpl.Title+' ('+GUIDToString(currCypherImpl.CypherGUID)+')'
        );
{$ENDIF}
                  SDUMessageDlg(
                             'Failed to decrypt encrypted block.'+CRLF+
                             CRLF+
                             'Cypher:'+CRLF+
                             '  '+currCypherDriver.DeviceKernelModeName+CRLF+
                             '  '+GUIDToString(currCypherImpl.CypherGUID),
                             mtError, [mbOK], 0
                            );
                  // Bail out...
                  raise EFreeOTFEReadCDBFailure.Create('Failed to decrypt encrypted block');
                  end;


                paddedCheckMAC := Copy(plaintextEncryptedBlock, 1, (CDB_MAX_MAC_LENGTH div 8));
                volumeDetailsBlock := Copy(
                                           plaintextEncryptedBlock,
                                           ((CDB_MAX_MAC_LENGTH div 8)+1),
                                           (Length(plaintextEncryptedBlock) - (CDB_MAX_MAC_LENGTH div 8))
                                          );

                allOK := MACData(
                                  MACAlgorithm,
                                  currHashDriver.DeviceKernelModeName,
                                  currHashImpl.HashGUID,
                                  currCypherDriver.DeviceKernelModeName,
                                  currCypherImpl.CypherGUID,
                                  criticalDataKey,
                                  volumeDetailsBlock,
                                  generatedMAC,
                                  -1
                                 );
                if not(allOK) then
                  begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg(
         'FAILED TO GENERATE MAC:'+CRLF+
         'Hash:'+CRLF+
         currHashImpl.Title+' ('+GUIDToString(currHashImpl.HashGUID)+')'+CRLF+
         'Cypher:'+CRLF+
         currCypherImpl.Title+' ('+GUIDToString(currCypherImpl.CypherGUID)+')'
        );
{$ENDIF}
                  SDUMessageDlg(
                             'Failed to generate check MAC.'+CRLF+
                             'Hash:'+CRLF+
                             '  '+currHashDriver.DeviceKernelModeName+CRLF+
                             '  '+GUIDToString(currHashImpl.HashGUID)+CRLF+
                             'Cypher:'+CRLF+
                             '  '+currCypherDriver.DeviceKernelModeName+CRLF+
                             '  '+GUIDToString(currCypherImpl.CypherGUID)+CRLF,
                             mtError, [mbOK], 0
                            );
                  // Bail out...
                  raise EFreeOTFEReadCDBFailure.Create('Failed to generate MAC');
                  end;


                // Truncate generated MAC to CDB_MAX_MAC_LENGTH bits, if it's too
                // long
                // Note that we can't do this by specifying a MAC length to
                // MACData(...) as this would right-pad the MAC with 0x00 if it
                // wasn't long enough
                if (length(generatedMAC) > (CDB_MAX_MAC_LENGTH * 8)) then
                  begin
                  // +1 because we index from 1, not zero
                  Delete(
                         generatedMAC,
                         ((CDB_MAX_MAC_LENGTH * 8)+1),
                         (length(generatedMAC) - (CDB_MAX_MAC_LENGTH * 8))
                        );
                  end;

                checkMAC := Copy(paddedCheckMAC, 1, Length(generatedMAC));

                // Check if we've found a valid hash/cypher combination
                if (checkMAC = generatedMAC) then
                  begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('SUCCESS! Found valid hash/cypher combination!');
DebugMsg('Volume details block is length: '+inttostr(length(volumeDetailsBlock)));
{$ENDIF}

                  if not(ParseVolumeDetailsBlock(volumeDetailsBlock, currVolumeDetails)) then
                    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('ERROR: FAILED to parse plaintext volume details block to structure');
{$ENDIF}
                    // At this point, there has been some failure (e.g.
                    // critical data block version ID or block structure
                    // incorrect)
                    // Note that we *don't* set allOK to FALSE; this failure
                    // is valid and only indicates that the current
                    // cypher/hash combination are incorrect for the volume;
                    // not that something has gone wrong with the mount
                    // operation
                    end
                  else
                    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('OK so far...');
{$ENDIF}
                    // Sanity check...
                    if
                       (
                        (currCypherImpl.KeySize >= 0) AND
                        (currVolumeDetails.MasterKeyLength <> DWORD(currCypherImpl.KeySize))
                       ) then
                      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('ERROR: Sanity check failed(!)');
{$ENDIF}
                      // Skip to next loop iteration...
                      Continue;
                      end;


                    // Add the decrypted critical data area onto an array
                    // containing other successfully decrypted copies,
                    // together with the hash/cypher combination successfully used



                    // v2 volumes only use HMAC for MAC and PBKDF2 of
                    // user's salted password as KDF
                    currCDBMetaData.MACAlgorithm := MACAlgorithm;
                    currCDBMetaData.KDFAlgorithm := KDFAlgorithm;

                    currCDBMetaData.CypherDriverKernelModeName := cypherDrivers[ci].DeviceKernelModeName;
                    currCDBMetaData.CypherGUID                 := cypherDrivers[ci].Cyphers[cj].CypherGUID;
                    currCDBMetaData.HashDriverKernelModeName   := hashDrivers[hi].DeviceKernelModeName;
                    currCDBMetaData.HashGUID                   := hashDrivers[hi].Hashes[hj].HashGUID;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Got valid decryption cypher/hash combination');
{$ENDIF}
                    SetLength(validVolumeDetails, (Length(validVolumeDetails)+1));
                    SetLength(validCDBMetaData, (Length(validCDBMetaData)+1));
                    validVolumeDetails[Length(validVolumeDetails)-1] := currVolumeDetails;
                    validCDBMetaData[Length(validCDBMetaData)-1] := currCDBMetaData;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('OK.');
{$ENDIF}

                    // Additional: Store internal information, if dumping
                    if DumpFlag then
                      begin
                      DumpCriticalDataKey        := criticalDataKey;
                      DumpCheckMAC               := checkMAC;
                      DumpPlaintextEncryptedBlock:= plaintextEncryptedBlock;
                      DumpVolumeDetailsBlock     := volumeDetailsBlock;
                      end;

                    end;  // ELSE PART - if not(ParseVolumeDetailsBlock(volumeDetailsBlock, currVolumeDetails)) then

                  end
                else
                  begin
                  // Do nothing - just loop around for the next one...
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Check hash and generated hash do not match.');
{$ENDIF}
                  end;  // ELSE PART - if (checkMAC = generatedMAC) then

                end;  // for cj:=low(currCypherDriver.Cyphers) to high(currCypherDriver.Cyphers) do

              end;  // for ci:=low(cypherDrivers) to high(cypherDrivers) do

            end; // for hj:=low(hashDrivers[hi].Hashes) to high(hashDrivers[hi].Hashes) do

          end;  // for hi:=low(hashDrivers) to high(hashDrivers) do

      except
        on EFreeOTFEReadCDBFailure do
          begin
          allOK := FALSE;
          end;

      end;

      end;  // if (allOK) then


{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Completed checking for all hash/cypher combinations');
{$ENDIF}


  finally
    Screen.Cursor := prevCursor;
  end;


  // Depending on how many hash/cypher combinations were found, select the
  // appropriate one to use
  if (allOK) then
    begin
    allOK := ChooseFromValid(
                             validVolumeDetails,
                             validCDBMetaData,
                             volumeDetails,
                             CDBMetaData
                            );
    end;

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Finished function.');
{$ENDIF}

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.ChooseFromValid(
                             validVolumeDetails: TVolumeDetailsBlockArray;
                             validCDBMetaDatas: TCDBMetaDataArray;
                             var volumeDetails: TVolumeDetailsBlock;
                             var CDBMetaData: TCDBMetaData
                            ): boolean;
var
  allOK: boolean;
  hashCypherSelectDlg: TfrmSelectHashCypher;
  i: integer;
begin
  allOK := TRUE;

  if (Length(validCDBMetaDatas) = 0) then
    begin
    // No valid combination of hash/cypher could be found
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('No valid hash/cypher combination found');
{$ENDIF}
    allOK := FALSE;
    end
  else if (Length(validCDBMetaDatas) = 1) then
    begin
    // There is only one valid hash/cypher combination that can be used; use it
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Exactly one hash/cypher combination found successfully');
{$ENDIF}
    volumeDetails := validVolumeDetails[low(validVolumeDetails)];
    CDBMetaData := validCDBMetaDatas[low(validCDBMetaDatas)];
    end
  else
    begin
    // More than one valid hash/cypher combination; offer the user the choice
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('More than one hash/cypher combination found');
{$ENDIF}
    hashCypherSelectDlg:= TfrmSelectHashCypher.Create(nil);
    try
      hashCypherSelectDlg.FreeOTFEObj := self;
      for i:=low(validCDBMetaDatas) to high(validCDBMetaDatas) do
        begin
        hashCypherSelectDlg.AddCombination(
                                           validCDBMetaDatas[i].HashDriverKernelModeName,
                                           validCDBMetaDatas[i].HashGUID,
                                           validCDBMetaDatas[i].CypherDriverKernelModeName,
                                           validCDBMetaDatas[i].CypherGUID
                                          );
        end;

      if (hashCypherSelectDlg.ShowModal() <> mrOK) then
        begin
        LastErrorCode := OTFE_ERR_USER_CANCEL;
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('User cancelled');
{$ENDIF}
        allOK := FALSE;
        end
      else
        begin
        // Locate the selected hash/cypher
        for i:=low(validCDBMetaDatas) to high(validCDBMetaDatas) do
          begin
          if (
              (validCDBMetaDatas[i].HashDriverKernelModeName = hashCypherSelectDlg.SelectedHashDriverKernelModeName()) and
              IsEqualGUID(validCDBMetaDatas[i].HashGUID, hashCypherSelectDlg.SelectedHashGUID()) and
              (validCDBMetaDatas[i].CypherDriverKernelModeName = hashCypherSelectDlg.SelectedCypherDriverKernelModeName()) and
              IsEqualGUID(validCDBMetaDatas[i].CypherGUID, hashCypherSelectDlg.SelectedCypherGUID())
             ) then
            begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('User''s selected hash/cypher combination found.');
{$ENDIF}
            volumeDetails := validVolumeDetails[i];
            CDBMetaData := validCDBMetaDatas[i];
            break;
            end;

          end;

        end;

    finally
      hashCypherSelectDlg.Free();
    end;

    end;


  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Write out a critical data block
// If the file specified doesn't already exist, it will be created
// ALL members of these two "volumeDetails" and "CDBMetaData" structs MUST
// be populated
// *EXCEPT* for:
//    volumeDetails.CDBFormatID - this is force set to the latest value
//    CDBMetaData.MACAlgorithm - set to appropriate value
//    CDBMetaData.KDFAlgorithm - set to appropriate value
function TOTFEFreeOTFE.WriteVolumeCriticalData(
    filename: string;
    offsetWithinFile: int64;
    userPassword: string;
    salt: string;
    keyIterations: integer;

    volumeDetails: TVolumeDetailsBlock;
    CDBMetaData: TCDBMetaData;

    randomPadData: string // Note: If insufficient is supplied, the write will
                          //       *fail*
                          //       The maximum that could possibly be required
                          //       is CRITICAL_DATA_LENGTH bits.
    ): boolean;
var
  encryptedBlock: string;
  allOK: boolean;
  sl_bits: integer;  // Salt value length, in *bits*
  criticalDataKey: string;
  IV: string;
  cypherDetails: TFreeOTFECypher;
  hashDetails: TFreeOTFEHash;
  volumeDetailsBlockNoPadding: string;
  volumeDetailsBlock: string;
  criticalDataBlock: string;
  cdkSize_bits: integer;  // In *bits*
  randomPaddingTwoLength_bits: integer;  // In *bits*
  leb_bits: integer;  // In *bits*
  volumeDetailsBlockLength_bits: integer;  // In *bits*
  volumeDetailsBlockNoPaddingLength_bits: integer;  // In *bits*
  checkMAC: string;
  paddedCheckMAC: string;
  randomPaddingThreeLength_bits: integer;  // In *bits*
  plaintextEncryptedBlock: string;
  randomPaddingOneLength_bits: integer;  // In *bits*
begin
  allOK := TRUE;

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('In WriteVolumeCriticalData');
{$ENDIF}

  // Force set; this function only supports this one CDB format
  CDBMetaData.MACAlgorithm:= fomacHMAC;
  CDBMetaData.KDFAlgorithm:= fokdfPBKDF2;


  // Set to 0 to get rid of compiler warning
  randomPaddingOneLength_bits := 0;

  sl_bits := Length(salt) * 8;

  // Note: No need to populate CDBMetaData.MACAlgorithm or
  //       CDBMetaData.KDFAlgorithm as the function which writes the CDB out
  //       populates them automatically


  
  // Get details of the cypher so we know the key and blocksize...
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('About to get cypher details...');
{$ENDIF}
    if GetSpecificCypherDetails(CDBMetaData.cypherDriverKernelModeName, CDBMetaData.cypherGUID, cypherDetails) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Cypher driver: '+CDBMetaData.cypherDriverKernelModeName);
DebugMsg('Cypher impl: '+cypherDetails.Title+' ('+GUIDToString(CDBMetaData.CypherGUID)+')');
{$ENDIF}
      //
      end
    else
      begin
      LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
      allOK := FALSE;
      end;
    end;


  // Get details of the hash so we know the hash length...
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('About to get hash details...');
{$ENDIF}
    if GetSpecificHashDetails(CDBMetaData.hashDriverKernelModeName, CDBMetaData.hashGUID, hashDetails) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Hash driver: '+CDBMetaData.hashDriverKernelModeName);
DebugMsg('Hash impl: '+hashDetails.Title+' ('+GUIDToString(CDBMetaData.HashGUID)+')');
{$ENDIF}
      end
    else
      begin
      LastErrorCode := OTFE_ERR_DRIVER_FAILURE;
      allOK := FALSE;
      end;
    end;



  // Sanity checks
  if (
      (cypherDetails.BlockSize <= 0) AND
      (volumeDetails.SectorIVGenMethod <> foivgNone)
     ) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Sector IVs cannot be used if cypher blocksize <= 0');
    end;
  if (
      (cypherDetails.BlockSize <= 0) AND
      (
       (volumeDetails.VolumeIVLength > 0) OR
       (Length(volumeDetails.VolumeIV) > 0)
      )
     ) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Volume IV cannot be used if cypher blocksize <= 0');
    end;
  if (hashDetails.Length = 0) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Hash selected cannot be used; hash length must be greater than zero due to PBKDF2');
    end;
  if (hashDetails.Length < 0) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Hash selected cannot be used; hash length must fixed due to PBKDF2');
    end;
  if (hashDetails.BlockSize = 0) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Hash selected cannot be used; zero blocksize means useless check MAC may will be generated (zero bytes) long with HMAC)');
    end;
  if (hashDetails.BlockSize < 0) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Hash selected cannot be used; hash blocksize must be fixed due to HMAC algorithm');
    end;
  if ((volumeDetails.MasterKeyLength mod 8) <> 0) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Master key length must be multiple of 8');
    end;
  if (
      (cypherDetails.KeySize >= 0) AND
      (volumeDetails.MasterKeyLength <> DWORD(cypherDetails.KeySize))
     ) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Master key length must be the same as the cypher''s keysize for cyphers with a fixed keysize');
    end;
  if ((volumeDetails.VolumeIVLength mod 8) <> 0) then
    begin
    raise EFreeOTFEInternalError.Create('Invalid: Volume IV length must be multiple of 8');
    end;



  // Determine size of "critical data key"...
  cdkSize_bits := cypherDetails.KeySize;  // In *bits*
  if (cdkSize_bits < 0) then
    begin
    cdkSize_bits := 512;
    end;


  // Derive the "critical data key"
  if (allOK) then
    begin
    allOK := DeriveKey(
                    CDBMetaData.KDFAlgorithm,
                    CDBMetaData.hashDriverKernelModeName,
                    CDBMetaData.hashGUID,
                    CDBMetaData.cypherDriverKernelModeName,
                    CDBMetaData.cypherGUID,
                    userPassword,
                    salt,
                    keyIterations,
                    cdkSize_bits,
                    criticalDataKey
                   );
    if not(allOK) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg(
         'FAILED TO DERIVE CRITICAL DATA KEY:'+CRLF+
         'Hash:'+CRLF+
         CDBMetaData.hashDriverKernelModeName+' ('+GUIDToString(CDBMetaData.hashGUID)+')'+CRLF+
         'Cypher:'+CRLF+
         CDBMetaData.cypherDriverKernelModeName+' ('+GUIDToString(CDBMetaData.cypherGUID)+')'
        );
{$ENDIF}
      LastErrorCode := OTFE_ERR_KDF_FAILURE;
      allOK := FALSE;
      end;
    end;


  // Create plaintext version of the "Volume details block" in-memory
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Converting volume details block structure to string rep...');
{$ENDIF}
    // Note: The string returned by BuildVolumeDetailsBlock(...) does *not*
    //       have any random data appended to it
    if not(BuildVolumeDetailsBlock(volumeDetails, volumeDetailsBlockNoPadding)) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('ERROR: FAILED to convert critical data structure to string rep.');
{$ENDIF}
      LastErrorCode := OTFE_ERR_UNKNOWN_ERROR;
      allOK := FALSE;
      end
    else
      begin
      if (cypherDetails.BlockSize > 8) then
        begin
        leb_bits := (((CRITICAL_DATA_LENGTH - sl_bits) div cypherDetails.BlockSize) * cypherDetails.BlockSize);
        end
      else
        begin
        leb_bits := (CRITICAL_DATA_LENGTH - sl_bits);
        end;

      randomPaddingOneLength_bits := (CRITICAL_DATA_LENGTH - sl_bits) - leb_bits;

      // Note: This *includes* the length of the "Random padding #2"
      volumeDetailsBlockLength_bits := leb_bits - CDB_MAX_MAC_LENGTH;

      // Note: This *excludes* the length of the "Random padding #2"
      volumeDetailsBlockNoPaddingLength_bits := (Length(volumeDetailsBlockNoPadding) * 8);

      randomPaddingTwoLength_bits := volumeDetailsBlockLength_bits -
                                     volumeDetailsBlockNoPaddingLength_bits;


      // Pad out with "random data #2"...
      if (randomPaddingTwoLength_bits > 0) then
        begin
        volumeDetailsBlock := volumeDetailsBlockNoPadding +
                              Copy(
                                   randomPadData,
                                   1,
                                   (randomPaddingTwoLength_bits div 8)
                                  );
        // 1 because we index from 1, not zero
        Delete(randomPadData, 1, (randomPaddingTwoLength_bits div 8));
        end;

      end;

    end;


  // Generate the check MAC of the "volume details block"
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('About to HMAC critical data to create check data...');
{$ENDIF}
    allOK := MACData(
                      CDBMetaData.MACAlgorithm,
                      CDBMetaData.hashDriverKernelModeName,
                      CDBMetaData.hashGUID,
                      CDBMetaData.cypherDriverKernelModeName,
                      CDBMetaData.cypherGUID,
                      criticalDataKey,
                      volumeDetailsBlock,
                      checkMAC,
                      -1
                     );
    if not(allOK) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg(
         'FAILED TO GENERATE MAC:'+CRLF+
         'Hash:'+CRLF+
         hashDetails.Title+' ('+GUIDToString(hashDetails.HashGUID)+')'+CRLF+
         'Cypher:'+CRLF+
         cypherDetails.Title+' ('+GUIDToString(cypherDetails.CypherGUID)+')'
        );
{$ENDIF}
      LastErrorCode := OTFE_ERR_MAC_FAILURE;
      allOK := FALSE;
      end;

    end;


  // Truncate/pad check data as appropriate
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('MAC check data generated OK.');
{$ENDIF}
    // If the checkdata newly generated is less than CDB_MAX_MAC_LENGTH bits,
    // right pad it with random data bits to CDB_MAX_MAC_LENGTH
    // If the checkdata newly generated is greater than CDB_MAX_MAC_LENGTH
    // bits, truncate it to the first CDB_MAX_MAC_LENGTH bits.
    if ((Length(checkMAC) * 8) < CDB_MAX_MAC_LENGTH) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Check data generated hash less than CDB_MAX_MAC_LENGTH bits; padding...');
{$ENDIF}
      randomPaddingThreeLength_bits := CDB_MAX_MAC_LENGTH - (Length(checkMAC) * 8);
      paddedCheckMAC := checkMAC + Copy(
                                        randomPadData,
                                        1,
                                        (randomPaddingThreeLength_bits div 8)
                                       );
      // 1 because we index from 1, not zero
      Delete(randomPadData, 1, (randomPaddingThreeLength_bits div 8));
      end
    else if ((Length(checkMAC) * 8) > CDB_MAX_MAC_LENGTH) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Check data generated hash greater than CDB_MAX_MAC_LENGTH bits; truncating...');
{$ENDIF}
      paddedCheckMAC := checkMAC;
      // +1 because we index from 1, not zero
      Delete(
             paddedCheckMAC,
             ((CDB_MAX_MAC_LENGTH div 8)+1),
             (Length(paddedCheckMAC) - (CDB_MAX_MAC_LENGTH div 8))
            );
      end
    else
      begin
      // No padding/truncation needed
      paddedCheckMAC := checkMAC;
      end;

    end;

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Padded check MAC is: ');
DebugMsg('-- begin --');
DebugMsgBinary(paddedCheckMAC);
DebugMsg('-- end --');
{$ENDIF}


  if (allOK) then
    begin
    // Prepend the check MAC (padded if necessary) to the volume details block
    // to form a plaintext version of the "Encrypted block"
    plaintextEncryptedBlock := paddedCheckMAC + volumeDetailsBlock;
    end;


  // Encrypt the plaintextEncryptedBlock, using
  //  - Zero'd IV
  //  - Critical data key
  if (allOK) then
    begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Creating zeroed IV');
{$ENDIF}
    // Zero the IV for decryption
    IV := '';
    if (cypherDetails.BlockSize > 0) then
      begin
      IV := StringOfChar(#0, (cypherDetails.BlockSize div 8));
      end;


{$IFDEF FREEOTFE_DEBUG}
DebugMsg('CriticalData.cypherDriverKernelModeName: '+CDBMetaData.cypherDriverKernelModeName);
DebugMsg('CriticalData.cypherGUID: '+GUIDToString(CDBMetaData.cypherGUID));
DebugMsg('CriticalDataKey len: '+inttostr(length(criticalDataKey)));
DebugMsg('IV len: '+inttostr(length(IV)));

DebugMsg('Critical data key:');
DebugMsgBinary(criticalDataKey);
DebugMsg('before encryption, decrypted block is:');
DebugMsgBinary(plaintextEncryptedBlock);
{$ENDIF}
    if not(EncryptData(
                       CDBMetaData.cypherDriverKernelModeName,
                       CDBMetaData.cypherGUID,
                       criticalDataKey,
                       IV,
                       plaintextEncryptedBlock,
                       encryptedBlock
                      )) then
      begin
{$IFDEF FREEOTFE_DEBUG}
DebugMsg('ERROR: FAILED to encrypt data block');
{$ENDIF}
      LastErrorCode := OTFE_ERR_CYPHER_FAILURE;
      allOK := FALSE;
      end;
      
    end;



  // Form the critical datablock using the salt bytes, the encrypted data block,
  // and random padding data #1
  if (allOK) then
    begin
    criticalDataBlock := salt +
                         encryptedBlock;
                         Copy(
                              randomPadData,
                              1,
                              (randomPaddingOneLength_bits div 8)
                             );
    // 1 because we index from 1, not zero
    Delete(randomPadData, 1, (randomPaddingOneLength_bits div 8));
    end;


  // Finally, write the critical data block out to the volume file, starting
  // from the appropriate offset
  if (allOK) then
    begin
    allOK := WriteRawVolumeCriticalData(
                                        filename,
                                        offsetWithinFile,
                                        criticalDataBlock
                                       );
    end;

{$IFDEF FREEOTFE_DEBUG}
DebugMsg('Exiting function');
{$ENDIF}


  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetHashList(var hashDispTitles, hashKernelModeDriverNames, hashGUIDs: TStringList): boolean;
var
  hashDrivers: array of TFreeOTFEHashDriver;
  i, j, k: integer;
  currHashDriver: TFreeOTFEHashDriver;
  currHash: TFreeOTFEHash;
  allOK: boolean;
  tmpPrettyTitle: string;
  inserted: boolean;
begin
  allOK := FALSE;

  SetLength(hashDrivers, 0);
  if (GetHashDrivers(TFreeOTFEHashDriverArray(hashDrivers))) then
    begin
    hashDispTitles.Clear();
    hashKernelModeDriverNames.Clear();
    hashGUIDs.Clear();

    for i:=low(hashDrivers) to high(hashDrivers) do
      begin
      currHashDriver := hashDrivers[i];
      for j:=low(hashDrivers[i].Hashes) to high(hashDrivers[i].Hashes) do
        begin
        currHash := hashDrivers[i].Hashes[j];

        tmpPrettyTitle := GetHashDisplayTitle(currHash);

        // Attempt to insert in alphabetic  order...
        inserted := FALSE;
        for k:=0 to (hashDispTitles.count-1) do
          begin
          if (tmpPrettyTitle < hashDispTitles[k]) then
            begin
            hashDispTitles.Insert(k, tmpPrettyTitle);
            hashKernelModeDriverNames.Insert(k, currHashDriver.DeviceKernelModeName);
            hashGUIDs.Insert(k, GUIDToString(currHash.HashGUID));
            inserted := TRUE;
            break;
            end;

          end;

        if not(inserted) then
          begin
          hashDispTitles.Add(tmpPrettyTitle);
          hashKernelModeDriverNames.Add(currHashDriver.DeviceKernelModeName);
          hashGUIDs.Add(GUIDToString(currHash.HashGUID));
          end;
          
        end;

      end;

    allOK := TRUE;
    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.GetCypherList(var cypherDispTitles, cypherKernelModeDriverNames, cypherGUIDs: TStringList): boolean;
var
  cypherDrivers: array of TFreeOTFECypherDriver;
  i, j, k: integer;
  currCypherDriver: TFreeOTFECypherDriver;
  currCypher: TFreeOTFECypher;
  allOK: boolean;
  tmpPrettyTitle: string;
  inserted: boolean;
begin
  allOK:= FALSE;
  
  SetLength(cypherDrivers, 0);
  if (GetCypherDrivers(TFreeOTFECypherDriverArray(cypherDrivers))) then
    begin
    for i:=low(cypherDrivers) to high(cypherDrivers) do
      begin
      currCypherDriver := cypherDrivers[i];
      for j:=low(cypherDrivers[i].Cyphers) to high(cypherDrivers[i].Cyphers) do
        begin
        currCypher := cypherDrivers[i].Cyphers[j];

        tmpPrettyTitle := GetCypherDisplayTitle(currCypher);

        // Attempt to insert in alphabetic  order...
        inserted := FALSE;
        for k:=0 to (cypherDispTitles.count-1) do
          begin
          if (tmpPrettyTitle < cypherDispTitles[k]) then
            begin
            cypherDispTitles.Insert(k, tmpPrettyTitle);
            cypherKernelModeDriverNames.Insert(k, currCypherDriver.DeviceKernelModeName);
            cypherGUIDs.Insert(k, GUIDToString(currCypher.CypherGUID));
            inserted := TRUE;
            break;
            end;

          end;

        if not(inserted) then
          begin
          cypherDispTitles.Add(tmpPrettyTitle);
          cypherKernelModeDriverNames.Add(currCypherDriver.DeviceKernelModeName);
          cypherGUIDs.Add(GUIDToString(currCypher.CypherGUID));
          end;
          
        end;

      end;

    allOK := TRUE;
    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Generate a prettyprinted title for a specific hash
// Returns '' on error
function TOTFEFreeOTFE.GetHashDisplayTitle(hash: TFreeOTFEHash): string;
begin
  Result := hash.Title + ' (' + inttostr(hash.Length) + '/' + inttostr(hash.BlockSize) + ')';

end;


// ----------------------------------------------------------------------------
// Generate a prettyprinted title for a specific cypher
// Returns '' on error
function TOTFEFreeOTFE.GetCypherDisplayTitle(cypher: TFreeOTFECypher): string;
var
  strCypherMode: string;
  strCypherSizes: string;
begin
  // This part prettifies the cypher details and populates the combobox
  // as appropriate
  strCypherMode := '';
  if (cypher.Mode = focmUnknown) then
    begin
    // Note the semicolon-space on the end
    strCypherMode := 'Mode unknown; ';
    end
  else if (cypher.Mode <> focmNone) then
    begin
    // Note the semicolon-space on the end
    strCypherMode := FreeOTFECypherModeTitle[cypher.Mode]+'; ';
    end;

  strCypherSizes := inttostr(cypher.KeySize) + '/' + inttostr(cypher.BlockSize);

  Result := cypher.Title + ' (' + strCypherMode + strCypherSizes + ')';

end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.CreateFreeOTFEVolumeWizard(): boolean;
var
  frmWizard: TfrmWizardCreateVolume;
  allOK: boolean;
  mr: integer;
begin
  allOK:= FALSE;
  LastErrorCode := OTFE_ERR_UNKNOWN_ERROR;

  CheckActive();

  if WarnIfNoHashOrCypherDrivers() then
    begin
    frmWizard:= TfrmWizardCreateVolume.Create(nil);
    try
      frmWizard.OTFEFreeOTFEObj := self;

      mr := frmWizard.ShowModal();
      if (mr = mrOK) then
        begin
        LastErrorCode := OTFE_ERR_SUCCESS;
        allOK:= TRUE;
        end
      else if (mr = mrCancel) then
        begin
        LastErrorCode := OTFE_ERR_USER_CANCEL;
        end;

    finally
      frmWizard.Free();
    end;

    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.CreateLinuxVolumeWizard(): boolean;
var
  volSizeDlg: TfrmNewVolumeSize;
  mr: integer;
  allOK: boolean;
  userCancel: boolean;
begin
  allOK := FALSE;

  volSizeDlg:= TfrmNewVolumeSize.Create(nil);
  try
    mr := volSizeDlg.ShowModal;
    if (mr = mrCancel) then
      begin
      LastErrorCode := OTFE_ERR_USER_CANCEL;
      end
    else if (mr = mrOK) then
      begin
      if not(SDUCreateLargeFile(volSizeDlg.Filename, volSizeDlg.VolumeSize, TRUE, userCancel)) then
        begin
        if not(userCancel) then
          begin
          SDUMessageDlg('An error occured while trying to create your volume file', mtError, [mbOK], 0);
          end;

        end
      else
        begin
        allOK := TRUE;
        end;
      end;

  finally
    volSizeDlg.Free();
  end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.CreateLinuxGPGKeyfile(): boolean;
var
  allOK: boolean;
begin
  allOK:= FALSE;

// xxx - implement GPG integration

  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Identify the next drive letter to mount a volume as
// i.e. If userDriveLetter is not set to #0, then return either that letter, or
//      the next free drive letter.
//      If userDriveLetter is set to #0, then use requiredDriveLetter instead
//      of userDriveLetter
// Returns: Drive letter to mount as, or #0 on error
function TOTFEFreeOTFE.GetNextDriveLetter(userDriveLetter, requiredDriveLetter: char): char;
var
  retVal: char;
  freeDriveLetters: string;
  searchDriveLetter: char;
begin
  retVal:= #0;

  searchDriveLetter := userDriveLetter;
  if (searchDriveLetter = #0) then
    begin
    searchDriveLetter := requiredDriveLetter;
    end;

  // If still #0, just get the next one after C:
  if (searchDriveLetter = #0) then
    begin
    searchDriveLetter := 'C';
    end;


  freeDriveLetters := uppercase(SDUGetUnusedDriveLetters());
  searchDriveLetter := upcase(searchDriveLetter);

  // Delete drive letters from the free drive letters, until we hit one which
  // appears after the one we've been requested - or we run out of free drive
  // letters
  while (freeDriveLetters <> '') and (freeDriveLetters[1]<searchDriveLetter) do
    begin
    Delete(freeDriveLetters, 1, 1);
    end;

  if (freeDriveLetters <> '') then
    begin
    retVal := freeDriveLetters[1];
    end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Display control for controlling underlying drivers
// You *shouldn't* call this while the component is active, as the user may
// try to do something dumn like uninstall the main FreeOTFE driver while we're
// connected to it!
procedure TOTFEFreeOTFE.ShowDriverControlDlg();
var
  dlg: TfrmDriverControl;
begin
  // TfrmDriverControl.Create(nil) will raise an exception if the user doens't
  // have the required privs to access driver control
  dlg:= TfrmDriverControl.Create(nil);
  try
    dlg.ShowModal();
  finally
    dlg.Free();
  end;

  // In case drivers were added/removed...
  CachesFlush();
end;


// ----------------------------------------------------------------------------
{$IFDEF FREEOTFE_DEBUG}
procedure TOTFEFreeOTFE.DebugMsg(msg: string);
begin
  DebugStrings.Add(msg);
  SendDebug(msg);
  if DebugShowMessage then
    begin
    showmessage(msg);
    end;
end;
{$ENDIF}

{$IFDEF FREEOTFE_DEBUG}
procedure TOTFEFreeOTFE.DebugClear();
begin
  DebugStrings.Clear();
end;
{$ENDIF}

{$IFDEF FREEOTFE_DEBUG}
procedure TOTFEFreeOTFE.DebugMsgBinary(msg: string);
var
  prettyStrings: TStringList;
  i: integer;
begin
  prettyStrings := TStringList.Create();
  try
    SDUPrettyPrintHex(msg, 0, Length(msg), TStringList(prettyStrings), 16);
    for i := 0 to (prettyStrings.Count - 1) do
      begin
      SendDebug(prettyStrings[i]);
      end;
    DebugStrings.AddStrings(prettyStrings);
    if DebugShowMessage then
      begin
      showmessage(prettyStrings.Text);
      end;
  finally
    prettyStrings.Free();
  end;

end;
{$ENDIF}

{$IFDEF FREEOTFE_DEBUG}
procedure TOTFEFreeOTFE.DebugMsgBinaryPtr(msg: PChar; len:integer);
var
  prettyStrings: TStringList;
  i: integer;
begin
  prettyStrings := TStringList.Create();
  try
    SDUPrettyPrintHex(Pointer(msg), 0, len, TStringList(prettyStrings), 16);
    for i := 0 to (prettyStrings.Count - 1) do
      begin
      SendDebug(prettyStrings[i]);
      end;
    DebugStrings.AddStrings(prettyStrings);
    if DebugShowMessage then
      begin
      showmessage(prettyStrings.Text);
      end;
  finally
    prettyStrings.Free();
  end;

end;
{$ENDIF}
 

// ----------------------------------------------------------------------------
//PortableDriverFilenames
function TOTFEFreeOTFE.PortableStart(driverFilenames: TStringList; showProgress: boolean): boolean;
var
  retVal: boolean;
  DriverControlObj: TOTFEFreeOTFEDriverControl;
  i: integer;
  status: DWORD;
  installedOK: boolean;
  startedOK: boolean;
  currDriverName: string;
  progressDlg: TSDUProgress_F;
  serviceState: DWORD;
  prevCursor: TCursor;
begin
  retVal := TRUE;

  progressDlg := nil;
  prevCursor := Screen.Cursor;
  if showProgress then
    begin
    progressDlg := TSDUProgress_F.Create(nil);
    end;

  try
    if (progressDlg <> nil) then
      begin
      progressDlg.Min := 0;
      // Yes, this is correct; it's driverFilenames.count and *not*
      // (driverFilenames.count - 1)
      progressDlg.Max := driverFilenames.count;
      progressDlg.Position := 0;
      progressDlg.Title := 'Starting portable mode drivers...';
      progressDlg.ShowStatusText := TRUE;
      progressDlg.Show();
      Screen.Cursor := crAppStart;
      Application.ProcessMessages();
      end;

    DriverControlObj := TOTFEFreeOTFEDriverControl.Create();
    try
      for i:=0 to (driverFilenames.count-1) do
        begin
        progressDlg.StatusText := 'Starting: ' +
                                  ExtractFilename(driverFilenames[i]) +
                                  '...';
        Application.ProcessMessages();

        // If a driver which uses the same filename (without the path) is
        // already installed, skip installation - we use the existing one
        currDriverName := DriverControlObj.GetServiceNameForFilename(driverFilenames[i]);
        if not(DriverControlObj.IsDriverInstalled(
                                                  currDriverName,
                                                  installedOK
                                                 )) then
          begin
          // Sanity... Force FALSE if we couldn't determine if it was installed
          // or not
          installedOK := FALSE;
          end;

        if not(installedOK) then
          begin
          status := DriverControlObj.InstallDriver(
                                                   driverFilenames[i],
                                                   TRUE,
                                                   currDriverName
                                                  );
          installedOK := ((status AND DRIVER_BIT_SUCCESS) = DRIVER_BIT_SUCCESS);
          end;

        // Start the driver, if it hasn't already been started.
        startedOK := FALSE;
        if (installedOK) then
          begin
          if DriverControlObj.GetServiceState(
                                              currDriverName,
                                              serviceState
                                             ) then
            begin
            startedOK := (serviceState = SERVICE_RUNNING);
            end;

          if not(startedOK) then
            begin
            // Start the driver, if needed
            startedOK := DriverControlObj.StartStopService(
                                                           currDriverName,
                                                           TRUE
                                                          );
            end;
              
          end;

        retVal := (retVal AND startedOK);

        if (progressDlg <> nil) then
          begin
          progressDlg.IncPosition();

          Application.ProcessMessages();

          if progressDlg.Cancel then
            begin
            break;
            end;
          end;

        end;

    finally
      DriverControlObj.Free();
    end;
  finally
    if showProgress then
      begin
      if (progressDlg <> nil) then
        begin
        progressDlg.Free();
        Screen.Cursor := prevCursor;
        end;
      end;
  end;

  // In case drivers were added/removed...
  CachesFlush();

  Result := retVal;
end;


function TOTFEFreeOTFE.PortableStop(): boolean;
var
  retVal: boolean;
  DriverControlObj: TOTFEFreeOTFEDriverControl;
  allDrivers: TStringList;
  wasInstalledPortable: boolean;
  i: integer;
  status: DWORD;
  allOK: boolean;
begin
  retVal := TRUE;

  DriverControlObj := TOTFEFreeOTFEDriverControl.Create();
  try
    allDrivers:= TStringList.Create();
    try
      if DriverControlObj.GetFreeOTFEDrivers(allDrivers) then
        begin
        for i:=0 to (allDrivers.count-1) do
          begin
          // If the driver was installed in portable mode, stop and uninstall it
          if DriverControlObj.IsDriverInstalledPortable(allDrivers[i], wasInstalledPortable) then
            begin
            if wasInstalledPortable then
              begin
              // Stop the driver
              DriverControlObj.StartStopService(allDrivers[i], FALSE);

              // Uninstall the driver
              status := DriverControlObj.UninstallDriver(allDrivers[i]);
              allOK := ((status AND DRIVER_BIT_SUCCESS) = DRIVER_BIT_SUCCESS);
              retVal := (retVal AND allOK);
              end;

            end;  // if DriverControlObj.IsDriverInstalledPortable(allDrivers[i], wasInstalledPortable) then

          end;  // for i:=0 to (allDrivers.count-1) do

        end;  // if DriverControlObj.GetFreeOTFEDrivers(allDrivers) then

    finally
      allDrivers.Free();
    end;

  finally
    DriverControlObj.Free();
  end;

  // In case drivers were added/removed...
  CachesFlush();

  Result := retVal;
end;


// -----------------------------------------------------------------------------
// Identify how many drivers are currently running in portable mode
// Returns -1 on failure
function TOTFEFreeOTFE.DriversInPortableMode(): integer;
var
  retVal: integer;
  DriverControlObj: TOTFEFreeOTFEDriverControl;
  allDrivers: TStringList;
  wasInstalledPortable: boolean;
  i: integer;
begin
  retVal := -1;

  try
    DriverControlObj := TOTFEFreeOTFEDriverControl.Create();
    try
      allDrivers:= TStringList.Create();
      try
        if DriverControlObj.GetFreeOTFEDrivers(allDrivers) then
          begin
          retVal := 0;
          for i:=0 to (allDrivers.count-1) do
            begin
            // If the driver was installed in portable mode, stop and uninstall it
            if DriverControlObj.IsDriverInstalledPortable(allDrivers[i], wasInstalledPortable) then
              begin
              if wasInstalledPortable then
                begin
                inc(retVal);
                end;

              end;  // if DriverControlObj.IsDriverInstalledPortable(allDrivers[i], wasInstalledPortable) then

            end;  // for i:=0 to (allDrivers.count-1) do

          end;  // if DriverControlObj.GetFreeOTFEDrivers(allDrivers) then

      finally
        allDrivers.Free();
      end;

    finally
      DriverControlObj.Free();
    end;

  except
    on EFreeOTFENeedAdminPrivs do
      begin
      // Do nothing - retVal already set to -1
      end;

  end;

  Result := retVal;
end;


// -----------------------------------------------------------------------------
// Determine if OTFE component can mount devices.
// Returns TRUE if it can, otherwise FALSE
function TOTFEFreeOTFE.CanMountDevice(): boolean;
begin
  // Supported
  Result := TRUE
end;


// -----------------------------------------------------------------------------
// Prompt the user for a device (if appropriate) and password (and drive
// letter if necessary), then mount the device selected
// Returns the drive letter of the mounted devices on success, #0 on failure
function TOTFEFreeOTFE.MountDevices(): string;
var
  selectedPartition: string;
  mountedAs: string;
  frmSelectVolumeType: TfrmSelectVolumeType;
  retVal: string;
  mr: integer;
begin
  mountedAs := '';

  CheckActive();

  if CanMountDevice() then
    begin
    // Ask the user if they want to mount their partitions as FreeOTFE or Linux
    // partitions
    frmSelectVolumeType:= TfrmSelectVolumeType.Create(nil);
    try
      mr := frmSelectVolumeType.ShowModal();
      if (mr = mrCancel) then
        begin
        LastErrorCode := OTFE_ERR_USER_CANCEL;
        end
      else if (mr = mrOK) then
        begin
        // Ask the user which partition they want to mount
        selectedPartition:= SelectPartition();
        if (selectedPartition = '') then
          begin
          LastErrorCode := OTFE_ERR_USER_CANCEL;
          end
        else
          begin
          // Mount the partition as the appropriate type
          if frmSelectVolumeType.FreeOTFEVolume then
            begin
            retVal := MountFreeOTFE(selectedPartition, FALSE);
            end
          else
            begin
            retVal := MountLinux(selectedPartition, FALSE);
            end;
          end;

        end;

    finally
      frmSelectVolumeType.Free();;
    end;

    end;

  Result := retVal;

end;


// -----------------------------------------------------------------------------
function TOTFEFreeOTFE.BackupVolumeCriticalData(
                                   srcFilename: string;
                                   srcOffsetWithinFile: int64;
                                   destFilename: string
                                  ): boolean;
var
  criticalData: string;
  destFileStream: TFileStream;
  allOK: boolean;
begin
  // Get the FreeOTFE driver to read a critical data block from the named file
  // starting from the specified offset
  allOK := ReadRawVolumeCriticalData(
                             srcFilename,
                             srcOffsetWithinFile,
                             criticalData
                            );

  if (allOK) then
    begin
    // Because we're just writing to a straight file, we don't need the
    // FreeOTFE driver to do this.
    // Also, because the file shouldn't already exist, the FreeOTFE driver
    // *won't* do this - it expects the target to already exist
    destFileStream := TFileStream.Create(destFilename, fmCreate OR fmShareDenyNone);
    try
      if length(criticalData)>0 then
        begin
        allOK := (destFileStream.Write(criticalData[1], length(criticalData)) = length(criticalData));
        end;
    finally
      destFileStream.Free();
    end;

    end;

  Result := allOK;
end;


// -----------------------------------------------------------------------------
function TOTFEFreeOTFE.RestoreVolumeCriticalData(
                                    srcFilename: string;
                                    destFilename: string;
                                    destOffsetWithinFile: int64
                                   ): boolean;
var
  criticalData: string;
  allOK: boolean;
begin
  // Get the FreeOTFE driver to read a critical data block from the named file
  // starting from the specified offset
  allOK := ReadRawVolumeCriticalData(
                             srcFilename,
                             0,
                             criticalData
                            );

  if (allOK) then
    begin
    // Get the FreeOTFE driver to write the critical data block to the named file
    // starting from the specified offset
    allOK := WriteRawVolumeCriticalData(
                             destFilename,
                             destOffsetWithinFile,
                             criticalData
                            );
    end;

  Result := allOK;
end;


// -----------------------------------------------------------------------------
function TOTFEFreeOTFE.CreateKeyfile(
                               srcFilename: string;
                               srcOffsetWithinFile: int64;
                               srcUserPassword: string;
                               srcSaltLength: integer;  // In bits
                               srcKeyIterations: integer;
                               keyFilename: string;
                               keyfileUserPassword: string;
                               keyfileSaltBytes: string;
                               keyfileKeyIterations: integer;
                               keyfileRequestedDrive: char;
                               keyfileRandomPadData: string  // Note: If insufficient is supplied, the write will *fail*
                                                             //       The maximum that could possibly be required is CRITICAL_DATA_LENGTH bits.
                              ): boolean;
var
  volumeDetails: TVolumeDetailsBlock;
  CDBMetaData: TCDBMetaData;
  allOK: boolean;
  userCancel: boolean;
begin
  allOK := FALSE;

  if ReadVolumeCriticalData(
                           srcFilename,
                           srcOffsetWithinFile,
                           srcUserPassword,
                           srcSaltLength,  // In bits
                           srcKeyIterations,
                           volumeDetails,
                           CDBMetaData
                          ) then
    begin
    volumeDetails.RequestedDriveLetter := keyfileRequestedDrive;

    // Just create a dummy file which will be populated with the keyfile's
    // contents (the driver doesn't create files, it just writes data to them)
    if (not(FileExists(keyFilename))) then
      begin
      SDUCreateLargeFile(keyFilename, (CRITICAL_DATA_LENGTH div 8), FALSE, userCancel);
      end;

    allOK := WriteVolumeCriticalData(
                                     keyFilename,
                                     0,
                                     keyfileUserPassword,
                                     keyfileSaltBytes,
                                     keyfileKeyIterations,
                                     volumeDetails,
                                     CDBMetaData,
                                     keyfileRandomPadData
                                    );
    end;

  Result := allOK;
end;


// -----------------------------------------------------------------------------
function TOTFEFreeOTFE.ChangeVolumePassword(
                               filename: string;
                               offsetWithinFile: int64;
                               oldUserPassword: string;
                               oldSaltLength: integer;  // In bits
                               oldKeyIterations: integer;
                               newUserPassword: string;
                               newSaltBytes: string;
                               newKeyIterations: integer;
                               newRequestedDriveLetter: char;
                               newRandomPadData: string  // Note: If insufficient is supplied, the write will *fail*
                                                         //       The maximum that could possibly be required is CRITICAL_DATA_LENGTH bits.
                              ): boolean;
var
  volumeDetails: TVolumeDetailsBlock;
  CDBMetaData: TCDBMetaData;
  allOK: boolean;
begin
  allOK := FALSE;

  if ReadVolumeCriticalData(
                           filename,
                           offsetWithinFile,
                           oldUserPassword,
                           oldSaltLength,  // In bits
                           oldKeyIterations,
                           volumeDetails,
                           CDBMetaData
                          ) then
    begin
    volumeDetails.RequestedDriveLetter := newRequestedDriveLetter;

    allOK := WriteVolumeCriticalData(
                                     filename,
                                     offsetWithinFile,
                                     newUserPassword,
                                     newSaltBytes,
                                     newKeyIterations,
                                     volumeDetails,
                                     CDBMetaData,
                                     newRandomPadData
                                    );
    end;

  Result := allOK;
end;


// -----------------------------------------------------------------------------
function TOTFEFreeOTFE.DumpCriticalDataToFile(
                               volFilename: string;
                               offsetWithinFile: int64;
                               userPassword: string;
                               saltLength: integer;  // In bits
                               keyIterations: integer;
                               dumpFilename: string
                              ): boolean;
var
  volumeDetails: TVolumeDetailsBlock;
  CDBMetaData: TCDBMetaData;
  allOK: boolean;
  dumpReport: TStringList;
  prettyPrintData: TStringList;
  criticalDataBuffer: string;
  hashTitle: string;
  cypherTitle: string;
  hashDetails: TFreeOTFEHash;
  cypherDetails: TFreeOTFECypher;
  readTimeStart: TDateTime;
  readTimeStop: TDateTime;
  readTimeDiff: TDateTime;
  Hour, Min, Sec, MSec: Word;
begin
  LastErrorCode := OTFE_ERR_UNKNOWN_ERROR;
  allOK := FALSE;

  CheckActive();

  prettyPrintData := TStringList.Create();
  try
    DumpFlag := TRUE;

    readTimeStart:= Now();

    // Read in the raw CDB to display the encrypted version, and then read it in
    // again to get the decrypt it
    if (
        ReadRawVolumeCriticalData(
                               volFilename,
                               offsetWithinFile,
                               criticalDataBuffer
                              )

          AND

        ReadVolumeCriticalData(
                               volFilename,
                               offsetWithinFile,
                               userPassword,
                               saltLength,  // In bits
                               keyIterations,
                               volumeDetails,
                               CDBMetaData
                              )
      ) then
      begin
      readTimeStop:= Now();

      // Generate the report
      dumpReport := TStringList.Create();
      try
        dumpReport.Add('Critical Data Block Dump');
        dumpReport.Add('========================');
        dumpReport.Add('');
        dumpReport.Add('Dump Created By');
        dumpReport.Add('---------------');
        dumpReport.Add('Platform              : PC ('+INSTALLED_OS_TITLE[SDUInstalledOS()]+')');
        dumpReport.Add('Application version   : v'+SDUGetVersionInfoString(ParamStr(0)));
        dumpReport.Add('Driver ID             : '+VersionStr());
        dumpReport.Add('');
        dumpReport.Add('');
        dumpReport.Add('User Supplied Information');
        dumpReport.Add('-------------------------');
        dumpReport.Add('Filename (user mode)  : '+volFilename);
        dumpReport.Add('Filename (kernel mode): '+GetKernelModeVolumeFilename(volFilename));
        dumpReport.Add('Password              : ');
        dumpReport.Add('  Length: '+inttostr((Length(userPassword) * 8))+' bits');
        dumpReport.Add('  Data  : ');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          userPassword,
                          0,
                          Length(userPassword),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);
        dumpReport.Add('Offset                : '+inttostr(offsetWithinFile)+' bytes');
        dumpReport.Add('Salt length           : '+inttostr(saltLength)+' bits');
        dumpReport.Add('Key iterations        : '+inttostr(keyIterations));

        dumpReport.Add('');
        dumpReport.Add('');
        dumpReport.Add('Plaintext Information');
        dumpReport.Add('---------------------');
        dumpReport.Add('Salt data             :');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          criticalDataBuffer,
                          0,
                          (saltLength div 8),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);


        dumpReport.Add('');
        dumpReport.Add('');
        dumpReport.Add('Dump Performance');
        dumpReport.Add('----------------');
        readTimeDiff := (readTimeStop - readTimeStart);
        DecodeTime(readTimeDiff, Hour, Min, Sec, MSec);
        dumpReport.Add('Time to process CDB   : '+inttostr(Hour)+' hours, '+inttostr(Min)+' mins, '+inttostr(Sec)+'.'+inttostr(MSec)+' secs');

        
        dumpReport.Add('');
        dumpReport.Add('');
        dumpReport.Add('Autodetermined Information');
        dumpReport.Add('--------------------------');

        hashTitle := 'ERROR: Unable to determine hash title?!';
        if GetSpecificHashDetails(
                                  CDBMetaData.HashDriverKernelModeName,
                                  CDBMetaData.HashGUID,
                                  hashDetails
                                 ) then
          begin
          hashTitle := GetHashDisplayTitle(hashDetails);
          end;

        cypherTitle := 'ERROR: Unable to determine cypher title?!';
        if GetSpecificCypherDetails(
                                    CDBMetaData.CypherDriverKernelModeName,
                                    CDBMetaData.CypherGUID,
                                    cypherDetails
                                   ) then
          begin
          cypherTitle := GetCypherDisplayTitle(cypherDetails);
          end;

        dumpReport.Add('Hash pretty title     : '+hashTitle);
        dumpReport.Add('Hash driver KM name   : '+CDBMetaData.HashDriverKernelModeName);
        dumpReport.Add('Hash GUID             : '+GUIDToString(CDBMetaData.HashGUID));
        dumpReport.Add('Cypher pretty title   : '+cypherTitle);
        dumpReport.Add('Cypher driver KM name : '+CDBMetaData.CypherDriverKernelModeName);
        dumpReport.Add('Cypher GUID           : '+GUIDToString(CDBMetaData.CypherGUID));
        dumpReport.Add('Critical data key     : ');
        dumpReport.Add('  KDF   : '+FreeOTFEKDFTitle[CDBMetaData.KDFAlgorithm]);
        dumpReport.Add('  Length: '+inttostr((Length(DumpCriticalDataKey) * 8))+' bits');
        dumpReport.Add('  Key   : ');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          DumpCriticalDataKey,
                          0,
                          Length(DumpCriticalDataKey),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);
        dumpReport.Add('Plaintext encrypted block: ');
        dumpReport.Add('  Length: '+inttostr((Length(DumpPlaintextEncryptedBlock) * 8))+' bits');
        dumpReport.Add('  Data  : ');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          DumpPlaintextEncryptedBlock,
                          0,
                          Length(DumpPlaintextEncryptedBlock),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);
        dumpReport.Add('Check MAC             : ');
        dumpReport.Add('  MAC algorithm: '+FreeOTFEMACTitle[CDBMetaData.MACAlgorithm]);
        dumpReport.Add('  Length       : '+inttostr((Length(DumpCheckMAC) * 8))+' bits');
        dumpReport.Add('  Data         : ');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          DumpCheckMAC,
                          0,
                          Length(DumpCheckMAC),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);
        dumpReport.Add('Volume details block  : ');
        dumpReport.Add('  Length       : '+inttostr((Length(DumpVolumeDetailsBlock) * 8))+' bits');
        dumpReport.Add('  Data         : ');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          DumpVolumeDetailsBlock,
                          0,
                          Length(DumpVolumeDetailsBlock),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);


        dumpReport.Add('');
        dumpReport.Add('');
        dumpReport.Add('Volume Details Block');
        dumpReport.Add('--------------------');
        dumpReport.Add('CDB format ID         : '+inttostr(volumeDetails.CDBFormatID));

        // Decode the VolumeFlags to human-readable format
        dumpReport.Add('Volume flags          : '+inttostr(volumeDetails.VolumeFlags));
        dumpReport.Add('                        '+
                       DumpCriticalDataToFileBitmap(
                                 volumeDetails.VolumeFlags,
                                 VOL_FLAGS_SECTOR_ID_ZERO_VOLSTART,
                                 'Sector ID zero is at the start of the encrypted data',
                                 'Sector ID zero is at the start of the host file/partition'
                                ));

        dumpReport.Add('Partition length      : '+inttostr(volumeDetails.PartitionLen)+' bytes');

        dumpReport.Add('Master key            : ');
        dumpReport.Add('  Length: '+inttostr(volumeDetails.MasterKeyLength)+' bits');
        dumpReport.Add('  Key   :');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          volumeDetails.MasterKey,
                          0,
                          (volumeDetails.MasterKeyLength div 8),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);
        dumpReport.Add('Sector IV generation  : '+FreeOTFESectorIVGenMethodTitle[volumeDetails.SectorIVGenMethod]);
        dumpReport.Add('Volume IV             : ');
        dumpReport.Add('  Length : '+inttostr(volumeDetails.VolumeIVLength)+' bits');
        dumpReport.Add('  IV data:');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          volumeDetails.VolumeIV,
                          0,
                          (volumeDetails.VolumeIVLength div 8),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);

        if (volumeDetails.RequestedDriveLetter = #0) then
          begin
          dumpReport.Add('Requested drive letter: None; use default');
          end
        else
          begin
          dumpReport.Add('Requested drive letter: '+volumeDetails.RequestedDriveLetter+':');
          end;

        dumpReport.Add('');
        dumpReport.Add('');
        dumpReport.Add('Encrypted CDB');
        dumpReport.Add('-------------');
        prettyPrintData.Clear();
        SDUPrettyPrintHex(
                          criticalDataBuffer,
                          0,
                          (CRITICAL_DATA_LENGTH div 8),
                          prettyPrintData
                         );
        dumpReport.AddStrings(prettyPrintData);


        // Save the report out to disk...
        dumpReport.SaveToFile(dumpFilename);
        allOK := TRUE;
      finally
        DumpFlag := FALSE;
        DumpCriticalDataKey        := '';
        DumpCheckMAC               := '';
        DumpPlaintextEncryptedBlock:= '';
        DumpVolumeDetailsBlock     := '';

        dumpReport.Clear();
        dumpReport.Free();
      end;

      end;

  finally
    prettyPrintData.Free();
  end;

  if (allOK) then
    begin
    LastErrorCode := OTFE_ERR_SUCCESS;
    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Return a nicely formatted string, decoding one bit of a bitmapped value
// bit - The zero-offset bit (i.e. the LSB is bit 0)
function TOTFEFreeOTFE.DumpCriticalDataToFileBitmap(
                               value: DWORD;
                               bit: integer;
                               unsetMeaning: string;
                               setMeaning: string
                              ): string;
var
  x: DWORD;
  i: integer;
  retVal: string;
begin
  retVal := '';

  x := 1;
  for i:=1 to bit do
    begin
    x := x * 2;
    end;

  if ((value AND x) = x) then
    begin
    retVal := 'Bit '+inttostr(bit)+': 1 ('+setMeaning+')';
    end
  else
    begin
    retVal := 'Bit '+inttostr(bit)+': 0 ('+unsetMeaning+')';
    end;

  Result := retval;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.WizardChangePassword(): boolean;
var
  dlg: TfrmWizardChangePasswordCreateKeyfile;
  allOK: boolean;
begin
  allOK := FALSE;

  if WarnIfNoHashOrCypherDrivers() then
    begin
    dlg:= TfrmWizardChangePasswordCreateKeyfile.Create(nil);
    try
      dlg.ChangePasswordCreateKeyfile := opChangePassword;
      dlg.OTFEFreeOTFE := self;
      allOK := (dlg.ShowModal() = mrOK);
    finally
      dlg.Free();
    end;

    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.WizardCreateKeyfile(): boolean;
var
  dlg: TfrmWizardChangePasswordCreateKeyfile;
  allOK: boolean;
begin
  allOK := FALSE;

  if WarnIfNoHashOrCypherDrivers() then
    begin
    dlg:= TfrmWizardChangePasswordCreateKeyfile.Create(nil);
    try
      dlg.ChangePasswordCreateKeyfile := opCreateKeyfile;
      dlg.OTFEFreeOTFE := self;
      allOK := (dlg.ShowModal() = mrOK);
    finally
      dlg.Free();
    end;
    
    end;

  Result := allOK;
end;


// ----------------------------------------------------------------------------
// Convert a kernel mode volume filename to a user mode volume filename
function TOTFEFreeOTFE.GetUserModeVolumeFilename(kernelModeFilename: string): string;
var
  retVal: string;
begin
  retVal := kernelModeFilename;


  // Kernel mode drivers take volume filenames of the format:
  //  \??\UNC<unc path>
  //  \??\<filename>
  //  <devicename>
  // e.g.:
  //   \??\UNC\myOtherComputer\shareName\dirname\file.dat
  //   \??\C:\file.dat
  //   \Device\Harddisk0\Partition1\path\filedisk.img
  //   \Device\Harddisk0\Partition1


  // Strip out the filename prefix
  if Pos('\??\UNC\', retVal) = 1 then
    begin
    // \\server\share\path\filedisk.img
    delete(retVal, 1, length('\??\UNC\'));
    retVal := '\\' + retVal;
    end
  else if Pos('\??\', retVal) = 1 then
    begin
    // C:\path\filedisk.img
    delete(retVal, 1, length('\??\'));
    end
  else
    begin
    // \Device\Harddisk0\Partition1\path\filedisk.img
    // Do nothing.
    end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Convert a user mode volume filename to a kernel mode volume filename
function TOTFEFreeOTFE.GetKernelModeVolumeFilename(userModeFilename: string): string;
var
  retVal: string;
begin
  retVal := '';


  // Kernel mode drivers take volume filenames of the format:
  //  \??\UNC<unc path>
  //  \??\<filename>
  //  <devicename>
  // e.g.:
  //   \??\UNC\myOtherComputer\shareName\dirname\file.dat
  //   \??\C:\file.dat
  //   \Device\Harddisk0\Partition1\path\filedisk.img
  //   \Device\Harddisk0\Partition1


  if (Pos('\\', userModeFilename) = 1) then
    begin
    // \\server\share\path\filedisk.img
    // Remove first "\"
    Delete(userModeFilename, 1, 1);
    retVal := '\??\UNC'+userModeFilename;
    end
  else if (
           (Pos(':\', userModeFilename) = 2) OR
           (Pos(':/', userModeFilename) = 2)
          ) then
    begin
    // C:\path\filedisk.img
    retVal := '\??\'+userModeFilename;
    end
  else
    begin
    // \Device\Harddisk0\Partition1\path\filedisk.img
    retVal := userModeFilename;
    end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Return TRUE/FALSE, depending on whether the specified KERNEL MODE volume
// filename refers to a partition/file
function TOTFEFreeOTFE.IsPartition(kernelModeFilename: string): boolean;
begin
  // In kernel mode, partitions start with just "\", not "\??\"
  Result := (Pos('\??\', kernelModeFilename) <> 1);

end;


// ----------------------------------------------------------------------------
// Get the FreeOTFE driver to read a critical data block from the named file
// starting from the specified offset
// criticalData - This will be set to the string representation of the raw
//                (encrypted) critical data block
function TOTFEFreeOTFE.ReadRawVolumeCriticalData(
                         filename: string;
                         offsetWithinFile: int64;
                         var criticalData: string
                        ): boolean;
begin
  Result := ReadRawVolumeData(
                             filename,
                             offsetWithinFile,
                             (CRITICAL_DATA_LENGTH div 8),
                             criticalData
                            );

end;


// ----------------------------------------------------------------------------
// Get the FreeOTFE driver to read data from the named file
// starting from the specified offset
// data - This will be set to the string representation of the raw data read
function TOTFEFreeOTFE.ReadRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;
var
  retVal: boolean;
begin
  retVal := ReadRawVolumeDataSimple(
                                    filename,
                                    offsetWithinFile,
                                    dataLength,
                                    data
                                   );
  if not(retVal) then
    begin
    retVal := ReadRawVolumeDataBounded(
                                       filename,
                                       offsetWithinFile,
                                       dataLength,
                                       data
                                      );
    end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.ReadRawVolumeDataBounded(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;
var
  preData: DWORD;
  postData: DWORD;
  tmpData: string;
  retVal: boolean;
begin
  preData := (offsetWithinFile mod ASSUMED_HOST_SECTOR_SIZE);
  // Yes, this is correct.
  // We use preData to avoid the need to add offsetWithinFile (a 64 bit value)
  // and then do the mod
  postData := ASSUMED_HOST_SECTOR_SIZE - ((preData + dataLength) mod ASSUMED_HOST_SECTOR_SIZE);

  retVal := ReadRawVolumeDataSimple(
                                    filename,
                                    (offsetWithinFile - preData),
                                    (preData + dataLength + postData),
                                    tmpData
                                   );
  if retVal then
    begin
    // Trim off the pre/post data garbage
    data := Copy(tmpData, preData, dataLength);
    
    // Overwrite...
    tmpData := StringOfChar(#0, (preData + dataLength + postData));
    end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.ReadRawVolumeDataSimple(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var Data: string
                        ): boolean;
var
  BytesReturned: DWORD;
  retVal: boolean;
  kmFilename: string;
  DIOCBufferIn: TDIOC_GET_RAW_DATA_IN;
  ptrDIOCBufferOut: PDIOC_GET_RAW_DATA_OUT;
begin
  retVal := FALSE;

  CheckActive();

  ptrDIOCBufferOut := allocmem(dataLength);
  try

    // Determine the filename as the kernel sees it
    kmFilename:= GetKernelModeVolumeFilename(filename);

    StrPCopy(DIOCBufferIn.Filename, kmFilename);
    DIOCBufferIn.Offset := offsetWithinFile;
    DIOCBufferIn.DataLength := dataLength;

  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Read raw data from: '+filename);
  {$ENDIF}
    if (DeviceIoControl(
                        DriverHandle,
                        IOCTL_FREEOTFE_GET_RAW,
                        @DIOCBufferIn,
                        sizeof(DIOCBufferIn),
                        ptrDIOCBufferOut,
                        dataLength,
                        BytesReturned,
                        nil
                        )) then
      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('DIOC OK (bytes: '+inttostr(BytesReturned)+')');
  {$ENDIF}
      if (BytesReturned = dataLength) then
        begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Bytes OK');
  {$ENDIF}
        // Set data so that it has enough characters which can be
        // overwritten with StrMove
        data := StringOfChar(#0, dataLength);
        StrMove(PChar(data), @ptrDIOCBufferOut.Data, dataLength);

        retVal := TRUE;
        end;

      end;
  finally
    FillChar(ptrDIOCBufferOut^, dataLength, 0);
    FreeMem(ptrDIOCBufferOut);
  end;


  Result := retVal;
end;


// ----------------------------------------------------------------------------
{$IFDEF FREEOTFE_DEBUG}
// Expose ReadRawVolumeCriticalData(...) for debug
function TOTFEFreeOTFE.DEBUGReadRawVolumeCriticalData(
                             filename: string;
                             offsetWithinFile: int64;
                             var criticalData: string
                        ): boolean;
begin
  Result := ReadRawVolumeCriticalData(
                         filename,
                         offsetWithinFile,
                         criticalData
                        );
end;
{$ENDIF}

// ----------------------------------------------------------------------------
{$IFDEF FREEOTFE_DEBUG}
// Expose WriteRawVolumeCriticalData(...) for debug
function TOTFEFreeOTFE.DEBUGWriteRawVolumeCriticalData(
                         filename: string;
                         offsetWithinFile: int64;
                         criticalData: string
                        ): boolean;
begin
  Result := WriteRawVolumeCriticalData(
                         filename,
                         offsetWithinFile,
                         criticalData
                        );
end;
{$ENDIF}

// ----------------------------------------------------------------------------
{$IFDEF FREEOTFE_DEBUG}
// Expose ReadRawVolumeData(...) for debug
function TOTFEFreeOTFE.DEBUGReadRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         dataLength: DWORD;  // In bytes
                         var data: string
                        ): boolean;
begin
  Result := ReadRawVolumeData(
                         filename,
                         offsetWithinFile,
                         dataLength,
                         data
                        );
end;
{$ENDIF}

// ----------------------------------------------------------------------------
{$IFDEF FREEOTFE_DEBUG}
// Expose WriteRawVolumeData(...) for debug
function TOTFEFreeOTFE.DEBUGWriteRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         data: string
                        ): boolean;
begin
  Result := WriteRawVolumeData(
                         filename,
                         offsetWithinFile,
                         data
                        );
end;
{$ENDIF}

// ----------------------------------------------------------------------------
{$IFDEF FREEOTFE_DEBUG}
// Expose ReadWritePlaintextToVolume(...) for debug
function TOTFEFreeOTFE.DEBUGReadWritePlaintextToVolume(
                    readNotWrite: boolean;

                    volFilename: string;
                    volumeKey: string;
                    sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
                    volumeIV: string;
                    IVHashDriver: string;
                    IVHashGUID: TGUID;
                    IVCypherDriver: string;
                    IVCypherGUID: TGUID;
                    mainCypherDriver: string;
                    mainCypherGUID: TGUID;
                    VolumeFlags: integer;
                    mountMountAs: TFreeOTFEMountAs;

                    dataOffset: int64;  // Offset from within mounted volume from where to read/write data
                    dataLength: integer;  // Length of data to read/write. In bytes
                    var data: string;  // Data to read/write

                    offset: int64 = 0;
                    size: int64 = 0;
                    storageMediaType: TFreeOTFEStorageMediaType = mtFixedMedia
                  ): boolean;
begin
  Result := ReadWritePlaintextToVolume(
                    readNotWrite,

                    volFilename,
                    volumeKey,
                    sectorIVGenMethod,
                    volumeIV,
                    IVHashDriver,
                    IVHashGUID,
                    IVCypherDriver,
                    IVCypherGUID,
                    mainCypherDriver,
                    mainCypherGUID,
                    VolumeFlags,
                    mountMountAs,

                    dataOffset,
                    dataLength,
                    data,

                    offset,
                    size,
                    storageMediaType
                  );
end;
{$ENDIF}

// ----------------------------------------------------------------------------
// Get the FreeOTFE driver to write a critical data block from the named file
// starting from the specified offset
// criticalData - This should be set to the string representation of a raw
//                (encrypted) critical data block
function TOTFEFreeOTFE.WriteRawVolumeCriticalData(
                         filename: string;
                         offsetWithinFile: int64;
                         criticalData: string
                        ): boolean;
begin
  if (length(criticalData) <> (CRITICAL_DATA_LENGTH div 8)) then
    begin
    EFreeOTFEInvalidParameter.Create('Invalid CDB passed in for raw writing');
    end;

  Result := WriteRawVolumeData(
                               filename,
                               offsetWithinFile,
                               criticalData
                              );
end;


// ----------------------------------------------------------------------------
// Get the FreeOTFE driver to write a critical data block from the named file
// starting from the specified offset
// criticalData - This should be set to the string representation of a raw
//                (encrypted) critical data block
function TOTFEFreeOTFE.WriteRawVolumeData(
                         filename: string;
                         offsetWithinFile: int64;
                         data: string
                        ): boolean;
var
  BytesReturned: DWORD;
  retVal: boolean;
  kmFilename: string;
  ptrDIOCBuffer: PDIOC_SET_RAW_DATA;
  bufferSize: integer;
begin
  retVal := FALSE;

  CheckActive();

  // Determine the volume filename as the kernel sees it
  kmFilename:= GetKernelModeVolumeFilename(filename);

  // Subtract sizeof(char) - once for the volume key, once for the volumeIV
  bufferSize := sizeof(TDIOC_SET_RAW_DATA)
                - sizeof(ptrDIOCBuffer^.Data) + Length(data);
  // Round up to next "DIOC boundry". Although this always results in an
  // oversized buffer, it's guaranteed to be big enough.
  bufferSize := bufferSize + (DIOC_BOUNDRY - (bufferSize mod DIOC_BOUNDRY));
  ptrDIOCBuffer := allocmem(bufferSize);
  try
    StrPCopy(ptrDIOCBuffer.Filename, kmFilename);
    ptrDIOCBuffer.Offset := offsetWithinFile;
    ptrDIOCBuffer.DataLength := Length(data);

    StrMove(@ptrDIOCBuffer.Data, PChar(data), Length(data));

  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Write raw data from: '+filename);
  {$ENDIF}
    if (DeviceIoControl(
                        DriverHandle,
                        IOCTL_FREEOTFE_SET_RAW,
                        ptrDIOCBuffer,
                        bufferSize,
                        nil,
                        0,
                        BytesReturned,
                        nil
                        )) then
      begin
  {$IFDEF FREEOTFE_DEBUG}
  DebugMsg('Written OK');
  {$ENDIF}
      retVal := TRUE;
      end;

  finally
    FillChar(ptrDIOCBuffer^, bufferSize, 0);
    FreeMem(ptrDIOCBuffer);
  end;

  Result := retVal;

end;


// ----------------------------------------------------------------------------
// Display dialog to allow user to select a partition
// Returns '' if the user cancels/on error, otherwise returns a partition
// identifier
function TOTFEFreeOTFE.SelectPartition(): string;
var
  dlg: TfrmSelectPartition;
  retVal: string;
begin
  retVal := '';

  dlg:= TfrmSelectPartition.Create(nil);
  try
    dlg.OTFEFreeOTFE := self;
    if (dlg.ShowModal() = mrOK) then
      begin
      retVal := dlg.Partition;
      end;

  finally
    dlg.Free();
  end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Generate a list of HDDs and partitions
function TOTFEFreeOTFE.HDDDeviceList(hddDevices: TStringList; title: TStringList): boolean;
var
  disk: integer;
  partition: integer;
  junk: string;
  currDevice: string;
  retVal: boolean;
  prevCursor: TCursor;
begin
  retVal:= TRUE;

  prevCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try

    for disk:=0 to MAX_HDDS do
      begin
      // Check to see if the entire disk exists, before checking for partitions
      currDevice := Format(PARTITION_DEVICE, [disk, 0]);
      if ReadRawVolumeCriticalData(
                                currDevice,
                                0,
                                junk
                               ) then
        begin
        // The entire disk
        hddDevices.Add(currDevice);
        title.Add('Hard disk #'+inttostr(disk)+' (entire disk)');

        // Note: We start from 1; partition 0 indicates the entire disk
        for partition:=1 to MAX_PARTITIONS do
          begin
          currDevice := Format(PARTITION_DEVICE, [disk, partition]);
          if ReadRawVolumeCriticalData(
                                    currDevice,
                                    0,
                                    junk
                                   ) then
            begin
            hddDevices.Add(currDevice);
            title.Add('Hard disk #'+inttostr(disk)+', partition #'+inttostr(partition));
            end;
          
          end;

        end;
      end;


  finally
    Screen.Cursor := prevCursor;
  end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
// Generate a list of HDDs and partitions
function TOTFEFreeOTFE.CDROMDeviceList(cdromDevices: TStringList; title: TStringList): boolean;
var
  cdrom: integer;
  junk: string;
  currDevice: string;
  retVal: boolean;
  prevCursor: TCursor;
begin
  retVal:= TRUE;

  prevCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try

    for cdrom:=0 to MAX_CDROMS do
      begin
      // Check to see if the entire disk exists, before checking for partitions
      currDevice := Format(CDROM_DEVICE, [cdrom]);
      if ReadRawVolumeCriticalData(
                                currDevice,
                                0,
                                junk
                               ) then
        begin
        cdromDevices.Add(currDevice);
        title.Add('CDROM drive #'+inttostr(cdrom));
        end;

      end;


  finally
    Screen.Cursor := prevCursor;
  end;

  Result := retVal;
end;


// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.CachesCreate();
begin
  fCachedVersionID := VERSION_ID_NOT_CACHED;

  fCachedHashDriverDetails := TStringList.Create();
  fCachedCypherDriverDetails := TStringList.Create();
end;

// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.CachesDestroy();
var
  i: integer;
begin
  fCachedVersionID := VERSION_ID_NOT_CACHED;

  for i:=0 to (fCachedHashDriverDetails.Count-1) do
    begin
    if (fCachedHashDriverDetails.Objects[i] <> nil) then
      begin
      Dispose(PFreeOTFEHashDriver(fCachedHashDriverDetails.Objects[i]));
      end;
    end;
  fCachedHashDriverDetails.Free();

  for i:=0 to (fCachedCypherDriverDetails.Count-1) do
    begin
    if (fCachedCypherDriverDetails.Objects[i] <> nil) then
      begin
      Dispose(PFreeOTFECypherDriver(fCachedCypherDriverDetails.Objects[i]));
      end;
    end;
  fCachedCypherDriverDetails.Free();

end;

// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.CachesFlush();
begin
  CachesDestroy();
  CachesCreate();
end;

// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.CachesAddHashDriver(hashKernelModeDeviceName: string; hashDriverDetails: TFreeOTFEHashDriver);
var
  ptrRec: PFreeOTFEHashDriver;
begin
  new(ptrRec);
  ptrRec^ := hashDriverDetails;
  fCachedHashDriverDetails.AddObject(hashKernelModeDeviceName, Pointer(ptrRec));
end;

// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.CachesGetHashDriver(hashKernelModeDeviceName: string; var hashDriverDetails: TFreeOTFEHashDriver): boolean;
var
  ptrRec: PFreeOTFEHashDriver;
  idx: integer;
  cacheHit: boolean;
begin
  cacheHit := FALSE;

  idx := fCachedHashDriverDetails.IndexOf(hashKernelModeDeviceName);
  if (idx >= 0) then
    begin
    ptrRec := PFreeOTFEHashDriver(fCachedHashDriverDetails.Objects[idx]);
    hashDriverDetails := ptrRec^;
    cacheHit := TRUE;
    end;

  Result := cacheHit;
end;

// ----------------------------------------------------------------------------
procedure TOTFEFreeOTFE.CachesAddCypherDriver(cypherKernelModeDeviceName: string; cypherDriverDetails: TFreeOTFECypherDriver);
var
  ptrRec: PFreeOTFECypherDriver;
begin
  new(ptrRec);
  ptrRec^ := cypherDriverDetails;
  fCachedCypherDriverDetails.AddObject(cypherKernelModeDeviceName, Pointer(ptrRec));
end;

// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.CachesGetCypherDriver(cypherKernelModeDeviceName: string; var cypherDriverDetails: TFreeOTFECypherDriver): boolean;
var
  ptrRec: PFreeOTFECypherDriver;
  idx: integer;
  cacheHit: boolean;
begin
  cacheHit := FALSE;

  idx := fCachedCypherDriverDetails.IndexOf(cypherKernelModeDeviceName);
  if (idx >= 0) then
    begin
    ptrRec := PFreeOTFECypherDriver(fCachedCypherDriverDetails.Objects[idx]);
    cypherDriverDetails := ptrRec^;
    cacheHit := TRUE;
    end;

  Result := cacheHit;
end;


// ----------------------------------------------------------------------------
// Check to see if there's at least one cypher and one hash algorithm
// available for use.
// Warn user and return FALSE if not.
// Returns TRUE if no problems
function TOTFEFreeOTFE.WarnIfNoHashOrCypherDrivers(): boolean;

var
  displayTitles: TStringList;
  kernelModeDriverNames: TStringList;
  hashGUIDs: TStringList;
  hashCount: integer;
  cypherCount: integer;
begin
  hashCount := 0;
  cypherCount := 0;

  displayTitles:= TStringList.Create();
  try
    kernelModeDriverNames:= TStringList.Create();
    try
      hashGUIDs:= TStringList.Create();
      try

        if (GetHashList(displayTitles, kernelModeDriverNames, hashGUIDs)) then
          begin
          hashCount := displayTitles.count;
          end;

        if (hashCount = 0) then
          begin
          SDUMessageDlg('You do not appear to have any FreeOTFE hash drivers '+CRLF+
                     'installed and started.'+CRLF+
                     CRLF+
                     'If you have only just installed FreeOTFE, you may need to '+CRLF+
                     'restart your computer.',
                     mtError,
                     [mbOK],
                     0);
          end;


        if (GetCypherList(displayTitles, kernelModeDriverNames, hashGUIDs)) then
          begin
          cypherCount := displayTitles.count;
          end;

        if (cypherCount = 0) then
          begin
          SDUMessageDlg('You do not appear to have any FreeOTFE cypher drivers '+CRLF+
                     'installed and started.'+CRLF+
                     CRLF+
                     'If you have only just installed FreeOTFE, you may need to '+CRLF+
                     'restart your computer.',
                     mtError,
                     [mbOK],
                     0);
          end;

      finally
        hashGUIDs.Free();
      end;
    finally
      kernelModeDriverNames.Free();
    end;
  finally
    displayTitles.Free();
  end;

  Result := (hashCount > 0) AND (cypherCount > 0);
end;


// ----------------------------------------------------------------------------
// Use the FreeOTFE device driver to mount the specified volume, and either:
//   *) Read in and decrypt specified data, returning the decrypted version
//   *) Encrypt specified data, and write to volume
// readNotWrite - Set to TRUE to read, FALSE to write
function TOTFEFreeOTFE.ReadWritePlaintextToVolume(
  readNotWrite: boolean;

  volFilename: string;
  volumeKey: string;
  sectorIVGenMethod: TFreeOTFESectorIVGenMethod;
  volumeIV: string;
  IVHashDriver: string;
  IVHashGUID: TGUID;
  IVCypherDriver: string;
  IVCypherGUID: TGUID;
  mainCypherDriver: string;
  mainCypherGUID: TGUID;
  VolumeFlags: integer;
  mountMountAs: TFreeOTFEMountAs;

  dataOffset: int64;  // Offset from within mounted volume from where to read/write data
  dataLength: integer;  // Length of data to read/write. In bytes
  var data: string;  // Data to read/write

  offset: int64 = 0;  // Offset within volume where encrypted data starts
  size: int64 = 0;  // Size of volume
  storageMediaType: TFreeOTFEStorageMediaType = mtFixedMedia
): boolean;
var
  allOK: boolean;
  deviceName: string;
begin
  LastErrorCode := OTFE_ERR_SUCCESS;

  CheckActive();

  // Attempt to create a new device to be used
  deviceName := CreateDiskDevice(FreeOTFEMountAsDeviceType[mountMountAs]);
  allOK := (deviceName <> '');
  if (allOK) then
    begin
    try
      // Attempt to mount the device
      allOK := MountDiskDevice(
                         deviceName,
                         volFilename,
                         volumeKey,
                         sectorIVGenMethod,
                         '',
                         FALSE,
                         IVHashDriver,
                         IVHashGUID,
                         IVCypherDriver,  // IV cypher
                         IVCypherGUID,  // IV cypher
                         mainCypherDriver,  // Main cypher
                         mainCypherGUID,  // Main cypher
                         VolumeFlags,
                         '',
                         offset,
                         size,
                         FreeOTFEMountAsStorageMediaType[mountMountAs]
                        );
      if (allOK) then
        begin
        try
          // Read decrypted data from mounted device
          if (readNotWrite) then
            begin
            allOK := ReadRawVolumeData(
                                      deviceName,
                                      dataOffset,
                                      dataLength,
                                      data
                                     );
            end
          else
            begin
            allOK := WriteRawVolumeData(
                                      deviceName,
                                      dataOffset,
                                      data
                                     );
            end;

        finally
          // Unmount
          DismountDiskDevice(deviceName, TRUE);
        end;

        end;  // if (allOK) then

    finally
      // Destroy device
      DestroyDiskDevice(deviceName);
    end;

    end;  // if (allOK) then

  Result := allOK;
end;


// ----------------------------------------------------------------------------
function TOTFEFreeOTFE.CanUserManageDrivers(): boolean;
var
  retval: boolean;
  testControlObj: TOTFEFreeOTFEDriverControl;
begin
  retval := FALSE;

  try
    // If we can't create a drive control object (i.e. we don't have admin
    // access to open the SCManager), we'll get an exception here.
    testControlObj:= TOTFEFreeOTFEDriverControl.Create();
    testControlObj.Free();
    retval := TRUE;
  except
    on EFreeOTFENeedAdminPrivs do
      begin
      // Do nothing - just swallow the exception
      end;

  end;

  Result := retval;
end;


// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

END.


